﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace EventExample
{
    class Program
    {
        //STEP 1: 
        //Derive a class from EventArgs to hold event related data
        public class myEventArgs : EventArgs
        {
            //Create some data to pass around in the event arguments
            //eventCounter counts the number of events called.
            public int eventCounter;

            //eventHandlerCounter counts the number of event 
            //handlers called.
            public static int eventHandlerCounter = 0;
        }

        //STEP 2:
        //Define a delegate type that specifies the signature of 
        //the event handler.
        public delegate void myEventHandler(object senderObj, myEventArgs e);

        //Create an event publisher class
        public class myPublisher
        {
            //Create a static event counter
            static int count = 0;

            //STEP 3:
            //Define an event based on the delegate type 
            //"myEventHandler" declared in step 2 
            //(field-like version)
            public event myEventHandler myEvent;

            //Define an event based on the delegate type 
            //"myEventHandler" declared in step 2 
            //(property-like version)
            /*
            private event myEventHandler myPrivateEvent;
            public event myEventHandler myEvent
            {
              add { myPrivateEvent += value; }
              remove { myPrivateEvent -= value; }
            }
            */

            //STEP 4: 
            //Define a protected virtual method that raises
            //the event declared in step 3.
            protected virtual void OnEvent()
            {
                //Before raising the event, make sure there is
                //at least one registered subscriber for the event 
                try
                {
                    if (myEvent != null)
                    {
                        //Populate the derived argument class with event 
                        //specific data.
                        myEventArgs arg = new myEventArgs();

                        //Update the event counter and 
                        arg.eventCounter = ++count;
                        Console.WriteLine("EVENT: {0}", arg.eventCounter);

                        //raise the event by calling the associated
                        //delegate to multicast every registered
                        //subscriber event handler method in its
                        //invocation list.                    
                        myEvent(this, arg);
                    }
                }
                catch
                {
                    //Handle exceptions here
                }
            }

            //STEP 5:            
            //Define a public method to run the protected, virtual 
            //method declared in Step 4 for when the event occurs.
            public void RunOnEvent()
            {
                OnEvent();
            }
        }

        //SubscriberA contains an instance event handler as well   
        //as a way to internally register and unregister for the   
        //event along with an event handler.
        public class SubscriberA
        {
            private myPublisher currPub;

            public SubscriberA(myPublisher p)
            {
                currPub = p;
            }

            //STEP 6:
            //In each subscriber class, create its own event  
            //handling method with the same signature as the 
            //publisher's delegate.
            public void SubscriberA_EventHandlerA(object senderObj,
            myEventArgs e)
            {
                Console.WriteLine("Processing event handler in SubscriberA (instance method)");
            }

            //STEP 8:
            //Register the event
            public void register()
            {
                currPub.myEvent += this.SubscriberA_EventHandlerA;

                //Update event handler counter and display message on 
                //the screen.
                myEventArgs.eventHandlerCounter++;
                Console.WriteLine("\nRegistering SubscriberA. Event handlers registered:{0}", myEventArgs.eventHandlerCounter);
            }

            //STEP 9:
            //Unregister the event
            public void unregister()
            {
                currPub.myEvent -= this.SubscriberA_EventHandlerA;

                //Update event handler counter and display message on 
                //the screen.
                myEventArgs.eventHandlerCounter--;
                Console.WriteLine("\nUnregistering SubscriberA. Event handlers registered:{0}", myEventArgs.eventHandlerCounter);
            }
        }

        //SubscriberB contains an instance event handler. 
        //Registering and unregistering for the event must be 
        //done externally.
        public class SubscriberB
        {
            //STEP 6:
            //In each subscriber class, create its own event  
            //handling method with the same signature as the 
            //publisher's delegate.
            public void SubscriberB_EventHandlerB(object senderObj,
            myEventArgs e)
            {
                Console.WriteLine("Processing event handler in SubscriberB (instance method)");
            }
        }

        //SubscriberC contains a static handler for the event.
        public class SubscriberC
        {
            //STEP 6:
            //In each subscriber class, create its own event   
            //handling method with the same signature as the 
            //publisher's delegate.
            public static void SubscriberC_EventHandlerC(object
            senderObj, myEventArgs e)
            {
                Console.WriteLine("Processing event handler in SubscriberC (static method)");
            }
        }

        public class MyClass
        {
            public static void Main()
            {
                //STEP 7:
                //Instantiate the publisher object
                myPublisher myPublisherObj = new myPublisher();

                //STEP 7:
                //Instantiate a subscriber object named subA
                SubscriberA subA = new SubscriberA(myPublisherObj);

                //STEP 7:
                //Instantiate another subscriber object named subB
                SubscriberB subB = new SubscriberB();

                //STEP 8:
                //Register subscriber object subA for the event
                subA.register();

                //STEP 8:
                //Register subscriber object subB for the event either 
                //this way:
                myPublisherObj.myEvent += subB.SubscriberB_EventHandlerB;
                //or this way:
                //myPublisherObj.myEvent += new myEventHandler(subB.SubscriberB_EventHandlerB);

                //Update event handler counter and display message on 
                //the screen.
                myEventArgs.eventHandlerCounter++;
                Console.WriteLine("Registering SubscriberB. Event handlers registered: {0}", myEventArgs.eventHandlerCounter);

                //STEP 8:
                //Register event handler for subscriber class SubscriberC.
                myPublisherObj.myEvent +=
                SubscriberC.SubscriberC_EventHandlerC;

                //Update event handler counter and display message on 
                //the screen.
                myEventArgs.eventHandlerCounter++;
                Console.WriteLine("Registering SubscriberC. Event handlers registered: {0}", myEventArgs.eventHandlerCounter);

                //Raise an event for all the currently registered 
                //event handlers
                Console.WriteLine("\nRaise event for all {0} currently registered subscribers.\n", myEventArgs.eventHandlerCounter);
                myPublisherObj.RunOnEvent();

                //STEP 9:
                //Unregister event handler for subscriber object subA
                subA.unregister();

                //STEP 9:
                //Unregister event handler for SubscriberB object subB
                myPublisherObj.myEvent -= subB.SubscriberB_EventHandlerB;

                //Update event handler counter and display message on screen.
                myEventArgs.eventHandlerCounter--;
                Console.WriteLine("Unregistering SubscriberB. Event handlers registered: {0}", myEventArgs.eventHandlerCounter);

                //Raise an event for all the currently registered event 
                //handlers
                Console.WriteLine("\nRaise event for all {0} currently registered subscribers.\n", myEventArgs.eventHandlerCounter);
                myPublisherObj.RunOnEvent();

                Console.ReadLine();
            }
        }
    }
}
