﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ComplexNumberDemo
{
    class Program
    {

        #region Complex Number struct

        /////////////////////////////////////////////////////
        // The complex struct represents a complex number  //  
        // of the general format z = x + iy where x = Re(z)// 
        // = real part and y = Im(z) = imaginary part. A   //  
        // complex instance has public double-precision    //    
        // floating point data members real and imag.      //
        /////////////////////////////////////////////////////
        [Serializable]
        public struct Complex
        {
            /////////////////////////////////////////////////
            ////////////// Instance Variables ///////////////       
            /////////////////////////////////////////////////
            private double real;
            private double imag;

            /////////////////////////////////////////////////
            //////////////// Accessor Methods////////////////
            /////////////////////////////////////////////////
            public double Real
            {
                get { return (real); }
                set { real = value; }
            }

            public double Imag
            {
                get { return (imag); }
                set { imag = value; }
            }

            /////////////////////////////////////////////////
            ///////////////// Constructors //////////////////
            /////////////////////////////////////////////////
            public Complex(double x, double y)
            {
                this.real = x;
                this.imag = y;
            }

            public Complex(double x)
            {
                this.real = x;
                this.imag = 0.0;
            }

            /////////////////////////////////////////////////
            // Implicit and Explicit conversion operators ///
            /////////////////////////////////////////////////            
            public static implicit operator double(Complex z)
            {
                return z.real;
            }

            public static explicit operator Complex(double x)
            {
                return (new Complex(x, 0.0));
            }

            /////////////////////////////////////////////////
            ///////////// Complex Constants /////////////////
            /////////////////////////////////////////////////        
            public static Complex CZero = new Complex(0.0, 0.0);
            public static Complex COne = new Complex(1.0, 0.0);
            public static Complex i = new Complex(0.0, 1.0);
            public static Complex CNaN = new Complex(double.NaN, double.NaN);
            public static Complex CInfinity = new Complex(double.PositiveInfinity, double.PositiveInfinity);

            /////////////////////////////////////////////////
            ///////////// Complex Booleans //////////////////
            /////////////////////////////////////////////////
            public bool IsCZero
            { get { return ((real == 0.0) && (imag == 0.0)); } }

            public bool IsCOne
            { get { return ((real == 1.0) && (imag == 0.0)); } }

            public bool Isi
            { get { return ((real == 0.0) && (imag == 1.0)); } }

            public bool IsCNaN
            { get { return double.IsNaN(real) || double.IsNaN(imag); } }

            public bool IsCInfinity
            { get { return double.IsInfinity(real) || double.IsInfinity(imag); } }

            public bool IsReal
            { get { return (imag == 0.0); } }

            public bool IsImag
            { get { return (real == 0.0); } }

            ////////////////////////////////////////
            ////////////Complex Methods ////////////
            ////////////////////////////////////////
            // Returns the argument of a complex number.
            public static double CArg(Complex z)
            {
                return (Math.Atan2(z.imag, z.real));
            }

            // Returns the conjugate of a complex number z.
            public static Complex CConj(Complex z)
            {
                return (new Complex(z.real, -z.imag));
            }

            // Returns the norm (or modulus or absolute value) 
            // of a complex number.
            public static double CNorm(Complex z)
            {
                return (Math.Sqrt((z.real * z.real) + (z.imag * z.imag)));
            }

            // Returns the norm (or modulus) of a complex number 
            // avoiding potential overflow and/or underflow for
            // very small or very large numbers.
            public static double CNorm2(Complex z)
            {
                double x = z.real;
                double y = z.imag;

                if (Math.Abs(x) < Math.Abs(y))
                {
                    return (Math.Abs(y) * Math.Sqrt(1.0 + (x / y) * (x / y)));
                }
                else
                {
                    return (Math.Abs(x) * Math.Sqrt(1.0 + (y / x) * (y / x)));
                }
            }

            // Returns the inverse of a complex number.
            public static Complex Inv(Complex z)
            {
                return (1.0 / z);
            }

            // Returns the real part of a complex number.
            public static double Re(Complex z)
            {
                return (z.real);
            }

            // Returns the imaginary part of a complex number.
            public static double Im(Complex z)
            {
                return (z.imag);
            }

            // Converts:(r,theta) ----> (x,y)         
            public static Complex FromPolarToXY(double r, double theta)
            {
                return (new Complex((r * Math.Cos(theta)), (r * Math.Sin(theta))));
            }

            // Converts:    (x,y) ----> (r,theta)
            public static Complex FromXYToPolar(Complex z)
            {
                return (new Complex(CNorm(z), CArg(z)));
            }

            // Returns the negation of a complex number.
            public static Complex CNeg(Complex z)
            {
                return (-z);
            }

            // Returns the sum of two complex numbers z1 and z2.
            public static Complex CAdd(Complex z1, Complex z2)
            {
                return (z1 + z2);
            }

            // Returns the sum of a real number with a complex number.
            public static Complex CAdd(double x, Complex z)
            {
                return (x + z);
            }

            // Returns the sum of a complex number with a real number.
            public static Complex CAdd(Complex z, double x)
            {
                return (z + x);
            }

            // Returns the difference between two complex numbers z1 and z2.
            public static Complex CSub(Complex z1, Complex z2)
            {
                return (z1 - z2);
            }

            // Returns the difference between a real number and a complex number.
            public static Complex CSub(double x, Complex z)
            {
                return (x - z);
            }

            // Returns the difference between a complex number and a real number.
            public static Complex CSub(Complex z, double x)
            {
                return (z - x);
            }

            // Returns the product between two complex numbers z1 and z2.
            public static Complex CMult(Complex z1, Complex z2)
            {
                return (z1 * z2);
            }

            // Returns the product of a real number with a complex number.
            public static Complex CMult(double x, Complex z)
            {
                return (x * z);
            }

            // Returns the product of a complex number and a real number.
            public static Complex CMult(Complex z, double x)
            {
                return (z * x);
            }

            // Returns the quotient of dividing two complex numbers z1 and z2.
            public static Complex CDiv(Complex z1, Complex z2)
            {
                return (z1 / z2);
            }

            // Returns the quotient of dividing a real number by a complex number.
            public static Complex CDiv(double x, Complex z)
            {
                return (x / z);
            }

            // Returns the quotient of dividing a complex number by a real number.
            public static Complex CDiv(Complex z, double x)
            {
                return (z / x);
            }

            // Returns the quotient of dividing two complex 
            // numbers z1 and z2 avoiding potential underflow
            // and/or overflow for very small or very large numbers
            public static Complex CDiv2(Complex z1, Complex z2)
            {
                double x1 = z1.real; double y1 = z1.imag;
                double x2 = z2.real; double y2 = z2.imag;

                Complex u;
                double denom;

                if (z2.IsCZero) return Complex.CInfinity;

                if (Math.Abs(x2) < Math.Abs(y2))
                {
                    denom = x2 * (x2 / y2) + y2;
                    u.real = (x1 * (x2 / y2) + y1) / denom;
                    u.imag = (y1 * (x2 / y2) - x1) / denom;
                }
                else
                {
                    denom = x2 + y2 * (y2 / x2);
                    u.real = (x1 + y1 * (y2 / x2)) / denom;
                    u.imag = (y1 - x1 * (y2 / x2)) / denom;
                }
                return u;
            }

            ////////////////////////////////////////
            /////////Overloaded Operators //////////
            ////////////////////////////////////////
            public static Complex operator +(Complex z)
            {
                return z;
            }

            public static Complex operator +(Complex z1, Complex z2)
            {
                return (new Complex(z1.real + z2.real, z1.imag + z2.imag));
            }

            // Returns the sum of a real number with a complex number.
            public static Complex operator +(double x, Complex z)
            {
                return (new Complex(x + z.real, z.imag));
            }

            // Returns the sum of a complex number with a real number.
            public static Complex operator +(Complex z, double x)
            {
                return (new Complex(z.real + x, z.imag));
            }

            // Returns the negation of a complex number.
            public static Complex operator -(Complex z)
            {
                return (new Complex(-z.real, -z.imag));
            }

            // Returns the difference between two complex numbers.
            public static Complex operator -(Complex z1, Complex z2)
            {
                return (new Complex(z1.real - z2.real, z1.imag - z2.imag));
            }

            // Returns the difference of a real number with a complex number.
            public static Complex operator -(double x, Complex z)
            {
                return (new Complex(x - z.real, -z.imag));
            }

            // Returns the difference of a complex number with a real number.
            public static Complex operator -(Complex z, double x)
            {
                return (new Complex(z.real - x, z.imag));
            }

            // Returns the product of two complex numbers z1 * z2.
            public static Complex operator *(Complex z1, Complex z2)
            {
                double x = (z1.real * z2.real) - (z1.imag * z2.imag);
                double y = (z1.real * z2.imag) + (z1.imag * z2.real);
                return (new Complex(x, y));
            }

            // Returns the product of a real and a complex number.
            public static Complex operator *(double x, Complex z)
            {
                return (new Complex(x * z.real, x * z.imag));
            }

            // Returns the product of a complex number and a real.
            public static Complex operator *(Complex z, double x)
            {
                return (new Complex(z.real * x, z.imag * x));
            }

            // Returns the quotient of two complex numbers z1 / z2.
            public static Complex operator /(Complex z1, Complex z2)
            {
                if (z2.IsCZero) return Complex.CInfinity;
                double denom = (double)(Math.Pow(z2.real, 2.0) + Math.Pow(z2.imag, 2.0));
                double x = ((z1.real * z2.real) + (z1.imag * z2.imag)) / denom;
                double y = ((z1.imag * z2.real) - (z1.real * z2.imag)) / denom;
                return (new Complex(x, y));
            }

            // Returns the quotient of dividing a real number by a complex number.
            public static Complex operator /(double x, Complex z)
            {
                if (z.IsCZero) return Complex.CInfinity;
                double denom = (double)(Math.Pow(z.real, 2.0) + Math.Pow(z.imag, 2.0));
                double re = (x * z.real) / denom;
                double im = (0.0 - (x * z.imag)) / denom;
                return (new Complex(re, im));
            }

            // Returns the quotient of dividing a complex number by a real number.
            public static Complex operator /(Complex z, double x)
            {
                if (x == 0.0) return Complex.CInfinity;
                double re = z.real / x;
                double im = z.imag / x;
                return (new Complex(re, im));
            }

            // Tests for equality of two complex numbers z1 and z2.
            public static bool operator ==(Complex z1, Complex z2)
            {
                return ((z1.real == z2.real) && (z1.imag == z2.imag));
            }

            // Tests for inequality of two complex numbers z1 and z2.
            public static bool operator !=(Complex z1, Complex z2)
            {
                return (!(z1 == z2));
            }

            // Tests for equality of this complex number and another complex number.
            public override bool Equals(Object obj)
            {
                return ((obj is Complex) && (this == (Complex)obj));
            }

            // Returns an integer hash code for this complex number.
            // If you override Equals, override GetHashCode too.
            public override int GetHashCode()
            {
                //return this.ToString().GetHashCode();
                return (real.GetHashCode() ^ imag.GetHashCode());
            }

            // Returns a formatted string representation in 
            // the form z = x + iy for a complex number.
            public override string ToString()
            {
                return (String.Format("{0} + {1}i", real, imag));
            }
            
            /////////////////////////////////////////
            //////////Exponential Functions /////////
            /////////////////////////////////////////
            public static Complex CExp(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double expx = Math.Exp(x);
                return (new Complex(expx * Math.Cos(y),
                                    expx * Math.Sin(y)));
            }

            /////////////////////////////////////////
            //////////Logarithmic Functions /////////
            /////////////////////////////////////////
            // Logarithm of complex z to base e
            public static Complex CLog(Complex z)
            {
                return (new Complex(Math.Log(Complex.CNorm(z)),
                                    Math.Atan2(z.Imag, z.Real)));
            }

            // Another version of logarithm of complex z to base e
            public static Complex CLog2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double re = 0.5 * Math.Log(x * x + y * y);
                double im = Math.Atan2(y, x);
                return (new Complex(re, im));
            }

            // Logarithm of complex z to base 10
            public static Complex CLog10(Complex z)
            {
                return (Complex.CLog(z) / Complex.CLog((Complex)10.0));
            }

            // Logarithm of complex z1 to complex base z2
            public static Complex CLogb(Complex z1, Complex z2)
            {
                return (Complex.CLog(z1) / Complex.CLog(z2));
            }

            // Logarithm of real x to complex base z2
            public static Complex CLogb(double x, Complex z2)
            {
                return (Complex.CLog((Complex)x) / Complex.CLog(z2));
            }

            // Logarithm of complex z1 to real base x
            public static Complex CLogb(Complex z1, double x)
            {
                return (Complex.CLog(z1) / Complex.CLog((Complex)x));
            }

            /////////////////////////////////////////
            ////////////Power Functions /////////////
            /////////////////////////////////////////
            // Complex z raised to the power of complex w
            public static Complex CPow(Complex z, Complex w)
            {
                return (Complex.CExp(w * Complex.CLog(z)));
            }

            // Complex z raised to the power of complex w (ver 2)
            public static Complex CPow2(Complex z, Complex w)
            {
                double x1 = z.Real; double y1 = z.Imag;
                double x2 = w.Real; double y2 = w.Imag;

                double r1 = Math.Sqrt(x1 * x1 + y1 * y1);
                double theta1 = Math.Atan2(y1, x1);
                double phi = theta1 * x2 + y2 * Math.Log(r1);

                double re = Math.Pow(r1, x2) * Math.Exp(-theta1 * y2) * Math.Cos(phi);
                double im = Math.Pow(r1, x2) * Math.Exp(-theta1 * y2) * Math.Sin(phi);

                return (new Complex(re, im));
            }

            // Complex z raised to the power of real x 
            public static Complex CPow(Complex z, double x)
            {
                return (Complex.CExp(x * Complex.CLog(z)));
            }

            // Complex z raised to the power of real x (ver 2)
            public static Complex CPow2(Complex z, double x)
            {
                double x1 = z.Real;
                double y1 = z.Imag;

                double r1 = Math.Sqrt(x1 * x1 + y1 * y1);
                double theta1 = Math.Atan2(y1, x1);
                double phi = theta1 * x;

                double re = Math.Pow(r1, x) * Math.Cos(phi);
                double im = Math.Pow(r1, x) * Math.Sin(phi);

                return (new Complex(re, im));
            }

            // Real x raised to the power of complex z
            public static Complex CPow(double x, Complex z)
            {
                return (Complex.CExp(z * Math.Log(x)));
            }

            // Real x raised to the power of complex z (ver 2)
            public static Complex CPow2(double x, Complex w)
            {
                double x2 = w.Real;
                double y2 = w.Imag;

                double r1 = Math.Sqrt(x * x);
                double theta1 = Math.Atan2(0.0, x);
                double phi = theta1 * x2 + y2 * Math.Log(r1);

                double re = Math.Pow(r1, x2) * Math.Cos(phi);
                double im = Math.Pow(r1, x2) * Math.Sin(phi);

                return (new Complex(re, im));
            }

            ////////////////////////////////////////
            ////////////Root Functions /////////////
            ////////////////////////////////////////
            // Complex root w of complex number z
            public static Complex CRoot(Complex z, Complex w)
            {
                return (Complex.CExp(Complex.CLog(z) / w));
            }

            // Real root x of complex number z
            public static Complex CRoot(Complex z, double x)
            {
                return (Complex.CExp(Complex.CLog(z) / x));
            }

            // Complex root z of real number x
            public static Complex CRoot(double x, Complex z)
            {
                return (Complex.CExp(Math.Log(x) / z));
            }

            // Complex square root of complex number z
            public static Complex CSqrt(Complex z)
            {
                return (Complex.CExp(Complex.CLog(z) / 2.0));
            }

            /////////////////////////////////////////
            ////Complex Trigonometric Functions /////
            /////////////////////////////////////////   
            // Complex sine of complex number z
            public static Complex CSin(Complex z)
            {
                return ((Complex.CExp(i * z) - Complex.CExp(-i * z)) / (2.0 * i));
            }

            // Complex sine of complex number z (ver 2)
            public static Complex CSin2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double re = Math.Sin(x) * Math.Cosh(y);
                double im = Math.Cos(x) * Math.Sinh(y);
                return (new Complex(re, im));
            }

            // Complex cosine of complex number z 
            public static Complex CCos(Complex z)
            {
                return ((Complex.CExp(i * z) + Complex.CExp(-i * z)) / 2.0);
            }

            // Complex cosine of complex number z (ver 2)
            public static Complex CCos2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double re = Math.Cos(x) * Math.Cosh(y);
                double im = -Math.Sin(x) * Math.Sinh(y);
                return (new Complex(re, im));
            }

            // Complex tangent of complex number z 
            public static Complex CTan(Complex z)
            {
                return (Complex.CSin(z) / Complex.CCos(z));
            }

            // Complex tangent of complex number z (ver 2)
            public static Complex CTan2(Complex z)
            {
                double x2 = 2.0 * z.Real;
                double y2 = 2.0 * z.Imag;
                double denom = Math.Cos(x2) + Math.Cosh(y2);
                if (denom == 0.0) return Complex.CInfinity;
                double re = Math.Sin(x2) / denom;
                double im = Math.Sinh(y2) / denom;
                return (new Complex(re, im));
            }

            // Complex cotangent of complex number z 
            public static Complex CCot(Complex z)
            {
                return (Complex.CCos(z) / Complex.CSin(z));
            }

            // Complex cotangent of complex number z (ver 2)
            public static Complex CCot2(Complex z)
            {
                double x2 = 2.0 * z.Real;
                double y2 = 2.0 * z.Imag;
                double denom = Math.Cosh(y2) - Math.Cos(x2);
                if (denom == 0.0) return Complex.CInfinity;
                double re = Math.Sin(x2) / denom;
                double im = -Math.Sinh(y2) / denom;
                return (new Complex(re, im));
            }

            // Complex secant of complex number z 
            public static Complex CSec(Complex z)
            {
                return (1.0 / Complex.CCos(z));
            }

            // Complex secant of complex number z (ver 2)
            public static Complex CSec2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double denom = Math.Cos(x) * Math.Cos(x) + Math.Sinh(y) * Math.Sinh(y);
                if (denom == 0.0) return Complex.CInfinity;
                double re = Math.Cos(x) * Math.Cosh(y) / denom;
                double im = Math.Sin(x) * Math.Sinh(y) / denom;
                return (new Complex(re, im));
            }

            // Complex cosecant of complex number z 
            public static Complex CCsc(Complex z)
            {
                return (1.0 / Complex.CSin(z));
            }

            // Complex cosecant of complex number z (ver 2)
            public static Complex CCsc2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double denom = Math.Sin(x) * Math.Sin(x) +
                               Math.Sinh(y) * Math.Sinh(y);
                if (denom == 0.0) return Complex.CInfinity;
                double re = Math.Sin(x) * Math.Cosh(y) / denom;
                double im = -Math.Cos(x) * Math.Sinh(y) / denom;
                return (new Complex(re, im));
            }

            ///////////////////////////////////////////////
            ///Complex Inverse Trigonometric Functions ////
            ///////////////////////////////////////////////   
            // Complex ArcSine of complex number z
            public static Complex CArcSin(Complex z)
            {
                return (-i * Complex.CLog((i * z) + Complex.CSqrt(1.0 - (z * z))));
            }

            // Complex ArcSine of complex number z (ver 2)
            public static Complex CArcSin2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;

                double ysqd = y * y;
                double rtpos = Math.Sqrt(Math.Pow(x + 1.0, 2.0) + ysqd);
                double rtneg = Math.Sqrt(Math.Pow(x - 1.0, 2.0) + ysqd);
                double alpha = 0.5 * (rtpos + rtneg);
                double beta = 0.5 * (rtpos - rtneg);

                double InvSinZRe = Math.Asin(beta);
                double InvSinZIm = Math.Sign(y) * Math.Log(alpha + Math.Sqrt(alpha * alpha - 1.0));

                return (new Complex(InvSinZRe, InvSinZIm));
            }

            // Complex ArcCosine of complex number z
            public static Complex CArcCos(Complex z)
            {
                return (-i * Complex.CLog(z + i * Complex.CSqrt(1.0 - (z * z))));
            }

            // Complex ArcCosine of complex number z (ver 2)
            public static Complex CArcCos2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;

                double ysqd = y * y;
                double rtpos = Math.Sqrt(Math.Pow(x + 1.0, 2.0) + ysqd);
                double rtneg = Math.Sqrt(Math.Pow(x - 1.0, 2.0) + ysqd);
                double alpha = 0.5 * (rtpos + rtneg);
                double beta = 0.5 * (rtpos - rtneg);

                double InvCosZRe = Math.Acos(beta);
                double InvCosZIm = -Math.Sign(y) * Math.Log(alpha + Math.Sqrt(alpha * alpha - 1.0));

                return (new Complex(InvCosZRe, InvCosZIm));
            }

            // Complex ArcTangent of complex number z
            public static Complex CArcTan(Complex z)
            {
                return ((i / 2.0) * Complex.CLog((i + z) / (i - z)));
            }

            // Complex ArcTangent of complex number z (ver 2)
            public static Complex CArcTan2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;

                double xsqd = x * x;
                double ysqd = y * y;

                double InvTanZRe = 0.5 * Math.Atan2(2.0 * x, 1.0 - xsqd - ysqd);
                double InvTanZIm = 0.25 * Math.Log((xsqd + Math.Pow(y + 1.0, 2.0)) / (xsqd + Math.Pow(y - 1.0, 2.0)));

                return (new Complex(InvTanZRe, InvTanZIm));
            }

            // Complex ArcCotangent of complex number z
            public static Complex CArcCot(Complex z)
            {
                return (Complex.CArcTan(1.0 / z));
            }

            // Complex ArcCotangent of complex number z (ver 2)
            public static Complex CArcCot2(Complex z)
            {
                return ((i / 2.0) * (Complex.CLog((z - i) / (z + i))));
            }

            // Complex ArcSecant of complex number z
            public static Complex CArcSec(Complex z)
            {
                return (Complex.CArcCos(1.0 / z));
            }

            // Complex ArcSecant of complex number z (ver 2)
            public static Complex CArcSec2(Complex z)
            {
                return (i * Complex.CLog((Complex.CSqrt(1.0 - (z * z)) + 1.0) / z));
            }

            // Complex ArcCosecant of complex number z 
            public static Complex CArcCsc(Complex z)
            {
                return (Complex.CArcSin(1.0 / z));
            }

            // Complex ArcCosecant of complex number z (ver 2)
            public static Complex CArcCsc2(Complex z)
            {
                return (-i * Complex.CLog((Complex.CSqrt((z * z) - 1.0) + i) / z));
            }

            //////////////////////////////////////////
            //////Complex Hyperbolic Functions ///////
            //////////////////////////////////////////
            // Complex hyperbolic sine of complex number z
            public static Complex CSinh(Complex z)
            {
                return ((Complex.CExp(z) - Complex.CExp(-z)) / 2.0);
            }

            // Complex Hyperbolic Sine 
            // of complex number z (ver 2)
            public static Complex CSinh2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double SinhZRe = Math.Sinh(x) * Math.Cos(y);
                double SinhZIm = Math.Cosh(x) * Math.Sin(y);
                return (new Complex(SinhZRe, SinhZIm));
            }

            // Complex Hyperbolic Sine of complex number z (ver 3)
            public static Complex CSinh3(Complex z)
            {
                return (-i * Complex.CSin(i * z));
            }

            // Complex Hyperbolic Cosine of complex number z 
            public static Complex CCosh(Complex z)
            {
                return ((Complex.CExp(z) + Complex.CExp(-z)) / 2.0);
            }

            // Complex Hyperbolic Cosine of complex number z (ver 2)
            public static Complex CCosh2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double CoshZRe = Math.Cosh(x) * Math.Cos(y);
                double CoshZIm = Math.Sinh(x) * Math.Sin(y);
                return (new Complex(CoshZRe, CoshZIm));
            }

            // Complex Hyperbolic Cosine of complex number z (ver 3)
            public static Complex CCosh3(Complex z)
            {
                return (Complex.CCos(i * z));
            }

            // Complex Hyperbolic Tangent of complex number z 
            public static Complex CTanh(Complex z)
            {
                return (Complex.CSinh(z) / Complex.CCosh(z));
            }

            // Complex Hyperbolic Tangent of complex number z (ver 2)
            public static Complex CTanh2(Complex z)
            {
                double twox = 2.0 * z.Real;
                double twoy = 2.0 * z.Imag;
                double denom = Math.Cosh(twox) + Math.Cos(twoy);

                double TanhZRe = Math.Sinh(twox) / denom;
                double TanhZIm = Math.Sin(twoy) / denom;
                return (new Complex(TanhZRe, TanhZIm));
            }

            // Complex Hyperbolic Tangent of complex number z (ver 3)
            public static Complex CTanh3(Complex z)
            {
                return (-i * Complex.CTan(i * z));
            }

            // Complex Hyperbolic Cotangent of complex number z 
            public static Complex CCoth(Complex z)
            {
                return (Complex.CCosh(z) / Complex.CSinh(z));
            }

            // Complex Hyperbolic Cotangent of complex number z (ver 2)
            public static Complex CCoth2(Complex z)
            {
                return (Complex.CCosh2(z) / Complex.CSinh2(z));
            }

            // Complex Hyperbolic Cotangent of complex number z (ver 3)
            public static Complex CCoth3(Complex z)
            {
                double twox = 2.0 * z.Real;
                double twoy = 2.0 * z.Imag;
                double denom = Math.Cosh(twox) - Math.Cos(twoy);

                double CothZRe = Math.Sinh(twox) / denom;
                double CothZIm = -Math.Sin(twoy) / denom;
                return (new Complex(CothZRe, CothZIm));
            }

            // Complex Hyperbolic Cotangent of complex number z (ver 4)
            public static Complex CCoth4(Complex z)
            {
                return (i * Complex.CCot(i * z));
            }

            // Complex Hyperbolic Secant of complex number z
            public static Complex CSech(Complex z)
            {
                return (1.0 / Complex.CCosh(z));
            }

            // Complex Hyperbolic Secant of complex number z (ver 2)
            public static Complex CSech2(Complex z)
            {
                return (1.0 / Complex.CCosh2(z));
            }

            // Complex Hyperbolic Secant of complex number z (ver 3)
            public static Complex CSech3(Complex z)
            {
                double CoshX = Math.Cosh(z.Real);
                double CosY = Math.Cos(z.Imag);
                double SinhX = Math.Sinh(z.Real);
                double SinY = Math.Sin(z.Imag);

                double denom = CosY * CosY + SinhX * SinhX;

                double CSechZRe = (CoshX * CosY) / denom;
                double CSechZIm = -(SinhX * SinY) / denom;
                return (new Complex(CSechZRe, CSechZIm));
            }

            // Complex Hyperbolic Secant of complex number z (ver 4)
            public static Complex CSech4(Complex z)
            {
                return (Complex.CSec(i * z));
            }

            // Complex Hyperbolic Cosecant of complex number z 
            public static Complex CCsch(Complex z)
            {
                return (1.0 / Complex.CSinh(z));
            }

            // Complex Hyperbolic Cosecant of complex number z (ver 2)
            public static Complex CCsch2(Complex z)
            {
                return (1.0 / Complex.CSinh2(z));
            }

            // Complex Hyperbolic Cosecant of complex number z (ver 3)
            public static Complex CCsch3(Complex z)
            {
                double CoshX = Math.Cosh(z.Real);
                double CosY = Math.Cos(z.Imag);
                double SinhX = Math.Sinh(z.Real);
                double SinY = Math.Sin(z.Imag);

                double denom = SinY * SinY + SinhX * SinhX;

                double CSechZRe = (SinhX * CosY) / denom;
                double CSechZIm = -(CoshX * SinY) / denom;
                return (new Complex(CSechZRe, CSechZIm));
            }

            // Complex Hyperbolic Cosecant of complex number z (ver 4)
            public static Complex CCsch4(Complex z)
            {
                return (i * Complex.CCsc(i * z));
            }

            /////////////////////////////////////////////
            ////Complex Inverse Hyperbolic Functions ////
            /////////////////////////////////////////////
            // Complex Inverse Hyperbolic Sine of complex number z
            public static Complex CArcSinh(Complex z)
            {
                return (Complex.CLog(z + Complex.CSqrt((z * z) + 1.0)));
            }

            // Complex Inverse Hyperbolic Sine of complex number z (ver 2)
            public static Complex CArcSinh2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;

                double xsqd = x * x;
                double rtpos = Math.Sqrt(Math.Pow(y - 1.0, 2.0) + xsqd);
                double rtneg = Math.Sqrt(Math.Pow(y + 1.0, 2.0) + xsqd);
                double alphap = 0.5 * (rtpos + rtneg);
                double betap = 0.5 * (rtpos - rtneg);

                double InvSinhZRe = Math.Sign(x) *
                    Math.Log(alphap + Math.Sqrt(alphap * alphap - 1));
                double InvSinhZIm = -Math.Asin(betap);

                return (new Complex(InvSinhZRe, InvSinhZIm));
            }

            // Complex Inverse Hyperbolic Cosine of complex number z
            public static Complex CArcCosh(Complex z)
            {
                return (Complex.CLog(z + Complex.CSqrt(z * z - 1.0)));
            }

            // Complex Inverse Hyperbolic Cosine of complex number z (ver 2)
            public static Complex CArcCosh2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;

                double ysqd = y * y;
                double rtpos = Math.Sqrt(Math.Pow(x + 1.0, 2.0) + ysqd);
                double rtneg = Math.Sqrt(Math.Pow(x - 1.0, 2.0) + ysqd);
                double alpha = 0.5 * (rtpos + rtneg);
                double beta = 0.5 * (rtpos - rtneg);

                double InvCoshZRe =
                    Math.Log(alpha + Math.Sqrt(alpha * alpha - 1));
                double InvCoshZIm = Math.Sign(y) * Math.Acos(beta);

                return (new Complex(InvCoshZRe, InvCoshZIm));
            }

            // Complex Inverse Hyperbolic Tangent of complex number z
            public static Complex CArcTanh(Complex z)
            {
                return (0.5 * Complex.CLog((1.0 + z) / (1.0 - z)));
            }

            // Complex Inverse Hyperbolic Tangent of complex number z (ver 2)
            public static Complex CArcTanh2(Complex z)
            {
                double x = z.Real;
                double y = z.Imag;
                double xsqd = x * x;
                double ysqd = y * y;

                double InvTanhZRe = 0.25 * Math.Log((ysqd +
                    Math.Pow(x + 1.0, 2.0)) / (ysqd + Math.Pow(x - 1.0, 2.0)));

                double InvTanhZIm = 0.5 * Math.Atan2(2.0 * y, 1.0 - xsqd - ysqd);

                return (new Complex(InvTanhZRe, InvTanhZIm));
            }

            // Complex Inverse Hyperbolic Cotangent of complex number z 
            public static Complex CArcCoth(Complex z)
            {
                return (Complex.CArcTanh(1.0 / z));
            }

            // Complex Inverse Hyperbolic Cotangent of complex number z (ver 2)
            public static Complex CArcCoth2(Complex z)
            {
                return (0.5 * Complex.CLog((z + 1.0) / (z - 1.0)));
            }

            // Complex Inverse Hyperbolic Secant of complex number z 
            public static Complex CArcSech(Complex z)
            {
                return (Complex.CArcCosh(1.0 / z));
            }

            // Complex Inverse Hyperbolic Secant of complex number z (ver 2)
            public static Complex CArcSech2(Complex z)
            {
                return (Complex.CLog((1.0 + Complex.CSqrt(1.0 - (z * z))) / z));
            }

            // Complex Inverse Hyperbolic Cosecant of complex number z 
            public static Complex CArcCsch(Complex z)
            {
                return (Complex.CArcSinh(1.0 / z));
            }

            // Complex Inverse Hyperbolic Cosecant of complex number z (ver 2)
            public static Complex CArcCsch2(Complex z)
            {
                return (Complex.CLog((1.0 + Complex.CSqrt(1.0 + (z * z))) / z));
            }
        }
        #endregion

        #region Complex Number vectors

        ///////////////////////////////////////////////
        /////////// Complex Number Vectors ////////////
        ///////////////////////////////////////////////
        public struct CVector : ICloneable
        {
            // Fields
            private int ndim;
            private Complex[] vector;

            // Constructors
            public CVector(int ndim)
            {
                this.ndim = ndim;
                this.vector = new Complex[ndim];
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] = Complex.CZero;
                }
            }

            public CVector(Complex[] cv)
            {
                this.ndim = cv.Length;
                this.vector = cv;
            }

            //Indexers
            public Complex this[int i]
            {
                get
                {
                    if (i < 0 || i > ndim)
                    {
                        throw new Exception("i is out of range!");
                    }
                    return vector[i];
                }
                set { vector[i] = value; }
            }

            // Accessors 
            public int GetCVectorSize
            {
                get { return ndim; }
            }

            // Override Methods
            public override string ToString()
            {
                string str = "(";
                for (int i = 0; i < ndim - 1; i++)
                {
                    str += vector[i] + ", ";
                }
                str += vector[ndim - 1] + ")";
                return str;
            }

            public override bool Equals(object obj)
            {
                return (obj is CVector) && this.Equals((CVector)obj);
            }

            public bool Equals(CVector cv)
            {
                return vector == cv.vector;
            }

            public override int GetHashCode()
            {
                return vector.GetHashCode();
            }

            public static bool operator ==(CVector v1, CVector v2)
            {
                return v1.Equals(v2);
            }

            public static bool operator !=(CVector v1, CVector v2)
            {
                return !v1.Equals(v2);
            }

            public static CVector operator +(CVector cv)
            {
                return cv;
            }

            public static CVector operator +(CVector v1, CVector v2)
            {
                CVector result = new CVector(v1.GetCVectorSize);
                for (int i = 0; i < v1.GetCVectorSize; i++)
                {
                    result[i] = v1[i] + v2[i];
                }
                return result;
            }

            public static CVector operator +(CVector cv, double d)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] + d;
                }
                return result;
            }

            public static CVector operator +(double d, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] + d;
                }
                return result;
            }

            public static CVector operator +(CVector cv, Complex cn)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] + cn;
                }
                return result;
            }

            public static CVector operator +(Complex cn, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] + cn;
                }
                return result;
            }

            public static CVector operator -(CVector cv)
            {
                Complex[] result = new Complex[cv.GetCVectorSize];
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = -cv[i];
                }
                return new CVector(result);
            }

            public static CVector operator -(CVector v1, CVector v2)
            {
                CVector result = new CVector(v1.GetCVectorSize);
                for (int i = 0; i < v1.GetCVectorSize; i++)
                {
                    result[i] = v1[i] - v2[i];
                }
                return result;
            }

            public static CVector operator -(CVector cv, double d)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] - d;
                }
                return result;
            }

            public static CVector operator -(double d, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = d - cv[i];
                }
                return result;
            }

            public static CVector operator -(CVector cv, Complex cn)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] - cn;
                }
                return result;
            }

            public static CVector operator -(Complex cn, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cn - cv[i];
                }
                return result;
            }

            public static CVector operator *(CVector cv, double d)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] * d;
                }
                return result;
            }

            public static CVector operator *(double d, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = d * cv[i];
                }
                return result;
            }

            public static CVector operator *(CVector cv, Complex cn)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] * cn;
                }
                return result;
            }

            public static CVector operator *(Complex cn, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cn * cv[i];
                }
                return result;
            }

            public static CVector Product(CVector v1, CVector v2)
            {
                CVector result = new CVector(v1.GetCVectorSize);
                for (int i = 0; i < v1.GetCVectorSize; i++)
                {
                    result[i] = v1[i] * v2[i];
                }
                return result;
            }

            public static CVector operator /(CVector cv, double d)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] / d;
                }
                return result;
            }

            public static CVector operator /(double d, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = d / cv[i];
                }
                return result;
            }

            public static CVector operator /(CVector cv, Complex cn)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cv[i] / cn;
                }
                return result;
            }

            public static CVector operator /(Complex cn, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                for (int i = 0; i < cv.GetCVectorSize; i++)
                {
                    result[i] = cn / cv[i];
                }
                return result;
            }

            // Makes a clone copy of a complex vector
            public CVector Clone()
            {
                CVector cv = new CVector(vector);
                cv.vector = (Complex[])vector.Clone();
                return cv;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            // Methods
            // Calculates the dot product of a complex vector
            public static Complex DotProduct(CVector v1, CVector v2)
            {
                Complex result = Complex.CZero;
                for (int i = 0; i < v1.GetCVectorSize; i++)
                {
                    result += Complex.CConj(v1[i]) * v2[i];
                }
                return result;
            }

            // Calculates the norm of a complex vector
            public double GetNorm()
            {
                Complex result = Complex.CZero;
                for (int i = 0; i < this.GetCVectorSize; i++)
                {
                    result += Complex.CConj(vector[i]) * vector[i];
                }
                return Math.Sqrt(result.Real * result.Real + result.Imag * result.Imag);
            }

            // Calculates the square of the norm of a complex vector
            public double GetNormSquare()
            {
                Complex result = Complex.CZero;
                for (int i = 0; i < this.GetCVectorSize; i++)
                {
                    result += Complex.CConj(vector[i]) * vector[i];
                }
                return (result.Real * result.Real + result.Imag * result.Imag);
            }

            // Normalizes a complex vector
            public void Normalize()
            {
                double norm = GetNorm();
                if (norm == 0)
                {
                    throw new Exception("Normalized a vector with norm of zero!");
                }
                for (int i = 0; i < this.GetCVectorSize; i++)
                {
                    vector[i] /= norm;
                }
            }

            // Calculates the unit vector of a complex vector
            public CVector GetUnitVector()
            {
                CVector result = new CVector(vector);
                result.Normalize();
                return result;
            }

            // Calculates the complex conjugate of a complex vector
            public CVector GetConjugate()
            {
                for (int i = 0; i < this.GetCVectorSize; i++)
                {
                    vector[i] = Complex.CConj(vector[i]);
                }
                return new CVector(vector);
            }

            // Swaps entries in a complex vector
            public CVector SwapCVectorEntries(int m, int n)
            {
                Complex temp = vector[m];
                vector[m] = vector[n];
                vector[n] = temp;
                return new CVector(vector);
            }

            // Calculates the cross product between two complex vectors 
            public static CVector CrossProduct(CVector v1, CVector v2)
            {
                if (v1.GetCVectorSize != 3)
                {
                    throw new Exception("Vector v1 must be 3 dimensional!");
                }
                CVector result = new CVector(3);
                result[0] = v1[1] * v2[2] - v1[2] * v2[1];
                result[1] = v1[2] * v2[0] - v1[0] * v2[2];
                result[2] = v1[0] * v2[1] - v1[1] * v2[0];
                return result;
            }
        }

        #endregion

        #region Complex Number matrices
        
        ///////////////////////////////////////////
        ///////// Complex Number Matrices /////////
        ///////////////////////////////////////////
        public struct CMatrix : ICloneable
        {
            // Fields
            private int nRows;
            private int nCols;
            private Complex[,] matrix;

            // Constructors
            public CMatrix(int nRows, int nCols)
            {
                this.nRows = nRows;
                this.nCols = nCols;
                this.matrix = new Complex[nRows, nCols];
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        matrix[i, j] = Complex.CZero;
                    }
                }
            }

            public CMatrix(Complex[,] matrix)
            {
                this.nRows = matrix.GetLength(0);
                this.nCols = matrix.GetLength(1);
                this.matrix = matrix;
            }

            public CMatrix IdentityMatrix()
            {
                CMatrix m = new CMatrix(nRows, nCols);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        if (i == j)
                        {
                            m[i, j] = new Complex(1, 0);
                        }
                    }
                }
                return m;
            }

            // Accessors
            public int GetnRows
            { get { return nRows; } }

            public int GetnCols
            { get { return nCols; } }

            // Indexers
            public Complex this[int m, int n]
            {
                get
                {
                    if (m < 0 || m > nRows)
                    {
                        throw new Exception("m-th row is out of range!");
                    }
                    if (n < 0 || n > nCols)
                    {
                        throw new Exception("n-th col is out of range!");
                    }
                    return matrix[m, n];
                }
                set { matrix[m, n] = value; }
            }

            // Override Methods
            public override string ToString()
            {
                string strMatrix = "(";
                for (int i = 0; i < nRows; i++)
                {
                    string str = "";
                    for (int j = 0; j < nCols - 1; j++)
                    {
                        str += matrix[i, j].ToString() + ", ";
                    }
                    str += matrix[i, nCols - 1].ToString();
                    if (i != nRows - 1 && i == 0)
                        strMatrix += str + "\n";
                    else if (i != nRows - 1 && i != 0)
                        strMatrix += " " + str + "\n";
                    else
                        strMatrix += " " + str + ")";
                }
                return strMatrix;
            }

            public override bool Equals(object obj)
            {
                return (obj is CMatrix) && this.Equals((CMatrix)obj);
            }

            public bool Equals(CMatrix cm)
            {
                return matrix == cm.matrix;
            }

            public override int GetHashCode()
            {
                return matrix.GetHashCode();
            }

            public static bool operator ==(CMatrix cm1, CMatrix cm2)
            {
                return cm1.Equals(cm2);
            }

            public static bool operator !=(CMatrix cm1, CMatrix cm2)
            {
                return !cm1.Equals(cm2);
            }

            public static CMatrix operator +(CMatrix cm)
            {
                return cm;
            }

            public static CMatrix operator +(CMatrix cm1, CMatrix cm2)
            {
                if (!CMatrix.CompareDimension(cm1, cm2))
                {
                    throw new Exception("The dimensions of 2 matrices must be the same!");
                }
                CMatrix result = new CMatrix(cm1.GetnRows, cm1.GetnCols);
                for (int i = 0; i < cm1.GetnRows; i++)
                {
                    for (int j = 0; j < cm1.GetnCols; j++)
                    {
                        result[i, j] = cm1[i, j] + cm2[i, j];
                    }
                }
                return result;
            }

            public static CMatrix operator +(CMatrix cm, Complex cn)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] + cn;
                    }
                }
                return result;
            }

            public static CMatrix operator +(Complex cn, CMatrix cm)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] + cn;
                    }
                }
                return result;
            }

            public static CMatrix operator -(CMatrix cm)
            {
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        cm[i, j] = -cm[i, j];
                    }
                }
                return cm;
            }

            public static CMatrix operator -(CMatrix cm1, CMatrix cm2)
            {
                if (!CMatrix.CompareDimension(cm1, cm2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                CMatrix result = new CMatrix(cm1.GetnRows, cm1.GetnCols);
                for (int i = 0; i < cm1.GetnRows; i++)
                {
                    for (int j = 0; j < cm1.GetnCols; j++)
                    {
                        result[i, j] = cm1[i, j] - cm2[i, j];
                    }
                }
                return result;
            }

            public static CMatrix operator -(CMatrix cm, Complex cn)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] - cn;
                    }
                }
                return result;
            }

            public static CMatrix operator -(Complex cn, CMatrix cm)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cn - cm[i, j];
                    }
                }
                return result;
            }

            public static CMatrix operator *(CMatrix cm, Complex cn)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] * cn;
                    }
                }
                return result;
            }

            public static CMatrix operator *(Complex cn, CMatrix cm)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] * cn;
                    }
                }
                return result;
            }

            public static CMatrix operator *(CMatrix cm1, CMatrix cm2)
            {
                if (cm1.GetnCols != cm2.GetnRows)
                {
                    throw new Exception("# columns of the matrix 1 must = # columns of the matrix 2");
                }

                Complex ctmp;
                CMatrix result = new CMatrix(cm1.GetnRows, cm2.GetnCols);

                for (int i = 0; i < cm1.GetnRows; i++)
                {
                    for (int j = 0; j < cm2.GetnCols; j++)
                    {
                        ctmp = result[i, j];
                        for (int k = 0; k < result.GetnRows; k++)
                        {
                            ctmp += cm1[i, k] * cm2[k, j];
                        }
                        result[i, j] = ctmp;
                    }
                }
                return result;
            }

            public static CMatrix operator /(CMatrix cm, Complex cn)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] / cn;
                    }
                }
                return result;
            }

            public static CMatrix operator /(Complex cn, CMatrix cm)
            {
                CMatrix result = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i, j] = cm[i, j] / cn;
                    }
                }
                return result;
            }

            // Methods
            // Checks for a square matrix where #rows = #cols
            public bool IsSquared()
            {
                if (nRows == nCols)
                    return true;
                else
                    return false;
            }

            // Compares the dimension of two complex matrices
            public static bool CompareDimension(CMatrix cm1, CMatrix cm2)
            {
                if (cm1.GetnRows == cm2.GetnRows && cm1.GetnCols == cm2.GetnCols)
                    return true;
                else
                    return false;
            }

            // Makes a clone copy of a complex matrix
            public CMatrix Clone()
            {
                CMatrix cm = new CMatrix(matrix);
                cm.matrix = (Complex[,])matrix.Clone();
                return cm;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            // Sets up a call to calculate the transpose of a complex matrix
            public CMatrix GetTranspose()
            {
                CMatrix ct = this;
                ct.Transpose();
                return ct;
            }

            // Calculates the transpose of a complex matrix
            public void Transpose()
            {
                CMatrix cm = new CMatrix(nCols, nRows);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        cm[j, i] = matrix[i, j];
                    }
                }
                this = cm;
            }

            // Calculates the trace of a complex matrix
            public Complex GetTrace()
            {
                Complex sum_of_diag = Complex.CZero;
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        if (i == j)
                            sum_of_diag += matrix[i, j];
                    }
                }
                return sum_of_diag;
            }

            // Extracts a row vector from a complex matrix at specified row
            public CVector GetRowCVector(int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                CVector RowCVector = new CVector(nCols);
                for (int i = 0; i < nCols; i++)
                {
                    RowCVector[i] = matrix[m, i];
                }
                return RowCVector;
            }

            // Extracts a column vector from a complex matrix at specified column
            public CVector GetColCVector(int m)
            {
                if (m < 0 || m > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                CVector ColCVector = new CVector(nRows);
                for (int i = 0; i < nRows; i++)
                {
                    ColCVector[i] = matrix[i, m];
                }
                return ColCVector;
            }

            // Swaps specificed complex matrix row with another row
            public CMatrix SwapCMatrixRow(int m, int n)
            {
                Complex ctemp = Complex.CZero;
                for (int i = 0; i < nCols; i++)
                {
                    ctemp = matrix[m, i];
                    matrix[m, i] = matrix[n, i];
                    matrix[n, i] = ctemp;
                }
                return new CMatrix(matrix);
            }

            // Swaps specificed complex matrix column with another column
            public CMatrix SwapCMatrixColumn(int m, int n)
            {
                Complex ctemp = Complex.CZero;
                for (int i = 0; i < nRows; i++)
                {
                    ctemp = matrix[i, m];
                    matrix[i, m] = matrix[i, n];
                    matrix[i, n] = ctemp;
                }
                return new CMatrix(matrix);
            }

            // Calculates the transform of a complex matrix
            public static CVector CTransform(CMatrix cm, CVector cv)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                if (!cm.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (cm.GetnCols != cv.GetCVectorSize)
                {
                    throw new Exception("Vector size must = # rows in matrix");
                }
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    result[i] = Complex.CZero;
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i] += cm[i, j] * cv[j];
                    }
                }
                return result;
            }

            public static CVector CTransform(CVector cv, CMatrix cm)
            {
                CVector result = new CVector(cv.GetCVectorSize);
                if (!cm.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (cm.GetnRows != cv.GetCVectorSize)
                {
                    throw new Exception("Vector size must = # rows in matrix");
                }
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    result[i] = Complex.CZero;
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        result[i] += cv[j] * cm[j, i];
                    }
                }
                return result;
            }

            public static CMatrix CTransform(CVector cv1, CVector cv2)
            {
                if (cv1.GetCVectorSize != cv2.GetCVectorSize)
                {
                    throw new Exception("The vectors must have the same size!");
                }
                CMatrix result = new CMatrix(cv1.GetCVectorSize,
                                             cv1.GetCVectorSize);
                for (int i = 0; i < cv1.GetCVectorSize; i++)
                {
                    for (int j = 0; j < cv1.GetCVectorSize; j++)
                    {
                        result[j, i] = cv1[i] * cv2[j];
                    }
                }
                return result;
            }

            // Calculates the determinant of a complex matrix
            public static Complex Determinant(CMatrix cm)
            {
                Complex result = new Complex(0.0, 0.0);
                if (!cm.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (cm.GetnRows == 1)
                    result = cm[0, 0];
                else
                {
                    for (int i = 0; i < cm.GetnRows; i++)
                    {
                        result += Math.Pow(-1, i) * cm[0, i] * Determinant(CMatrix.Minor(cm, 0, i));
                    }
                }
                return result;
            }

            // Calculates the minor of a complex matrix at specified row and column
            public static CMatrix Minor(CMatrix cm, int row, int col)
            {
                CMatrix cmm = new CMatrix(cm.GetnRows - 1, cm.GetnCols - 1);
                int ii = 0, jj = 0;
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    if (i == row)
                        continue;
                    jj = 0;
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        if (j == col)
                            continue;
                        cmm[ii, jj] = cm[i, j];
                        jj++;
                    }
                    ii++;
                }
                return cmm;
            }

            // Calculates the adjoint of a complex matrix
            public static CMatrix Adjoint(CMatrix cm)
            {
                if (!cm.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                CMatrix ma = new CMatrix(cm.GetnRows, cm.GetnCols);
                for (int i = 0; i < cm.GetnRows; i++)
                {
                    for (int j = 0; j < cm.GetnCols; j++)
                    {
                        ma[i, j] = Math.Pow(-1, i + j) * (Determinant(Minor(cm, i, j)));
                    }
                }
                return ma.GetTranspose();
            }

            // Calculates the inverse of a complex matrix
            public static CMatrix Inverse(CMatrix cm)
            {
                if (Determinant(cm) == new Complex(0, 0))
                {
                    throw new Exception("Cannot inverse a matrix with 0 determinant!");
                }
                return (Adjoint(cm) / Determinant(cm));
            }

            // Replaces the n-th row of a complex matrix with contents of a complex vector
            public CMatrix ReplaceCRow(CVector cv, int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                if (cv.GetCVectorSize != nCols)
                {
                    throw new Exception("Vector size is out of range!");
                }
                for (int i = 0; i < nCols; i++)
                {
                    matrix[m, i] = cv[i];
                }
                return new CMatrix(matrix);
            }

            // Replaces the n-th column of a complex matrix with contents of a complex vector
            public CMatrix ReplaceCCol(CVector cv, int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                if (cv.GetCVectorSize != nRows)
                {
                    throw new Exception("Vector size is out of range!");
                }
                for (int i = 0; i < nRows; i++)
                {
                    matrix[i, n] = cv[i];
                }
                return new CMatrix(matrix);
            }
        }
        #endregion
        
        //////////////////////////////////////////
        ////////////////// MAIN //////////////////
        //////////////////////////////////////////
        static void Main(string[] args)
        {
            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing Complex Number Operations");
            Console.WriteLine("---------------------------------");
            //Declare some complex variables and
            //assign them some arbitrary data.
            Complex w1 = new Complex(1.0, 2.0);
            Complex w2 = new Complex(3.0, 4.0);
            Complex w3 = new Complex();

            //Declare a variable i to hold the complex
            //number i. This will make later operations easier.
            Complex i = Complex.i;

            //Declare a real variable and assign
            //it some arbitrary data.
            double x1 = 5.0;

            //Display all these variables out
            Console.WriteLine("w1 = {0}", w1);
            Console.WriteLine("w2 = {0}", w2);
            Console.WriteLine("w3 = {0}", w3);
            Console.WriteLine(" i = {0}", i);
            Console.WriteLine("x1 = {0}", x1);

            Console.WriteLine("----------------------------------");
            Console.WriteLine("Testing Complex Number Addition   ");
            Console.WriteLine("----------------------------------");
            w3 = w1 + w2;
            Console.WriteLine("w3 = ({0}) + ({1}) = {2}", w1, w2, w3);
            w3 = Complex.CAdd(w1, w2);
            Console.WriteLine("w3 = CAdd({0},{1}) = {2}", w1, w2, w3);
            w3 = w1 + x1;
            Console.WriteLine("w3 = ({0}) + ({1}) = {2}", w1, x1, w3);
            w3 = Complex.CAdd(w1, x1);
            Console.WriteLine("w3 = CAdd({0},{1}) = {2}", w1, x1, w3);
            w3 = x1 + w1;
            Console.WriteLine("w3 = ({0}) + ({1}) = {2}", x1, w1, w3);
            w3 = Complex.CAdd(x1, w1);
            Console.WriteLine("w3 = CAdd({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("----------------------------------");
            Console.WriteLine("Testing Complex Number Subtraction");
            Console.WriteLine("----------------------------------");
            w3 = w1 - w2;
            Console.WriteLine("w3 = ({0}) - ({1}) = {2}", w1, w2, w3);
            w3 = Complex.CSub(w1, w2);
            Console.WriteLine("w3 = CSub({0},{1}) = {2}", w1, w2, w3);
            w3 = w1 - x1;
            Console.WriteLine("w3 = ({0}) - ({1}) = {2}", w1, x1, w3);
            w3 = Complex.CSub(w1, x1);
            Console.WriteLine("w3 = CSub({0},{1}) = {2}", w1, x1, w3);
            w3 = x1 - w1;
            Console.WriteLine("w3 = ({0}) - ({1}) = {2}", x1, w1, w3);
            w3 = Complex.CSub(x1, w1);
            Console.WriteLine("w3 = CSub({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------------------");
            Console.WriteLine("Testing Complex Number Multiplication");
            Console.WriteLine("-------------------------------------");
            w3 = w1 * w2;
            Console.WriteLine("w3 = ({0}) * ({1}) = {2}", w1, w2, w3);
            w3 = Complex.CMult(w1, w2);
            Console.WriteLine("w3 = CMult({0},{1}) = {2}", w1, w2, w3);
            w3 = w1 * x1;
            Console.WriteLine("w3 = ({0}) * ({1}) = {2}", w1, x1, w3);
            w3 = Complex.CMult(w1, x1);
            Console.WriteLine("w3 = CMult({0},{1}) = {2}", w1, x1, w3);
            w3 = x1 * w1;
            Console.WriteLine("w3 = ({0}) * ({1}) = {2}", x1, w1, w3);
            w3 = Complex.CMult(x1, w1);
            Console.WriteLine("w3 = CMult({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------------------");
            Console.WriteLine("Testing Complex Number Division      ");
            Console.WriteLine("-------------------------------------");
            w3 = w1 / w2;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", w1, w2, w3);
            w3 = Complex.CDiv(w1, w2);
            Console.WriteLine("w3 = CDiv({0},{1}) = {2}", w1, w2, w3);
            w3 = Complex.CDiv2(w1, w2);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", w1, w2, w3);
            w3 = w1 / x1;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", w1, x1, w3);
            w3 = Complex.CDiv(w1, x1);
            Console.WriteLine("w3 = CDiv({0},{1}) = {2}", w1, x1, w3);
            w3 = Complex.CDiv2(w1, (Complex)x1);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", w1, x1, w3);
            w3 = x1 / w1;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", x1, w1, w3);
            w3 = Complex.CDiv(x1, w1);
            Console.WriteLine("w3 = CDiv({0},{1}) = {2}", x1, w1, w3);
            w3 = Complex.CDiv2((Complex)x1, w1);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------------------");
            Console.WriteLine("Testing Complex Number Division by 0 ");
            Console.WriteLine("-------------------------------------");
            x1 = 0.0;
            w2 = 0.0 + i * 0.0;
            w3 = w1 / w2;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", w1, w2, w3);
            w3 = Complex.CDiv(w1, w2);
            Console.WriteLine("w3 = CDiv({0},{1}) = {2}", w1, w2, w3);
            w3 = Complex.CDiv2(w1, w2);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", w1, w2, w3);
            w3 = w1 / x1;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", w1, x1, w3);
            w3 = Complex.CDiv(w1, x1);
            Console.WriteLine("w3 =  CDiv({0},{1}) = {2}", w1, x1, w3);
            w3 = Complex.CDiv2(w1, (Complex)x1);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", w1, x1, w3);
            w3 = x1 / w1;
            Console.WriteLine("w3 = ({0}) / ({1}) = {2}", x1, w1, w3);
            w3 = Complex.CDiv(x1, w1);
            Console.WriteLine("w3 = CDiv({0},{1}) = {2}", x1, w1, w3);
            w3 = Complex.CDiv2((Complex)x1, w1);
            Console.WriteLine("w3 = CDiv2({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------");
            Console.WriteLine("Testing Polar-to-z and   ");
            Console.WriteLine("z-to-Polar Conversions   ");
            Console.WriteLine("-------------------------");
            w1 = 1.0 + i;
            Console.WriteLine("w1 = {0}", w1);
            w2 = Complex.FromXYToPolar(w1);
            Console.WriteLine("w2 = FromXYToPolar({0}) = {1}", w1, w2);
            w3 = Complex.FromPolarToXY(w2.Real, w2.Imag);
            Console.WriteLine("w3 = FromPolarToXY({0}) = {1}", w2, w3);

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing Complex Number Equality");
            Console.WriteLine("-------------------------------");
            if (w1.Equals(w1))
                Console.WriteLine("Yes, w1 = w1");
            else
                Console.WriteLine("No, w1 != w1");

            if (w1.Equals(w3))
                Console.WriteLine("Yes, w1 = w3");
            else
                Console.WriteLine("No, w1 != w3");

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing Complex-to-Double and  ");
            Console.WriteLine("Double-to-Complex Conversion   ");
            Console.WriteLine("-------------------------------");
            w1 = 1.0 + i * 2.0;
            w2 = 3.0 + i * 4.0;
            Console.WriteLine("w1 = {0}", w1);
            Console.WriteLine("w2 = {0}", w2);
            Console.WriteLine("---------------------------");
            //Implicit conversion of Complex-to-Doubl
            x1 = w2;
            Console.Write("Complex-to-Double Conversion: ");
            Console.WriteLine("x1 <-- w2 = {0}", x1);
            //Explicit conversion of double-to-Complex
            w3 = (Complex)x1;
            Console.Write("Double-to-Complex Conversion: ");
            Console.WriteLine("w3 <-- x1 = {0}", w3);

            Console.WriteLine("---------------------------");
            Console.WriteLine("Testing CConjugate Function");
            Console.WriteLine("---------------------------");
            w3 = Complex.CConj(w1);
            Console.WriteLine("w1  = {0}", w1);
            Console.WriteLine("w1* = {0}", w3);

            Console.WriteLine("----------------------");
            Console.WriteLine("Testing CNorm Function");
            Console.WriteLine("----------------------");
            x1 = Complex.CNorm(w1);
            Console.WriteLine(" w1  = {0}", w1);
            Console.WriteLine("|w1| = {0}", x1);

            Console.WriteLine("-----------------------");
            Console.WriteLine("Testing CNorm2 Function");
            Console.WriteLine("-----------------------");
            x1 = Complex.CNorm2(w1);
            Console.WriteLine(" w1  = {0}", w1);
            Console.WriteLine("|w1| = {0}", x1);

            Console.WriteLine("-----------------------");
            Console.WriteLine("Testing Inv Function   ");
            Console.WriteLine("-----------------------");
            w3 = Complex.Inv(w1);
            Console.WriteLine(" 1/({0}) = {1}", w1, w3);

            Console.WriteLine("-----------------------");
            Console.WriteLine("Testing Re Function   ");
            Console.WriteLine("-----------------------");
            x1 = Complex.Re(w1);
            Console.WriteLine(" Re({0}) = {1}", w1, x1);

            Console.WriteLine("-----------------------");
            Console.WriteLine("Testing Im Function   ");
            Console.WriteLine("-----------------------");
            x1 = Complex.Im(w1);
            Console.WriteLine(" Im({0}) = {1}", w1, x1);

            Console.WriteLine("------------------------");
            Console.WriteLine("Testing CExp(z) Function");
            Console.WriteLine("------------------------");
            w3 = Complex.CExp(w1);
            Console.WriteLine("CExp({0}) = {1}", w1, w3);

            Console.WriteLine("------------------------");
            Console.WriteLine("Testing CLog(z) Function");
            Console.WriteLine("------------------------");
            w1 = Complex.CLog(w3);
            Console.WriteLine("CLog({0}) = {1}", w3, w1);

            Console.WriteLine("-------------------------");
            Console.WriteLine("Testing CLog2(z) Function");
            Console.WriteLine("-------------------------");
            w1 = Complex.CLog2(w3);
            Console.WriteLine("CLog2({0}) = {1}", w3, w1);

            Console.WriteLine("-----------------------------");
            Console.WriteLine("Testing CLog10(z) Function   ");
            Console.WriteLine("-----------------------------");
            w2 = Complex.CLog10(w1);
            Console.WriteLine("CLog10({0}) = {1}", w1, w2);

            w2 = 3.0 + 4.0 * i;
            Console.WriteLine("-----------------------------");
            Console.WriteLine("Testing CLogb(z1,z2) Function");
            Console.WriteLine("-----------------------------");
            w3 = Complex.CLogb(w1, w2);
            Console.WriteLine("CLogb({0},{1}) = {2}", w1, w2, w3);

            Console.WriteLine("-----------------------------");
            Console.WriteLine("Testing CLogb(x,z2) Function ");
            Console.WriteLine("-----------------------------");
            w3 = Complex.CLogb((Complex)x1, w2);
            Console.WriteLine("CLogb({0},{1}) = {2}", (Complex)x1, w2, w3);

            Console.WriteLine("-----------------------------");
            Console.WriteLine("Testing CLogb(z1,x) Function ");
            Console.WriteLine("-----------------------------");
            w3 = Complex.CLogb(w1, x1);
            Console.WriteLine("CLogb({0},{1}) = {2}", w1, (Complex)x1, w3);

            Console.WriteLine("------------------------------");
            Console.WriteLine("Testing Power & Root Functions");
            Console.WriteLine("------------------------------");

            //Test data
            //Complex number to be raised to a positive real number
            w1 = 1.0 + 2.0 * i;

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower(z,x) Function     ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w3 = Complex.CPow(w1, 2);
            Console.WriteLine("CPow({0},2) = {1}", w1, w3);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower(z,1/x) Function   ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CPow(w3, 0.5);
            Console.WriteLine("CPow({0},0.5) = {1}", w3, w1);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CRoot(z,x) Function      ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CRoot(w3, 2);
            Console.WriteLine("CRoot({0},2) = {1}", w3, w1);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower2(z,x) Function    ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w3 = Complex.CPow2(w1, 2);
            Console.WriteLine("CPow2({0},2) = {1}", w1, w3);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower2(z,1/x) Function  ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CPow2(w3, 0.5);
            Console.WriteLine("CPow2({0},0.5) = {1}", w3, w1);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CRoot2(z,x) Function     ");
            Console.WriteLine("where x = a positive real number.");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CRoot(w3, (Complex)2);
            Console.WriteLine("CRoot2({0},2) = {1}", w3, w1);

            //Test data
            //Complex number raised to another complex number power
            w1 = 1.0 + 2.0 * i;
            w2 = 3.0 + 4.0 * i;

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower(z,w) Function     ");
            Console.WriteLine("where w = another complex number ");
            Console.WriteLine("---------------------------------");
            w3 = Complex.CPow(w1, w2);
            Console.WriteLine("CPow({0},{1}) = {2}", w1, w2, w3);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CPower(z,1/w) Function   ");
            Console.WriteLine("where w = another complex number ");
            Console.WriteLine("Note how taking the complex root ");
            Console.WriteLine("will not return the original     ");
            Console.WriteLine("complex number.                  ");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CPow(w3, 1.0 / w2);
            Console.WriteLine("CPow({0},1.0/({1})) = {2}", w3, w2, w1);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CRoot(z,w) Function      ");
            Console.WriteLine("where w = another complex number ");
            Console.WriteLine("Note how taking the complex root ");
            Console.WriteLine("will not return the original     ");
            Console.WriteLine("complex number.                  ");
            Console.WriteLine("---------------------------------");
            w1 = Complex.CRoot(w3, w2);
            Console.WriteLine("CRoot({0},{1}) = {2}", w3, w2, w1);

            //Test data
            //Positive real number raised to a complex number power
            w1 = 1.0 + 2.0 * i;
            x1 = 2.0;

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing CPower(x,z) Function   ");
            Console.WriteLine("where z = a complex number     ");
            Console.WriteLine("and x = a positive real number.");
            Console.WriteLine("-------------------------------");
            w3 = Complex.CPow(x1, w1);
            Console.WriteLine("CPow({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing CPower(x,1/z) Function ");
            Console.WriteLine("where z = a complex number     ");
            Console.WriteLine("and x = a positive real number.");
            Console.WriteLine("-------------------------------");
            w2 = Complex.CPow(w3, 1.0 / w1);
            Console.WriteLine("CPow({0},1.0/({1})) = {2}", w3, w1, w2);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CRoot(x,z) Function      ");
            Console.WriteLine("where z = another complex number ");
            Console.WriteLine("and x = a positive real number.  ");
            Console.WriteLine("---------------------------------");
            w2 = Complex.CRoot(w3, w1);
            Console.WriteLine("CRoot({0},{1}) = {2}", w3, w1, w2);

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing CPower2(x,z) Function  ");
            Console.WriteLine("where z = a complex number     ");
            Console.WriteLine("and x = a positive real number.");
            Console.WriteLine("-------------------------------");
            w3 = Complex.CPow2(x1, w1);
            Console.WriteLine("CPow2({0},{1}) = {2}", x1, w1, w3);

            Console.WriteLine("-------------------------------");
            Console.WriteLine("Testing CPower2(x,1/z) Function");
            Console.WriteLine("where z = a complex number     ");
            Console.WriteLine("and x = a positive real number.");
            Console.WriteLine("-------------------------------");
            w2 = Complex.CPow2(w3, 1.0 / w1);
            Console.WriteLine("CPow2({0},1.0/({1})) = {2}", w3, w1, w2);

            Console.WriteLine("---------------------------------");
            Console.WriteLine("Testing CRoot(x,z) Function      ");
            Console.WriteLine("where z = another complex number ");
            Console.WriteLine("and x = a positive real number.  ");
            Console.WriteLine("---------------------------------");
            w2 = Complex.CRoot(w3, w1);
            Console.WriteLine("CRoot2({0},{1}) = {2}", w3, w1, w2);

            w1 = 1.0 + 2.0 * i;

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing:                        ");
            Console.WriteLine("Complex Trigonometric Functions ");
            Console.WriteLine("where input z = a complex number");
            Console.WriteLine("--------------------------------");
            w2 = Complex.CSin(w1);
            Console.WriteLine("CSin({0}) = {1}", w1, w2);
            w2 = Complex.CSin2(w1);
            Console.WriteLine("CSin2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCos(w1);
            Console.WriteLine("CCos({0}) = {1}", w1, w2);
            w2 = Complex.CCos2(w1);
            Console.WriteLine("CCos2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CTan(w1);
            Console.WriteLine("CTan({0}) = {1}", w1, w2);
            w2 = Complex.CTan2(w1);
            Console.WriteLine("CTan2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCot(w1);
            Console.WriteLine("CCot({0}) = {1}", w1, w2);
            w2 = Complex.CCot2(w1);
            Console.WriteLine("CCot2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CSec(w1);
            Console.WriteLine("CSec({0}) = {1}", w1, w2);
            w2 = Complex.CSec2(w1);
            Console.WriteLine("CSec2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCsc(w1);
            Console.WriteLine("CCsc({0}) = {1}", w1, w2);
            w2 = Complex.CCsc2(w1);
            Console.WriteLine("CCsc2({0}) = {1}", w1, w2);

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing: Inverse                ");
            Console.WriteLine("Complex Trigonometric Functions ");
            Console.WriteLine("where input z = a complex number");
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcSin(w1);
            Console.WriteLine("CArcSin({0}) = {1}", w1, w2);
            w2 = Complex.CArcSin2(w1);
            Console.WriteLine("CArcSin2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCos(w1);
            Console.WriteLine("CArcCos({0}) = {1}", w1, w2);
            w2 = Complex.CArcCos2(w1);
            Console.WriteLine("CArcCos2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcTan(w1);
            Console.WriteLine("CArcTan({0}) = {1}", w1, w2);
            w2 = Complex.CArcTan2(w1);
            Console.WriteLine("CArcTan2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCot(w1);
            Console.WriteLine("CArcCot({0}) = {1}", w1, w2);
            w2 = Complex.CArcCot2(w1);
            Console.WriteLine("CArcCot2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcSec(w1);
            Console.WriteLine("CArcSec({0}) = {1}", w1, w2);
            w2 = Complex.CArcSec2(w1);
            Console.WriteLine("CArcSec2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCsc(w1);
            Console.WriteLine("CArcCsc({0}) = {1}", w1, w2);
            w2 = Complex.CArcCsc2(w1);
            Console.WriteLine("CArcCsc2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing:                        ");
            Console.WriteLine("Complex Hyperbolic Functions    ");
            Console.WriteLine("where input z = a complex number");
            Console.WriteLine("--------------------------------");
            w2 = Complex.CSinh(w1);
            Console.WriteLine("CSinh({0}) = {1}", w1, w2);
            w2 = Complex.CSinh2(w1);
            Console.WriteLine("CSinh2({0}) = {1}", w1, w2);
            w2 = Complex.CSinh3(w1);
            Console.WriteLine("CSinh3({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCosh(w1);
            Console.WriteLine("CCosh({0}) = {1}", w1, w2);
            w2 = Complex.CCosh2(w1);
            Console.WriteLine("CCosh2({0}) = {1}", w1, w2);
            w2 = Complex.CCosh3(w1);
            Console.WriteLine("CCosh3({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CTanh(w1);
            Console.WriteLine("CTanh({0}) = {1}", w1, w2);
            w2 = Complex.CTanh2(w1);
            Console.WriteLine("CTanh2({0}) = {1}", w1, w2);
            w2 = Complex.CTanh3(w1);
            Console.WriteLine("CTanh3({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCoth(w1);
            Console.WriteLine("CCoth({0}) = {1}", w1, w2);
            w2 = Complex.CCoth2(w1);
            Console.WriteLine("CCoth2({0}) = {1}", w1, w2);
            w2 = Complex.CCoth3(w1);
            Console.WriteLine("CCoth3({0}) = {1}", w1, w2);
            w2 = Complex.CCoth4(w1);
            Console.WriteLine("CCoth4({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CSech(w1);
            Console.WriteLine("CSech({0}) = {1}", w1, w2);
            w2 = Complex.CSech2(w1);
            Console.WriteLine("CSech2({0}) = {1}", w1, w2);
            w2 = Complex.CSech3(w1);
            Console.WriteLine("CSech3({0}) = {1}", w1, w2);
            w2 = Complex.CSech4(w1);
            Console.WriteLine("CSech4({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CCsch(w1);
            Console.WriteLine("CCsch({0}) = {1}", w1, w2);
            w2 = Complex.CCsch2(w1);
            Console.WriteLine("CCsch2({0}) = {1}", w1, w2);
            w2 = Complex.CCsch3(w1);
            Console.WriteLine("CCsch3({0}) = {1}", w1, w2);
            w2 = Complex.CCsch4(w1);
            Console.WriteLine("CCsch4({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing: Inverse                ");
            Console.WriteLine("Complex Hyperbolic Functions    ");
            Console.WriteLine("where input z = a complex number");
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcSinh(w1);
            Console.WriteLine("CArcSinh({0}) = {1}", w1, w2);
            w2 = Complex.CArcSinh2(w1);
            Console.WriteLine("CArcSinh2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCosh(w1);
            Console.WriteLine("CArcCosh({0}) = {1}", w1, w2);
            w2 = Complex.CArcCosh2(w1);
            Console.WriteLine("CArcCosh2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcTanh(w1);
            Console.WriteLine("CArcTanh({0}) = {1}", w1, w2);
            w2 = Complex.CArcTanh2(w1);
            Console.WriteLine("CArcTanh2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCoth(w1);
            Console.WriteLine("CArcCoth({0}) = {1}", w1, w2);
            w2 = Complex.CArcCoth2(w1);
            Console.WriteLine("CArcCoth2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcSech(w1);
            Console.WriteLine("CArcSech({0}) = {1}", w1, w2);
            w2 = Complex.CArcSech2(w1);
            Console.WriteLine("CArcSech2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");
            w2 = Complex.CArcCsch(w1);
            Console.WriteLine("CArcCsch({0}) = {1}", w1, w2);
            w2 = Complex.CArcCsch2(w1);
            Console.WriteLine("CArcCsch2({0}) = {1}", w1, w2);
            Console.WriteLine("--------------------------------");

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing: Complex Vectors        ");
            Console.WriteLine("--------------------------------");
            CVector Cv1 = new CVector(new Complex[] { new Complex(1, 2), new Complex(2, 3), new Complex(3, 4), new Complex(4, 5) });
            CVector Cv2 = new CVector(new Complex[] { new Complex(2, 1), new Complex(3, 2), new Complex(4, 3), new Complex(5, 4) });
            Complex c = new Complex(5, 10);

            Console.WriteLine("\n Cv1 = {0}", Cv1);
            Console.WriteLine(" Cv2 = {0}", Cv2);
            Console.WriteLine(" c = {0}", c);
            Console.WriteLine(" Cv2 + Cv1 = {0}", (Cv2 + Cv1));
            Console.WriteLine(" Cv2 - Cv1 = {0}", (Cv2 - Cv1));
            Console.WriteLine(" Cv2 * c = {0}", (Cv2 * c));
            Console.WriteLine(" Cv2 / c = {0}", (Cv2 / c));
            Console.WriteLine(" Product of Cv1 and Cv2 = {0}", CVector.Product(Cv1, Cv2));
            Console.WriteLine(" Dot product of Cv1 and Cv2 = {0}", CVector.DotProduct(Cv1, Cv2));

            Console.WriteLine("--------------------------------");
            Console.WriteLine("Testing: Complex Matrices       ");
            Console.WriteLine("--------------------------------");
            // Create a complex matrix using a 2D complex array:
            CMatrix Cm1 = new CMatrix(new Complex[,]{{new Complex(1,1),new Complex(1,2), new Complex(1,3)},
                                                     {new Complex(2,1),new Complex(2,2), new Complex(2,3)},
                                                     {new Complex(3,1),new Complex(3,2), new Complex(3,3)}});

            //Create a complex matrix by directly defining its elements:
            CMatrix Cm2 = new CMatrix(3, 3);
            Cm2[0, 0] = new Complex(9, 8);
            Cm2[0, 1] = new Complex(3, -3);
            Cm2[0, 2] = new Complex(7, 6);
            Cm2[1, 0] = new Complex(4, -2);
            Cm2[1, 1] = new Complex(0, 1);
            Cm2[1, 2] = new Complex(6, 3);
            Cm2[2, 0] = new Complex(8, 9);
            Cm2[2, 1] = new Complex(7, -1);
            Cm2[2, 2] = new Complex(7, 0);
            CVector Cv = new CVector(new Complex[] { new Complex(1, 1), new Complex(1, 2), new Complex(1, 3) });
            Console.WriteLine("\n Original matrix: Cm1 = \n{0}", Cm1);
            Console.WriteLine("\n Original matrix: Cm2 = \n{0}", Cm2);
            Console.WriteLine("\n Cv = {0}", (Cv));
            Console.WriteLine("\n Cm1 + Cm2 = \n{0}", (Cm1 + Cm2));
            Console.WriteLine("\n Cm1 - Cm2 = \n{0}", (Cm1 - Cm2));
            Console.WriteLine("\n Cm1 * Cm2 = \n{0}", (Cm1 * Cm2));
            Console.WriteLine("\n Cm2 * Cm1 = \n{0}", (Cm2 * Cm1));
            Console.WriteLine("\n Cm1 * Cv = {0}", CMatrix.CTransform(Cm1, Cv));
            Console.WriteLine("\n Cv * Cm1 = {0}", CMatrix.CTransform(Cv, Cm2));
            Console.WriteLine("\n Inverse of Cm2 = \n{0}", CMatrix.Inverse(Cm2));
            Console.WriteLine("\n Comfirm: Cm2 * Inv(Cm2) = \n{0}", Cm2 * CMatrix.Inverse(Cm2));
            Console.WriteLine("\n Determinant of Cm1 = {0}", CMatrix.Determinant(Cm1));
            Console.WriteLine("\n Determinant of Cm2 = {0}", CMatrix.Determinant(Cm2));
            Console.WriteLine("Press ENTER key to terminate program");
            Console.ReadLine();
        }
    }
}
