﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter07_CodeExamples
{
    class Program
    {
        #region Real Vector Methods from chapter 3
        
        public struct RVector : ICloneable
        {
            private int ndim;
            private double[] vector;

            public RVector(int ndim)
            {
                this.ndim = ndim;
                this.vector = new double[ndim];
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] = 0.0;
                }
            }

            public RVector(double[] vector)
            {
                this.ndim = vector.Length;
                this.vector = vector;
            }

            public double this[int i]
            {
                get
                {
                    if (i < 0 || i > ndim)
                    {
                        throw new Exception("Requested vector index is out of range!");
                    }
                    return vector[i];
                }
                set { vector[i] = value; }
            }

            public int GetVectorSize
            {
                get { return ndim; }
            }

            public RVector Clone()
            {
                RVector v = new RVector(vector);
                v.vector = (double[])vector.Clone();
                return v;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public RVector SwapVectorEntries(int m, int n)
            {
                double temp = vector[m];
                vector[m] = vector[n];
                vector[n] = temp;
                return new RVector(vector);
            }

            public override string ToString()
            {
                string str = "(";
                for (int i = 0; i < ndim - 1; i++)
                {
                    str += vector[i].ToString() + ", ";
                }
                str += vector[ndim - 1].ToString() + ")";
                return str;
            }

            public override bool Equals(object obj)
            {
                return (obj is RVector) && this.Equals((RVector)obj);
            }

            public bool Equals(RVector v)
            {
                return vector == v.vector;
            }

            public override int GetHashCode()
            {
                return vector.GetHashCode();
            }

            public static bool operator ==(RVector v1, RVector v2)
            {
                return v1.Equals(v2);
            }

            public static bool operator !=(RVector v1, RVector v2)
            {
                return !v1.Equals(v2);
            }

            public static RVector operator +(RVector v)
            {
                return v;
            }

            public static RVector operator +(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] + v2[i];
                }
                return result;
            }

            public static RVector operator -(RVector v)
            {
                double[] result = new double[v.ndim];
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = -v[i];
                }
                return new RVector(result);
            }

            public static RVector operator -(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] - v2[i];
                }
                return result;
            }

            public static RVector operator *(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] * d;
                }
                return result;
            }

            public static RVector operator *(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = d * v[i];
                }
                return result;
            }

            public static RVector operator /(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static RVector operator /(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static double DotProduct(RVector v1, RVector v2)
            {
                double result = 0.0;
                for (int i = 0; i < v1.ndim; i++)
                {
                    result += v1[i] * v2[i];
                }
                return result;
            }

            public double GetNorm()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return Math.Sqrt(result);
            }

            public double GetNormSquare()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return result;
            }

            public void Normalize()
            {
                double norm = GetNorm();
                if (norm == 0)
                {
                    throw new Exception("Tried to normalize a vector with norm of zero!");
                }
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] /= norm;
                }
            }

            public RVector GetUnitVector()
            {
                RVector result = new RVector(vector);
                result.Normalize();
                return result;
            }

            public static RVector CrossProduct(RVector v1, RVector v2)
            {
                if (v1.ndim != 3)
                {
                    throw new Exception("Vector v1 must be 3 dimensional!");
                }
                if (v2.ndim != 3)
                {
                    throw new Exception("Vector v2 must be 3 dimensional!");
                }
                RVector result = new RVector(3);
                result[0] = v1[1] * v2[2] - v1[2] * v2[1];
                result[1] = v1[2] * v2[0] - v1[0] * v2[2];
                result[2] = v1[0] * v2[1] - v1[1] * v2[0];
                return result;
            }
        }
        #endregion

        #region Real Matrix Methods from chapter 3

        public struct RMatrix : ICloneable
        {
            private int nRows;
            private int nCols;
            private double[,] matrix;

            public RMatrix(int nRows, int nCols)
            {
                this.nRows = nRows;
                this.nCols = nCols;
                this.matrix = new double[nRows, nCols];
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        matrix[i, j] = 0.0;
                    }
                }
            }

            public RMatrix(double[,] matrix)
            {
                this.nRows = matrix.GetLength(0);
                this.nCols = matrix.GetLength(1);
                this.matrix = matrix;
            }

            public RMatrix(RMatrix m)
            {
                nRows = m.GetnRows;
                nCols = m.GetnCols;
                matrix = m.matrix;
            }

            public RMatrix IdentityMatrix()
            {
                RMatrix m = new RMatrix(nRows, nCols);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        if (i == j)
                        {
                            m[i, j] = 1;
                        }
                    }
                }
                return m;
            }

            public double this[int m, int n]
            {
                get
                {
                    if (m < 0 || m > nRows)
                    {
                        throw new Exception("m-th row is out of range!");
                    }
                    if (n < 0 || n > nCols)
                    {
                        throw new Exception("n-th col is out of range!");
                    }
                    return matrix[m, n];
                }
                set { matrix[m, n] = value; }
            }

            public int GetnRows
            {
                get { return nRows; }
            }

            public int GetnCols
            {
                get { return nCols; }
            }

            public RMatrix Clone()
            {
                RMatrix m = new RMatrix(matrix);
                m.matrix = (double[,])matrix.Clone();
                return m;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public override string ToString()
            {
                string strMatrix = "(";
                for (int i = 0; i < nRows; i++)
                {
                    string str = "";
                    for (int j = 0; j < nCols - 1; j++)
                    {
                        str += matrix[i, j].ToString() + ", ";
                    }
                    str += matrix[i, nCols - 1].ToString();
                    if (i != nRows - 1 && i == 0)
                        strMatrix += str + "\n";
                    else if (i != nRows - 1 && i != 0)
                        strMatrix += " " + str + "\n";
                    else
                        strMatrix += " " + str + ")";
                }
                return strMatrix;
            }

            public override bool Equals(object obj)
            {
                return (obj is RMatrix) && this.Equals((RMatrix)obj);
            }

            public bool Equals(RMatrix m)
            {
                return matrix == m.matrix;
            }

            public override int GetHashCode()
            {
                return matrix.GetHashCode();
            }

            public static bool operator ==(RMatrix m1, RMatrix m2)
            {
                return m1.Equals(m2);
            }

            public static bool operator !=(RMatrix m1, RMatrix m2)
            {
                return !m1.Equals(m2);
            }

            public static RMatrix operator +(RMatrix m)
            {
                return m;
            }

            public static RMatrix operator +(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] + m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator -(RMatrix m)
            {
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        m[i, j] = -m[i, j];
                    }
                }
                return m;
            }

            public static RMatrix operator -(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] - m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnCols != m2.GetnRows)
                {
                    throw new Exception("The numbers of columns of the" +
                     " first matrix must be equal to the number of " +
                     " rows of the second matrix!");
                }
                double tmp;
                RMatrix result = new RMatrix(m1.GetnRows, m2.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m2.GetnCols; j++)
                    {
                        tmp = result[i, j];
                        for (int k = 0; k < result.GetnRows; k++)
                        {
                            tmp += m1[i, k] * m2[k, j];
                        }
                        result[i, j] = tmp;
                    }
                }
                return result;
            }

            public RMatrix GetTranspose()
            {
                RMatrix m = this;
                m.Transpose();
                return m;
            }

            public void Transpose()
            {
                RMatrix m = new RMatrix(nCols, nRows);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        m[j, i] = matrix[i, j];
                    }
                }
                this = m;
            }

            public double GetTrace()
            {
                double sum_of_diag = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    if (i < nCols)
                        sum_of_diag += matrix[i, i];
                }
                return sum_of_diag;
            }

            public bool IsSquared()
            {
                if (nRows == nCols)
                    return true;
                else
                    return false;
            }

            public static bool CompareDimension(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnRows == m2.GetnRows && m1.GetnCols == m2.GetnCols)
                    return true;
                else
                    return false;
            }

            public RVector GetRowVector(int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                RVector RowVector = new RVector(nCols);
                for (int i = 0; i < nCols; i++)
                {
                    RowVector[i] = matrix[m, i];
                }
                return RowVector;
            }

            public RVector GetColVector(int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                RVector ColVector = new RVector(nRows);
                for (int i = 0; i < nRows; i++)
                {
                    ColVector[i] = matrix[i, n];
                }
                return ColVector;
            }

            public RMatrix ReplaceRow(RVector vec, int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                if (vec.GetVectorSize != nCols)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nCols; i++)
                {
                    matrix[m, i] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix ReplaceCol(RVector vec, int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                if (vec.GetVectorSize != nRows)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nRows; i++)
                {
                    matrix[i, n] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapMatrixRow(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nCols; i++)
                {
                    temp = matrix[m, i];
                    matrix[m, i] = matrix[n, i];
                    matrix[n, i] = temp;
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapMatrixColumn(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    temp = matrix[i, m];
                    matrix[i, m] = matrix[i, n];
                    matrix[i, n] = temp;
                }
                return new RMatrix(matrix);
            }

            public static RVector Transform(RMatrix mat, RVector vec)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnCols != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of cols of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += mat[i, j] * vec[j];
                    }
                }
                return result;
            }

            public static RVector Transform(RVector vec, RMatrix mat)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of rows of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += vec[j] * mat[j, i];
                    }
                }
                return result;
            }

            public static RMatrix Transform(RVector v1, RVector v2)
            {
                if (v1.GetVectorSize != v2.GetVectorSize)
                {
                    throw new Exception("The vectors must have the same ndim!");
                }
                RMatrix result = new RMatrix(v1.GetVectorSize, v1.GetVectorSize);
                for (int i = 0; i < v1.GetVectorSize; i++)
                {
                    for (int j = 0; j < v1.GetVectorSize; j++)
                    {
                        result[j, i] = v1[i] * v2[j];
                    }
                }
                return result;
            }

            public static double Determinant(RMatrix mat)
            {
                double result = 0.0;
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows == 1)
                    result = mat[0, 0];
                else
                {
                    for (int i = 0; i < mat.GetnRows; i++)
                    {
                        result += Math.Pow(-1, i) * mat[0, i] * Determinant(RMatrix.Minor(mat, 0, i));
                    }
                }
                return result;
            }

            public static RMatrix Minor(RMatrix mat, int row, int col)
            {
                RMatrix mm = new RMatrix(mat.GetnRows - 1, mat.GetnCols - 1);
                int ii = 0, jj = 0;
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    if (i == row)
                        continue;
                    jj = 0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        if (j == col)
                            continue;
                        mm[ii, jj] = mat[i, j];
                        jj++;
                    }
                    ii++;
                }
                return mm;
            }

            public static RMatrix Adjoint(RMatrix mat)
            {
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                RMatrix ma = new RMatrix(mat.GetnRows, mat.GetnCols);
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        ma[i, j] = Math.Pow(-1, i + j) * (Determinant(Minor(mat, i, j)));
                    }
                }
                return ma.GetTranspose();
            }

            public static RMatrix Inverse(RMatrix mat)
            {
                if (Determinant(mat) == 0)
                {
                    throw new Exception("Cannot inverse a matrix with a zero determinant!");
                }
                return (Adjoint(mat) / Determinant(mat));
            }
        }
        #endregion

        #region Interpolation Methods

        public static double LinearInterpolation(double[] x, double[] y, double xval)
        {
            double yval = 0.0;
            for (int i = 0; i < x.Length - 1; i++)
            {
                if (xval >= x[i] && xval < x[i + 1])
                {
                    yval = y[i] + (xval - x[i]) * (y[i + 1] - y[i]) / (x[i + 1] - x[i]);
                }
            }
            return yval;
        }

        public static double[] LinearInterpolation(double[] x, double[] y, double[] xvals)
        {
            double[] yvals = new double[xvals.Length];
            for (int i = 0; i < xvals.Length; i++)
                yvals[i] = LinearInterpolation(x, y, xvals[i]);
            return yvals;
        }

        public static double BilinearInterpolation(double[] x, double[] y, double[,] z, double xval, double yval)
        {
            double zval = 0.0;
            for (int i = 0; i < x.Length - 1; i++)
            {
                for (int j = 0; j < y.Length - 1; j++)
                {
                    if (xval >= x[i] && xval < x[i + 1] && yval >= y[j] && yval < y[j + 1])
                    {
                        zval = z[i, j] * (x[i + 1] - xval) * (y[j + 1] - yval) / (x[i + 1] - x[i]) / (y[j + 1] - y[j]) +
                            z[i + 1, j] * (xval - x[i]) * (y[j + 1] - yval) / (x[i + 1] - x[i]) / (y[j + 1] - y[j]) +
                            z[i, j + 1] * (x[i + 1] - xval) * (yval - y[j]) / (x[i + 1] - x[i]) / (y[j + 1] - y[j]) +
                            z[i + 1, j + 1] * (xval - x[i]) * (yval - y[j]) / (x[i + 1] - x[i]) / (y[j + 1] - y[j]);
                    }
                }
            }
            return zval;
        }

        public static double[] BilinearInterpolation(double[] x, double[] y, double[,] z, double[] xvals, double[] yvals)
        {
            double[] zvals = new double[xvals.Length];
            for (int i = 0; i < xvals.Length; i++)
                zvals[i] = BilinearInterpolation(x, y, z, xvals[i], yvals[i]);
            return zvals;
        }

        public static double LagrangeInterpolation(double[] x, double[] y, double xval)
        {
            double yval = 0.0;
            double Products = y[0];
            for (int i = 0; i < x.Length; i++)
            {
                Products = y[i];
                for (int j = 0; j < x.Length; j++)
                {
                    if (i != j)
                    {
                        Products *= (xval - x[j]) / (x[i] - x[j]);
                    }
                }
                yval += Products;
            }
            return yval;
        }

        public static double[] LagrangeInterpolation(double[] x, double[] y, double[] xvals)
        {
            double[] yvals = new double[xvals.Length];
            for (int i = 0; i < xvals.Length; i++)
                yvals[i] = LagrangeInterpolation(x, y, xvals[i]);
            return yvals;
        }

        public static double BarycentricInterpolation(double[] x, double[] y, double xval)
        {
            double product;
            double deltaX;
            double bc1 = 0;
            double bc2 = 0;
            int size = x.Length;
            double[] weights = new double[size];

            for (int i = 0; i < size; i++)
            {
                product = 1;
                for (int j = 0; j < size; j++)
                {
                    if (i != j)
                    {
                        product *= (x[i] - x[j]);
                        weights[i] = 1.0 / product;
                    }
                }
            }

            for (int i = 0; i < size; i++)
            {
                deltaX = weights[i] / (xval - x[i]);
                bc1 += y[i] * deltaX;
                bc2 += deltaX;
            }
            return bc1 / bc2;
        }

        public static double[] BarycentricInterpolation(double[] x, double[] y, double[] xvals)
        {
            double[] yvals = new double[xvals.Length];
            for (int i = 0; i < xvals.Length; i++)
                yvals[i] = BarycentricInterpolation(x, y, xvals[i]);
            return yvals;
        }

        public static double NewtonDividedDifferenceInterpolation(double[] x, double[] y, double xval)
        {
            double yval;
            int size = x.Length;
            double[] tarray = new double[size];
            for (int i = 0; i < size; i++)
            {
                tarray[i] = y[i];
            }

            for (int i = 0; i < size - 1; i++)
            {
                for (int j = size - 1; j > i; j--)
                {
                    tarray[j] = (tarray[j - 1] - tarray[j]) / (x[j - 1 - i] - x[j]);
                }
            }

            yval = tarray[size - 1];

            for (int i = size - 2; i >= 0; i--)
            {
                yval = tarray[i] + (xval - x[i]) * yval;
            }
            return yval;
        }

        public static double[] NewtonDividedDifferenceInterpolation(double[] x, double[] y, double[] xvals)
        {
            double[] yvals = new double[xvals.Length];
            for (int i = 0; i < xvals.Length; i++)
                yvals[i] = NewtonDividedDifferenceInterpolation(x, y, xvals[i]);
            return yvals;
        }
        #endregion       

        #region Test Functions

        static void TestLinear()
        {
            double[] xdata = new double[] { 1, 3, 5, 7, 9 };
            double[] ydata = new double[] { 2, 6, 10, 14, 18 };
            double[] x = new double[] { 2, 4, 6, 8 };
            double[] y = LinearInterpolation(xdata, ydata, x);
            RVector xvec = new RVector(x);
            RVector yvec = new RVector(y);
            Console.Clear();
            Console.WriteLine("Running Linear Interpolation Test\n\n");
            Console.WriteLine(" x = " + xvec.ToString());
            Console.WriteLine(" y = " + yvec.ToString() + "\n\n");
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        static void TestBilinear()
        {
            double[] xdata = new double[] { 0, 2 };
            double[] ydata = new double[] { 0, 2 };
            double[,] zdata = new double[,] { { 0, 20 }, { 20, 10 } };
            double[] x = new double[18];
            double[] y = new double[18];
            RMatrix z = new RMatrix(18, 18);
            for (int i = 0; i < 9; i++)
            {
                x[i] = (i + 2.0) / 20.0;
                y[i] = (i + 2.0) / 20.0;
            }
            RVector xvec = new RVector(x);
            RVector yvec = new RVector(y);

            for (int i = 0; i < 18; i++)
            {
                for (int j = 0; j < 18; j++)
                {
                    z[i, j] = BilinearInterpolation(xdata, ydata, zdata, x[i], y[j]);
                }
            }
            Console.Clear();
            Console.WriteLine("Running Bilinear Interpolation Test\n\n");
            Console.WriteLine("x = " + xvec.ToString());
            Console.WriteLine("y = " + yvec.ToString());
            Console.WriteLine("\nResults z = \n" + z.ToString() + "\n\n");
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        static void TestLagrangian()
        {
            double[] xdata = new double[5] { 2, 4, 6, 8, 10 };
            double[] ydata = new double[5] { 2, 8, 18, 32, 50 };
            double[] x = new double[3] { 5.0, 7.0, 3.0 };

            double[] y = LagrangeInterpolation(xdata, ydata, x);
            RVector xvec = new RVector(x);
            RVector yvec = new RVector(y);
            Console.Clear();
            Console.WriteLine("Running Lagrangian Interpolation Test\n\n");
            Console.WriteLine(" x = " + xvec.ToString());
            Console.WriteLine(" y = " + yvec.ToString() + "\n\n");
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        static void TestBarycentric()
        {
            double[] xdata = new double[] { 0, 4, 8, 12, 16 };
            double[] ydata = new double[] { 0, 8, 32, 72, 128 };
            double[] x = new double[] { 2, 6, 10, 14 };
            double[] y = BarycentricInterpolation(xdata, ydata, x);
            RVector xvec = new RVector(x);
            RVector yvec = new RVector(y);
            Console.Clear();
            Console.WriteLine("Running Barycentric Interpolation Test\n\n");
            Console.WriteLine(" x = " + xvec.ToString());
            Console.WriteLine(" y = " + yvec.ToString() + "\n\n");
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        static void TestNewtonDividedDifference()
        {
            double[] xdata = new double[] { 50, 60, 70, 80, 90 };
            double[] ydata = new double[] { 75.0, 150.0, 200.0, 225.0, 250.0 };
            double[] x = new double[] { 55, 65, 75, 85 };
            double[] y = NewtonDividedDifferenceInterpolation(xdata, ydata, x);
            RVector xvec = new RVector(x);
            RVector yvec = new RVector(y);
            Console.Clear();
            Console.WriteLine("Running Newton Divided Difference Interpolation Test\n\n");
            Console.WriteLine(" x = " + xvec.ToString());
            Console.WriteLine(" y = " + yvec.ToString() + "\n\n");
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        public static void NaturalCubicSplineTest()
        {
            int size = 26;
            double[] xdata = new double[size];
            double[] ydata = new double[size];

            xdata[0] = 0.1;
            xdata[1] = 0.2;
            xdata[2] = 0.3;
            xdata[3] = 0.4;

            ydata[0] = -0.62049958;
            ydata[1] = -0.28398668;
            ydata[2] = 0.00660095;
            ydata[3] = 0.24842440;

            Console.Clear();
            NaturalCubicSpline(xdata, ydata, 0.25);
            Console.WriteLine("Press ENTER key to continue...");
            Console.ReadLine();
        }

        public static void ClampedCubicSplineTest()
        {
            int size = 26;
            double[] xdata = new double[size];
            double[] ydata = new double[size];

            xdata[0] = 0;
            xdata[1] = 1;
            xdata[2] = 2;
            xdata[3] = 3;

            ydata[0] = 1;
            ydata[1] = 2.718281828;
            ydata[2] = 7.389056099;
            ydata[3] = 20.08553692;

            double fp_0 = 1.0;
            double fp_n = 20.0855369;

            Console.Clear();
            ClampedCubicSpline(xdata, ydata, fp_0, fp_n, 2.0);
            Console.WriteLine("Press ENTER key to end program...");
            Console.ReadLine();
        }

        #endregion

        #region Natural Cubic Splines

        public static void NaturalCubicSpline(double[] x, double[] y, double xvalue)
        {
            int i, j, m;
            double S = 0.0;
            double delta = 0.0;

            int n = 4; //max number of coefficients: A,B,C,D for cubic spline

            double[] A = new double[n + 1];
            double[] B = new double[n + 1];
            double[] C = new double[n + 1];
            double[] D = new double[n + 1];
            double[] H = new double[n + 1];
            double[] XA = new double[n + 1];
            double[] XL = new double[n + 1];
            double[] XU = new double[n + 1];
            double[] XZ = new double[n + 1];

            for (i = 0; i < n; i++) A[i] = y[i];

            m = n - 1;
            for (i = 0; i <= m; i++)
                H[i] = x[i + 1] - x[i];

            for (i = 1; i <= m; i++)
                XA[i] = 3 * (A[i + 1] * H[i - 1] - A[i] * (x[i + 1] - x[i - 1]) + A[i - 1] * H[i]) / (H[i] * H[i - 1]);

            XL[0] = 1; XU[0] = 0; XZ[0] = 0;
            for (i = 1; i <= m; i++)
            {
                XL[i] = 2 * (x[i + 1] - x[i - 1]) - H[i - 1] * XU[i - 1];
                XU[i] = H[i] / XL[i];
                XZ[i] = (XA[i] - H[i - 1] * XZ[i - 1]) / XL[i];
            }
            XL[n] = 1; XZ[n] = 0; C[n] = 0;

            for (i = 0; i <= m; i++)
            {
                j = m - i;
                C[j] = XZ[j] - XU[j] * C[j + 1];
                B[j] = (A[j + 1] - A[j]) / H[j] - H[j] * (C[j + 1] + 2 * C[j]) / 3;
                D[j] = (C[j + 1] - C[j]) / (3 * H[j]);
            }

            Console.WriteLine("\n\nTesting natural cubic splines");
            Console.WriteLine("Input data stored in (x[i],y[i]). Value to interpolate = {0}\n", xvalue);
            Console.WriteLine("i\t" + "x[i]\t" + "y[i]\t" + "A[i]\t" + "B[i]\t" + "C[i]\t" + "D[i]\n");
            for (i = 0; i < n; i++)
            {
                Console.WriteLine(i.ToString() + "\t" + x[i].ToString("0.0000") + "\t" + y[i].ToString("0.0000") + "\t" + A[i].ToString("0.0000") + "\t" + B[i].ToString("0.0000") + "\t" + C[i].ToString("0.0000") + "\t" + D[i].ToString("0.0000") + "\n");
            }

            for (i = 0; i <= m; i++)
            {
                if (xvalue >= x[i] && xvalue < x[i + 1])
                {
                    delta = xvalue - x[i];
                    S = A[i] + B[i] * delta + C[i] * delta * delta + D[i] * delta * delta * delta;
                    Console.WriteLine("Interpolated x value = y({0}) = {1}\n", xvalue, S);
                }
            }
        }     

        #endregion

        #region Clamped Cubic Splines

        public static void ClampedCubicSpline(double[] x, double[] y, double fp0, double fpn, double xvalue)
        {
            int i, j, m;
            double S = 0.0;
            double delta = 0.0;

            int n = 4; //max number of coefficients: A,B,C,D for cubic spline

            double[] A = new double[n + 1];
            double[] B = new double[n + 1];
            double[] C = new double[n + 1];
            double[] D = new double[n + 1];
            double[] H = new double[n + 1];
            double[] XA = new double[n + 1];
            double[] XL = new double[n + 1];
            double[] XU = new double[n + 1];
            double[] XZ = new double[n + 1];

            for (i = 0; i < n; i++) A[i] = y[i];

            m = n - 1;
            for (i = 0; i <= m; i++)
                H[i] = x[i + 1] - x[i];

            XA[0] = 3 * (A[1] - A[0]) / H[0] - 3 * fp0;
            XA[n] = 3 * fpn - 3 * (A[n] - A[n - 1]) / H[n - 1];

            for (i = 1; i <= m; i++)
                XA[i] = 3 * (A[i + 1] * H[i - 1] - A[i] * (x[i + 1] - x[i - 1]) + A[i - 1] * H[i]) / (H[i] * H[i - 1]);

            XL[0] = 2 * H[0]; XU[0] = 0.5; XZ[0] = XA[0] / XL[0];
            for (i = 1; i <= m; i++)
            {
                XL[i] = 2 * (x[i + 1] - x[i - 1]) - H[i - 1] * XU[i - 1];
                XU[i] = H[i] / XL[i];
                XZ[i] = (XA[i] - H[i - 1] * XZ[i - 1]) / XL[i];
            }
            XL[n] = H[n - 1] * (2 - XU[n - 1]);
            XZ[n] = (XA[n] - H[n - 1] * XZ[n - 1]) / XL[n];
            C[n] = XZ[n];

            for (i = 1; i <= n; i++)
            {
                j = n - i;
                C[j] = XZ[j] - XU[j] * C[j + 1];
                B[j] = (A[j + 1] - A[j]) / H[j] - H[j] * (C[j + 1] + 2 * C[j]) / 3;
                D[j] = (C[j + 1] - C[j]) / (3 * H[j]);
            }

            Console.WriteLine("\n\nTesting clamped cubic splines");
            Console.WriteLine("Input data stored in (x[i],y[i]). Value to interpolate = {0}\n", xvalue);
            Console.WriteLine("i\t" + "x[i]\t" + "y[i]\t" + "A[i]\t" + "B[i]\t" + "C[i]\t" + "D[i]\n");
            for (i = 0; i < n; i++)
            {
                Console.WriteLine(i.ToString() + "\t" + x[i].ToString("0.0000") + "\t" + y[i].ToString("0.0000") + "\t" + A[i].ToString("0.0000") + "\t" + B[i].ToString("0.0000") + "\t" + C[i].ToString("0.0000") + "\t" + D[i].ToString("0.0000") + "\n");
            }

            for (i = 0; i <= m; i++)
            {
                if (xvalue >= x[i] && xvalue < x[i + 1])
                {
                    delta = xvalue - x[i];
                    S = A[i] + B[i] * delta + C[i] * delta * delta + D[i] * delta * delta * delta;
                    Console.WriteLine("Interpolated x value = y({0}) = {1}\n", xvalue, S);
                }
            }
        }

        #endregion

        static void Main(string[] args)
        {
            TestLinear();
            TestBilinear();
            TestLagrangian();
            TestBarycentric();
            TestNewtonDividedDifference();
            NaturalCubicSplineTest();
            ClampedCubicSplineTest();
        }
    }
}
