﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter08_CodeExamples
{
    class Program
    {
        #region Real Vector Methods from chapter 3
        /*
        *  real zero-based n-dimensional vectors
        *   v = (v[0],v[1],v[2], ... ,v[n-1])
        */
        public struct RVector : ICloneable
        {
            private int ndim;
            private double[] vector;

            public RVector(int ndim)
            {
                this.ndim = ndim;
                this.vector = new double[ndim];
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] = 0.0;
                }
            }

            public RVector(double[] vector)
            {
                this.ndim = vector.Length;
                this.vector = vector;
            }

            public double this[int i]
            {
                get
                {
                    if (i < 0 || i > ndim)
                    {
                        throw new Exception("Requested vector index is out of range!");
                    }
                    return vector[i];
                }
                set { vector[i] = value; }
            }

            public int GetVectorSize
            {
                get { return ndim; }
            }

            public RVector Clone()
            {
                RVector v = new RVector(vector);
                v.vector = (double[])vector.Clone();
                return v;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public RVector SwapVectorEntries(int m, int n)
            {
                double temp = vector[m];
                vector[m] = vector[n];
                vector[n] = temp;
                return new RVector(vector);
            }

            public override string ToString()
            {
                string str = "(";
                for (int i = 0; i < ndim - 1; i++)
                {
                    str += vector[i].ToString() + ", ";
                }
                str += vector[ndim - 1].ToString() + ")";
                return str;
            }

            public override bool Equals(object obj)
            {
                return (obj is RVector) && this.Equals((RVector)obj);
            }

            public bool Equals(RVector v)
            {
                return vector == v.vector;
            }

            public override int GetHashCode()
            {
                return vector.GetHashCode();
            }

            public static bool operator ==(RVector v1, RVector v2)
            {
                return v1.Equals(v2);
            }

            public static bool operator !=(RVector v1, RVector v2)
            {
                return !v1.Equals(v2);
            }

            public static RVector operator +(RVector v)
            {
                return v;
            }

            public static RVector operator +(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] + v2[i];
                }
                return result;
            }

            public static RVector operator -(RVector v)
            {
                double[] result = new double[v.ndim];
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = -v[i];
                }
                return new RVector(result);
            }

            public static RVector operator -(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] - v2[i];
                }
                return result;
            }

            public static RVector operator *(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] * d;
                }
                return result;
            }

            public static RVector operator *(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = d * v[i];
                }
                return result;
            }

            public static RVector operator /(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static RVector operator /(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static double DotProduct(RVector v1, RVector v2)
            {
                double result = 0.0;
                for (int i = 0; i < v1.ndim; i++)
                {
                    result += v1[i] * v2[i];
                }
                return result;
            }

            public double GetNorm()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return Math.Sqrt(result);
            }

            public double GetNormSquare()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return result;
            }

            public void Normalize()
            {
                double norm = GetNorm();
                if (norm == 0)
                {
                    throw new Exception("Tried to normalize a vector with norm of zero!");
                }
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] /= norm;
                }
            }

            public RVector GetUnitVector()
            {
                RVector result = new RVector(vector);
                result.Normalize();
                return result;
            }

            public static RVector CrossProduct(RVector v1, RVector v2)
            {
                if (v1.ndim != 3)
                {
                    throw new Exception("Vector v1 must be 3 dimensional!");
                }
                if (v2.ndim != 3)
                {
                    throw new Exception("Vector v2 must be 3 dimensional!");
                }
                RVector result = new RVector(3);
                result[0] = v1[1] * v2[2] - v1[2] * v2[1];
                result[1] = v1[2] * v2[0] - v1[0] * v2[2];
                result[2] = v1[0] * v2[1] - v1[1] * v2[0];
                return result;
            }
        }
        #endregion

        #region Real Matrix Methods from chapter 3
        /*
         *  real (row x col) matrix         
         */

        public struct RMatrix : ICloneable
        {
            private int nRows;
            private int nCols;
            private double[,] matrix;

            public RMatrix(int nRows, int nCols)
            {
                this.nRows = nRows;
                this.nCols = nCols;
                this.matrix = new double[nRows, nCols];
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        matrix[i, j] = 0.0;
                    }
                }
            }

            public RMatrix(double[,] matrix)
            {
                this.nRows = matrix.GetLength(0);
                this.nCols = matrix.GetLength(1);
                this.matrix = matrix;
            }

            public RMatrix(RMatrix m)
            {
                nRows = m.GetnRows;
                nCols = m.GetnCols;
                matrix = m.matrix;
            }

            public RMatrix IdentityMatrix()
            {
                RMatrix m = new RMatrix(nRows, nCols);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        if (i == j)
                        {
                            m[i, j] = 1;
                        }
                    }
                }
                return m;
            }

            public double this[int m, int n]
            {
                get
                {
                    if (m < 0 || m > nRows)
                    {
                        throw new Exception("m-th row is out of range!");
                    }
                    if (n < 0 || n > nCols)
                    {
                        throw new Exception("n-th col is out of range!");
                    }
                    return matrix[m, n];
                }
                set { matrix[m, n] = value; }
            }

            public int GetnRows
            {
                get { return nRows; }
            }

            public int GetnCols
            {
                get { return nCols; }
            }

            public RMatrix Clone()
            {
                RMatrix m = new RMatrix(matrix);
                m.matrix = (double[,])matrix.Clone();
                return m;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public override string ToString()
            {
                string strMatrix = "(";
                for (int i = 0; i < nRows; i++)
                {
                    string str = "";
                    for (int j = 0; j < nCols - 1; j++)
                    {
                        str += matrix[i, j].ToString() + ", ";
                    }
                    str += matrix[i, nCols - 1].ToString();
                    if (i != nRows - 1 && i == 0)
                        strMatrix += str + "\n";
                    else if (i != nRows - 1 && i != 0)
                        strMatrix += " " + str + "\n";
                    else
                        strMatrix += " " + str + ")";
                }
                return strMatrix;
            }

            public override bool Equals(object obj)
            {
                return (obj is RMatrix) && this.Equals((RMatrix)obj);
            }

            public bool Equals(RMatrix m)
            {
                return matrix == m.matrix;
            }

            public override int GetHashCode()
            {
                return matrix.GetHashCode();
            }

            public static bool operator ==(RMatrix m1, RMatrix m2)
            {
                return m1.Equals(m2);
            }

            public static bool operator !=(RMatrix m1, RMatrix m2)
            {
                return !m1.Equals(m2);
            }

            public static RMatrix operator +(RMatrix m)
            {
                return m;
            }

            public static RMatrix operator +(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] + m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator -(RMatrix m)
            {
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        m[i, j] = -m[i, j];
                    }
                }
                return m;
            }

            public static RMatrix operator -(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] - m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnCols != m2.GetnRows)
                {
                    throw new Exception("The numbers of columns of the" +
                     " first matrix must be equal to the number of " +
                     " rows of the second matrix!");
                }
                double tmp;
                RMatrix result = new RMatrix(m1.GetnRows, m2.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m2.GetnCols; j++)
                    {
                        tmp = result[i, j];
                        for (int k = 0; k < result.GetnRows; k++)
                        {
                            tmp += m1[i, k] * m2[k, j];
                        }
                        result[i, j] = tmp;
                    }
                }
                return result;
            }

            public RMatrix GetTranspose()
            {
                RMatrix m = this;
                m.Transpose();
                return m;
            }

            public void Transpose()
            {
                RMatrix m = new RMatrix(nCols, nRows);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        m[j, i] = matrix[i, j];
                    }
                }
                this = m;
            }

            public double GetTrace()
            {
                double sum_of_diag = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    if (i < nCols)
                        sum_of_diag += matrix[i, i];
                }
                return sum_of_diag;
            }

            public bool IsSquared()
            {
                if (nRows == nCols)
                    return true;
                else
                    return false;
            }

            public static bool CompareDimension(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnRows == m2.GetnRows && m1.GetnCols == m2.GetnCols)
                    return true;
                else
                    return false;
            }

            public RVector GetRowVector(int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                RVector RowVector = new RVector(nCols);
                for (int i = 0; i < nCols; i++)
                {
                    RowVector[i] = matrix[m, i];
                }
                return RowVector;
            }

            public RVector GetColVector(int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                RVector ColVector = new RVector(nRows);
                for (int i = 0; i < nRows; i++)
                {
                    ColVector[i] = matrix[i, n];
                }
                return ColVector;
            }

            public RMatrix ReplaceRow(RVector vec, int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                if (vec.GetVectorSize != nCols)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nCols; i++)
                {
                    matrix[m, i] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix ReplaceCol(RVector vec, int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                if (vec.GetVectorSize != nRows)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nRows; i++)
                {
                    matrix[i, n] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapMatrixRow(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nCols; i++)
                {
                    temp = matrix[m, i];
                    matrix[m, i] = matrix[n, i];
                    matrix[n, i] = temp;
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapMatrixColumn(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    temp = matrix[i, m];
                    matrix[i, m] = matrix[i, n];
                    matrix[i, n] = temp;
                }
                return new RMatrix(matrix);
            }

            public static RVector Transform(RMatrix mat, RVector vec)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnCols != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of cols of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += mat[i, j] * vec[j];
                    }
                }
                return result;
            }

            public static RVector Transform(RVector vec, RMatrix mat)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of rows of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += vec[j] * mat[j, i];
                    }
                }
                return result;
            }

            public static RMatrix Transform(RVector v1, RVector v2)
            {
                if (v1.GetVectorSize != v2.GetVectorSize)
                {
                    throw new Exception("The vectors must have the same ndim!");
                }
                RMatrix result = new RMatrix(v1.GetVectorSize, v1.GetVectorSize);
                for (int i = 0; i < v1.GetVectorSize; i++)
                {
                    for (int j = 0; j < v1.GetVectorSize; j++)
                    {
                        result[j, i] = v1[i] * v2[j];
                    }
                }
                return result;
            }

            public static double Determinant(RMatrix mat)
            {
                double result = 0.0;
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows == 1)
                    result = mat[0, 0];
                else
                {
                    for (int i = 0; i < mat.GetnRows; i++)
                    {
                        result += Math.Pow(-1, i) * mat[0, i] * Determinant(RMatrix.Minor(mat, 0, i));
                    }
                }
                return result;
            }

            public static RMatrix Minor(RMatrix mat, int row, int col)
            {
                RMatrix mm = new RMatrix(mat.GetnRows - 1, mat.GetnCols - 1);
                int ii = 0, jj = 0;
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    if (i == row)
                        continue;
                    jj = 0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        if (j == col)
                            continue;
                        mm[ii, jj] = mat[i, j];
                        jj++;
                    }
                    ii++;
                }
                return mm;
            }

            public static RMatrix Adjoint(RMatrix mat)
            {
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                RMatrix ma = new RMatrix(mat.GetnRows, mat.GetnCols);
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        ma[i, j] = Math.Pow(-1, i + j) * (Determinant(Minor(mat, i, j)));
                    }
                }
                return ma.GetTranspose();
            }

            public static RMatrix Inverse(RMatrix mat)
            {
                if (Determinant(mat) == 0)
                {
                    throw new Exception("Cannot inverse a matrix with a zero determinant!");
                }
                return (Adjoint(mat) / Determinant(mat));
            }
        }
        #endregion

        #region Real Linear Equations Methods
        const double epsilon = 1.0e-500;

        // Gauss-Jordan elimination to solve Ax = b for x
        public static RVector GaussJordan(RMatrix A, RVector b)
        {
            Triangulate(A, b);
            int bSize = b.GetVectorSize;
            RVector x = new RVector(bSize);
            for (int i = bSize - 1; i >= 0; i--)
            {
                double Aii = A[i, i];
                if (Math.Abs(Aii) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                x[i] = (b[i] - RVector.DotProduct(A.GetRowVector(i), x)) / Aii;
            }
            return x;
        }

        // Triangulate matrix A
        private static void Triangulate(RMatrix A, RVector b)
        {
            int nRows = A.GetnRows;
            for (int i = 0; i < nRows - 1; i++)
            {
                double diagonalElement = pivotGaussJordan(A, b, i);
                if (Math.Abs(diagonalElement) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                for (int j = i + 1; j < nRows; j++)
                {
                    double w = A[j, i] / diagonalElement;
                    for (int k = i + 1; k < nRows; k++)
                    {
                        A[j, k] -= w * A[i, k];
                    }
                    b[j] -= w * b[i];
                }
            }
        }

        private static double pivotGaussJordan(RMatrix A, RVector b, int q)
        {
            int bSize = b.GetVectorSize;
            int c = q;
            double d = 0.0;
            for (int j = q; j < bSize; j++)
            {
                double w = Math.Abs(A[j, q]);
                if (w > d)
                {
                    d = w;
                    c = j;
                }
            }
            if (c > q)
            {
                A.SwapMatrixRow(q, c);
                b.SwapVectorEntries(q, c);
            }
            return A[q, q];
        }

        // LU decomposition using the Crout algorithm with pivoting
        public static double LUCrout(RMatrix A, RVector b)
        {
            LUDecompose(A);
            return LUSubstitute(A, b);
        }

        private static void LUDecompose(RMatrix matrix)
        {
            int nRows = matrix.GetnRows;
            for (int i = 0; i < nRows; i++)
            {
                for (int j = 0; j < nRows; j++)
                {
                    double w = matrix[i, j];
                    for (int k = 0; k < Math.Min(i, j); k++)
                    {
                        w -= matrix[i, k] * matrix[k, j];
                    }
                    if (j > i)
                    {
                        double s = matrix[i, i];
                        if (Math.Abs(w) < epsilon)
                            throw new Exception("Diagonal element is too small!");
                        w /= s;
                    }
                    matrix[i, j] = w;
                }
            }
        }

        private static double LUSubstitute(RMatrix matrix, RVector vec) 
        {
            int size = vec.GetVectorSize;
            double det = 1.0;
            for (int i = 0; i < size; i++)               
            {
                double w = vec[i];
                for (int j = 0; j < i; j++)
                {
                    w -= matrix[i, j] * vec[j];
                }
                double p = matrix[i, i];
                if (Math.Abs(w) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                w /= p;
                vec[i] = w;
                det *= matrix[i, i];
            }
            for (int i = size - 1; i >= 0; i--)
            {
                double s = vec[i];
                for (int j = i + 1; j < size; j++)
                {
                    s -= matrix[i, j] * vec[j];
                }
                vec[i] = s;       
            }
            return det;
        }

        //LU Matrix Inverse
        public static RMatrix LUInverse(RMatrix matrix)
        {
            int nRows = matrix.GetnRows;
            RMatrix u = matrix.IdentityMatrix();
            LUDecompose(matrix);
            RVector uv = new RVector(nRows);
            for (int i = 0; i < nRows; i++)
            {
                uv = u.GetRowVector(i);
                LUSubstitute(matrix, uv);
                u.ReplaceRow(uv, i);
            }
            RMatrix inverse = u.GetTranspose();
            return inverse;
        }

        // Gauss-Jacobi method
        public static RVector GaussJacobi(RMatrix A, RVector b, int MaxIterations, double tolerance)
        {
            int bSize = b.GetVectorSize;
            RVector x = new RVector(bSize);
            for (int nIteration = 0; nIteration < MaxIterations; nIteration++)
            {
                RVector xOld = x.Clone();
                for (int i = 0; i < bSize; i++)
                {
                    double entry = b[i];
                    double diagonal = A[i, i];
                    if (Math.Abs(diagonal) < epsilon)
                        throw new Exception("Diagonal element is too small!");
                    for (int j = 0; j < bSize; j++)
                    {
                        if (j != i)
                        {
                            entry -= A[i, j] * xOld[j];
                        }
                    }
                    x[i] = entry / diagonal;
                }
                RVector dx = x - xOld;
                if (dx.GetNorm() < tolerance)
                {
                    return x;
                }
            }
            return x;
        }

        // Gauss-Seidel method
        public static RVector GaussSeidel(RMatrix A, RVector b, int MaxIterations, double tolerance)
        {
            int size = b.GetVectorSize;
            RVector x = new RVector(size);
            for (int nIteration = 0; nIteration < MaxIterations; nIteration++)
            {
                RVector xOld = x.Clone();
                for (int i = 0; i < size; i++)
                {
                    double entry = b[i];
                    double diagonal = A[i, i];
                    if (Math.Abs(diagonal) < epsilon)
                        throw new Exception("Diagonal element is too small!");
                    for (int j = 0; j < i; j++)
                    {
                        entry -= A[i, j] * x[j];
                    }
                    for (int j = i + 1; j < size; j++)
                    {
                        entry -= A[i, j] * xOld[j];
                    }
                    x[i] = entry / diagonal;
                }
                RVector dx = x - xOld;
                if (dx.GetNorm() < tolerance)
                {
                    return x;
                }
            }
            return x;
        }
        #endregion

        #region Eigenvalues and Eigenvectors
        public static void JacobiEigenValVec(double[,] a, int maxsize, int n, double epsilon, out double[,] eigenval, out double[,] eigenvec)
        {
            int i, j, p, q, flag;
            double[,] d = new double[maxsize, maxsize];
            double[,] s = new double[maxsize, maxsize];
            double[,] s1 = new double[maxsize, maxsize];
            double[,] s1t = new double[maxsize, maxsize];
            double[,] temp = new double[maxsize, maxsize];
            double theta, max;

            //Initialization of matrix d and s
            for (i = 1; i <= n; i++)
            {
                for (j = 1; j <= n; j++)
                {
                    d[i, j] = a[i, j];
                    s[i, j] = 0.0;
                }
            }
            for (i = 1; i <= n; i++) s[i, i] = 1.0;

            do
            {
                flag = 0;
                //Find largest off-diagonal element
                i = 1;
                j = 2;
                max = Math.Abs(d[1, 2]);
                for (p = 1; p <= n; p++)
                {
                    for (q = 1; q <= n; q++)
                    {
                        if (p != q) //off diagonal element
                        {
                            if (max < Math.Abs(d[p, q]))
                            {
                                max = Math.Abs(d[p, q]);
                                i = p;
                                j = q;
                            }
                        }
                    }
                }

                if (d[i, i] == d[j, j])
                {
                    if (d[i, j] > 0) theta = Math.PI / 4.0;
                    else theta = -Math.PI / 4.0;
                }
                else
                {
                    theta = 0.5 * Math.Atan(2.0 * d[i, j] / (d[i, i] - d[j, j]));
                }

                //Construction of the matrix s1 and s1t
                for (p = 1; p <= n; p++)
                {
                    for (q = 1; q <= n; q++)
                    {
                        s1[p, q] = 0.0;
                        s1t[p, q] = 0.0;
                    }
                }

                for (p = 1; p <= n; p++)
                {
                    s1[p, p] = 1.0;
                    s1t[p, p] = 1.0;
                }

                s1[i, i] = Math.Cos(theta); s1[j, j] = s1[i, i];
                s1[j, i] = Math.Sin(theta); s1[i, j] = -s1[j, i];
                s1t[i, i] = s1[i, i]; s1t[j, j] = s1[j, j];
                s1t[i, j] = s1[j, i]; s1t[j, i] = s1[i, j];

                //Product of s1t and d
                for (i = 1; i <= n; i++)
                {
                    for (j = 1; j <= n; j++)
                    {
                        temp[i, j] = 0.0;
                        for (p = 1; p <= n; p++)
                            temp[i, j] += s1t[i, p] * d[p, j];
                    }
                }

                //Product of temp and s1: d = s1t * d * s1
                for (i = 1; i <= n; i++)
                {
                    for (j = 1; j <= n; j++)
                    {
                        d[i, j] = 0.0;
                        for (p = 1; p <= n; p++)
                            d[i, j] += temp[i, p] * s1[p, j];
                    }
                }

                //Product of s and s1: s = s*s1
                for (i = 1; i <= n; i++)
                {
                    for (j = 1; j <= n; j++)
                    {
                        temp[i, j] = 0.0;
                        for (p = 1; p <= n; p++)
                            temp[i, j] += s[i, p] * s1[p, j];
                    }
                }

                for (i = 1; i <= n; i++)
                {
                    for (j = 1; j <= n; j++)
                    {
                        s[i, j] = temp[i, j];
                    }
                }

                //check to see if d is a diagonal matrix
                for (i = 1; i <= n; i++)
                {
                    for (j = 1; j <= n; j++)
                    {
                        if (i != j)
                            if (Math.Abs(d[i, j]) > epsilon)
                                flag = 1;
                    }
                }

            } while (flag == 1);
            //copy results to output matrices
            eigenval = d;
            eigenvec = s;
        }       
        #endregion

        #region Test Functions
        static void TestGaussJordan()
        {
            RMatrix A = new RMatrix(new double[3, 3] { { 2, 4, -6 }, { 6, -4, 2 }, { 4, 2, 6} });
            RVector b = new RVector(new double[3] { 6, -2, 4 });
            RVector x = GaussJordan(A, b);
            Console.WriteLine("Solution x = {0}", x);
        }

        static void TestLU()
        {
            RMatrix A = new RMatrix(new double[3, 3] { { 2, 4, -6 }, { 6, -4, 2 }, { 4, 2, 6 } });
            RVector b = new RVector(new double[3] { 4, 4, 8 });
            RMatrix Anew = A.Clone();
            RMatrix Bnew = A.Clone();
            double d = LUCrout(A, b);
            RMatrix inv = LUInverse(Anew);
            Console.WriteLine("\n Inverse of A = \n {0}", (inv));
            Console.WriteLine("\nSolution of the equations Ax = B is x = {0}", b);
            Console.WriteLine("\nDeterminant of A = {0}", d);
            Console.WriteLine("\nTest Inverse: BB*Inverse = \n {0}", Bnew * inv);
        }

        static void TestGaussJacobiIteration()
        {
            RMatrix A = new RMatrix(new double[3, 3] {{4,0,1},{0,3,2},{1,2,4}});
            RVector b = new RVector(new double[3] {2,1,3} );
            RMatrix A1 = A.Clone();
            RVector b1 = b.Clone();

            RVector x = GaussJacobi(A, b, 10, 1.0e-4);
            Console.WriteLine("\n Solution from the Gauss-Jacobi iteration:");
            Console.WriteLine(" x[0] = {0}", x[0]);
            Console.WriteLine(" x[1] = {0}", x[1]);
            Console.WriteLine(" x[2] = {0}", x[2]);
        }

        static void TestGaussSeidelIteration()
        {
            RMatrix A = new RMatrix(new double[3, 3] {{4,0,1},{0,3,2},{1,2,4}});
            RVector b = new RVector(new double[3] { 2, 1, 3 });
            RMatrix A1 = A.Clone();
            RVector b1 = b.Clone();

            RVector x1 = GaussSeidel(A1, b1, 10, 1.0e-4);
            Console.WriteLine("\n Solution from the Gauss-Seidel iteration:");
            Console.WriteLine(" x1[0] = {0}", x1[0]);
            Console.WriteLine(" x1[1] = {0}", x1[1]);
            Console.WriteLine(" x1[2] = {0}", x1[2]);
        }

        public static void TestingJacobiEigenValVec()
        {
            int nCol = 4;                 //size of input 4x4 matrix
            int maxMatrixSize = 10;     //max matrix size
            double Epsilon = 1.0e-04;   //tolerance
            double[,] A = new double[maxMatrixSize, maxMatrixSize];
            double[,] Eigenvalues = new double[maxMatrixSize, maxMatrixSize];
            double[,] Eigenvectors = new double[maxMatrixSize, maxMatrixSize];

            A[1, 1] = 1; A[1, 2] = 2; A[1, 3] = 3; A[1, 4] = 4;
            A[2, 1] = 2; A[2, 2] = -3; A[2, 3] = 3; A[2, 4] = 4;
            A[3, 1] = 3; A[3, 2] = 3; A[3, 3] = 4; A[3, 4] = 5;
            A[4, 1] = 4; A[4, 2] = 4; A[4, 3] = 5; A[4, 4] = 0;

            Console.WriteLine("\n\nTesting Jacobi's Method For Finding Eigenvalues and Eigenvectors\n");
            Console.WriteLine("The input matrix is given by\n");
            for (int i = 1; i <= nCol; i++)
            {
                for (int j = 1; j <= nCol; j++)
                {
                    if (j != nCol)
                        Console.Write(A[i, j].ToString("0.000000") + "\t");
                    else
                        Console.WriteLine(A[i, j].ToString("0.000000"));
                }
            }

            //Calculate eigenvalues and eigenvectors using Jacobi method
            JacobiEigenValVec(A, maxMatrixSize, nCol, Epsilon, out Eigenvalues, out Eigenvectors);

            //Output results
            Console.WriteLine("\nThe eigenvalues are:\n");
            for (int i = 1; i <= nCol; i++)
            {
                Console.WriteLine(Eigenvalues[i, i].ToString("0.00000"));
            }

            Console.WriteLine("\nThe corresponding eigenvectors are\n");
            for (int j = 1; j <= nCol; j++)
            {
                for (int i = 1; i <= nCol; i++)
                {
                    if (i != nCol)
                        Console.Write(Eigenvectors[i, j].ToString("0.00000") + "\t\t");
                    else
                        Console.WriteLine(Eigenvectors[i, j].ToString("0.00000"));
                }
            }
        }
        #endregion

        static void Main(string[] args)
        {                   
            TestGaussJordan();
            TestLU();
            TestGaussJacobiIteration();
            TestGaussSeidelIteration();
            TestingJacobiEigenValVec();
            Console.ReadLine();
        }
    }
}
