﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace Chapter15_CodeExamples
{
    class Program
    {
        #region Real Vector Methods from Chapter 3
        /*
        *  real zero-based n-dimensional vectors
        *   v = (v[0],v[1],v[2], ... ,v[n-1])
        */
        public struct RVector : ICloneable
        {
            private int ndim;
            private double[] vector;

            public RVector(int ndim)
            {
                this.ndim = ndim;
                this.vector = new double[ndim];
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] = 0.0;
                }
            }

            public RVector(double[] vector)
            {
                this.ndim = vector.Length;
                this.vector = vector;
            }

            public double this[int i]
            {
                get
                {
                    if (i < 0 || i > ndim)
                    {
                        throw new Exception("Requested vector index is out of range!");
                    }
                    return vector[i];
                }
                set { vector[i] = value; }
            }

            public int GetVectorSize
            {
                get { return ndim; }
            }

            public RVector Clone()
            {
                RVector v = new RVector(vector);
                v.vector = (double[])vector.Clone();
                return v;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public RVector SwapVectorEntries(int m, int n)
            {
                double temp = vector[m];
                vector[m] = vector[n];
                vector[n] = temp;
                return new RVector(vector);
            }

            public override string ToString()
            {
                string str = "(";
                for (int i = 0; i < ndim - 1; i++)
                {
                    str += vector[i].ToString() + ", ";
                }
                str += vector[ndim - 1].ToString() + ")";
                return str;
            }

            public override bool Equals(object obj)
            {
                return (obj is RVector) && this.Equals((RVector)obj);
            }

            public bool Equals(RVector v)
            {
                return vector == v.vector;
            }

            public override int GetHashCode()
            {
                return vector.GetHashCode();
            }

            public static bool operator ==(RVector v1, RVector v2)
            {
                return v1.Equals(v2);
            }

            public static bool operator !=(RVector v1, RVector v2)
            {
                return !v1.Equals(v2);
            }

            public static RVector operator +(RVector v)
            {
                return v;
            }

            public static RVector operator +(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] + v2[i];
                }
                return result;
            }

            public static RVector operator -(RVector v)
            {
                double[] result = new double[v.ndim];
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = -v[i];
                }
                return new RVector(result);
            }

            public static RVector operator -(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] - v2[i];
                }
                return result;
            }

            public static RVector operator *(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] * d;
                }
                return result;
            }

            public static RVector operator *(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = d * v[i];
                }
                return result;
            }

            public static RVector operator /(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static RVector operator /(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static double DotProduct(RVector v1, RVector v2)
            {
                double result = 0.0;
                for (int i = 0; i < v1.ndim; i++)
                {
                    result += v1[i] * v2[i];
                }
                return result;
            }

            public double GetNorm()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return Math.Sqrt(result);
            }

            public double GetNormSquare()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return result;
            }

            public void Normalize()
            {
                double norm = GetNorm();
                if (norm == 0)
                {
                    throw new Exception("Tried to normalize a vector with norm of zero!");
                }
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] /= norm;
                }
            }

            public RVector GetUnitVector()
            {
                RVector result = new RVector(vector);
                result.Normalize();
                return result;
            }

            public static RVector CrossProduct(RVector v1, RVector v2)
            {
                if (v1.ndim != 3)
                {
                    throw new Exception("Vector v1 must be 3 dimensional!");
                }
                if (v2.ndim != 3)
                {
                    throw new Exception("Vector v2 must be 3 dimensional!");
                }
                RVector result = new RVector(3);
                result[0] = v1[1] * v2[2] - v1[2] * v2[1];
                result[1] = v1[2] * v2[0] - v1[0] * v2[2];
                result[2] = v1[0] * v2[1] - v1[1] * v2[0];
                return result;
            }
        }
        #endregion

        #region Real Matrix Methods from Chapter 3
        /*
         *  real (row x col) matrix         
         */

        public struct RMatrix : ICloneable
        {
            private int nRows;
            private int nCols;
            private double[,] matrix;

            public RMatrix(int nRows, int nCols)
            {
                this.nRows = nRows;
                this.nCols = nCols;
                this.matrix = new double[nRows, nCols];
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        matrix[i, j] = 0.0;
                    }
                }
            }

            public RMatrix(double[,] matrix)
            {
                this.nRows = matrix.GetLength(0);
                this.nCols = matrix.GetLength(1);
                this.matrix = matrix;
            }

            public RMatrix(RMatrix m)
            {
                nRows = m.GetnRows;
                nCols = m.GetnCols;
                matrix = m.matrix;
            }

            public RMatrix IdentityMatrix()
            {
                RMatrix m = new RMatrix(nRows, nCols);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        if (i == j)
                        {
                            m[i, j] = 1;
                        }
                    }
                }
                return m;
            }

            public double this[int m, int n]
            {
                get
                {
                    if (m < 0 || m > nRows)
                    {
                        throw new Exception("m-th row is out of range!");
                    }
                    if (n < 0 || n > nCols)
                    {
                        throw new Exception("n-th col is out of range!");
                    }
                    return matrix[m, n];
                }
                set { matrix[m, n] = value; }
            }

            public int GetnRows
            {
                get { return nRows; }
            }

            public int GetnCols
            {
                get { return nCols; }
            }

            public RMatrix Clone()
            {
                RMatrix m = new RMatrix(matrix);
                m.matrix = (double[,])matrix.Clone();
                return m;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public override string ToString()
            {
                string strMatrix = "(";
                for (int i = 0; i < nRows; i++)
                {
                    string str = "";
                    for (int j = 0; j < nCols - 1; j++)
                    {
                        str += matrix[i, j].ToString() + ", ";
                    }
                    str += matrix[i, nCols - 1].ToString();
                    if (i != nRows - 1 && i == 0)
                        strMatrix += str + "\n";
                    else if (i != nRows - 1 && i != 0)
                        strMatrix += " " + str + "\n";
                    else
                        strMatrix += " " + str + ")";
                }
                return strMatrix;
            }

            public override bool Equals(object obj)
            {
                return (obj is RMatrix) && this.Equals((RMatrix)obj);
            }

            public bool Equals(RMatrix m)
            {
                return matrix == m.matrix;
            }

            public override int GetHashCode()
            {
                return matrix.GetHashCode();
            }

            public static bool operator ==(RMatrix m1, RMatrix m2)
            {
                return m1.Equals(m2);
            }

            public static bool operator !=(RMatrix m1, RMatrix m2)
            {
                return !m1.Equals(m2);
            }

            public static RMatrix operator +(RMatrix m)
            {
                return m;
            }

            public static RMatrix operator +(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] + m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator -(RMatrix m)
            {
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        m[i, j] = -m[i, j];
                    }
                }
                return m;
            }

            public static RMatrix operator -(RMatrix m1, RMatrix m2)
            {
                if (!RMatrix.CompareDimension(m1, m2))
                {
                    throw new Exception("The dimensions of two matrices must be the same!");
                }
                RMatrix result = new RMatrix(m1.GetnRows, m1.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m1.GetnCols; j++)
                    {
                        result[i, j] = m1[i, j] - m2[i, j];
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] * d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(RMatrix m, double d)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator /(double d, RMatrix m)
            {
                RMatrix result = new RMatrix(m.GetnRows, m.GetnCols);
                for (int i = 0; i < m.GetnRows; i++)
                {
                    for (int j = 0; j < m.GetnCols; j++)
                    {
                        result[i, j] = m[i, j] / d;
                    }
                }
                return result;
            }

            public static RMatrix operator *(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnCols != m2.GetnRows)
                {
                    throw new Exception("The numbers of columns of the" +
                     " first matrix must be equal to the number of " +
                     " rows of the second matrix!");
                }
                double tmp;
                RMatrix result = new RMatrix(m1.GetnRows, m2.GetnCols);
                for (int i = 0; i < m1.GetnRows; i++)
                {
                    for (int j = 0; j < m2.GetnCols; j++)
                    {
                        tmp = result[i, j];
                        for (int k = 0; k < result.GetnRows; k++)
                        {
                            tmp += m1[i, k] * m2[k, j];
                        }
                        result[i, j] = tmp;
                    }
                }
                return result;
            }

            public RMatrix GetTranspose()
            {
                RMatrix m = this;
                m.Transpose();
                return m;
            }

            public void Transpose()
            {
                RMatrix m = new RMatrix(nCols, nRows);
                for (int i = 0; i < nRows; i++)
                {
                    for (int j = 0; j < nCols; j++)
                    {
                        m[j, i] = matrix[i, j];
                    }
                }
                this = m;
            }

            public double GetTrace()
            {
                double sum_of_diag = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    if (i < nCols)
                        sum_of_diag += matrix[i, i];
                }
                return sum_of_diag;
            }

            public bool IsSquared()
            {
                if (nRows == nCols)
                    return true;
                else
                    return false;
            }

            public static bool CompareDimension(RMatrix m1, RMatrix m2)
            {
                if (m1.GetnRows == m2.GetnRows && m1.GetnCols == m2.GetnCols)
                    return true;
                else
                    return false;
            }

            public RVector GetRowVector(int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                RVector RowVector = new RVector(nCols);
                for (int i = 0; i < nCols; i++)
                {
                    RowVector[i] = matrix[m, i];
                }
                return RowVector;
            }

            public RVector GetColVector(int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                RVector ColVector = new RVector(nRows);
                for (int i = 0; i < nRows; i++)
                {
                    ColVector[i] = matrix[i, n];
                }
                return ColVector;
            }

            public RMatrix ReplaceRow(RVector vec, int m)
            {
                if (m < 0 || m > nRows)
                {
                    throw new Exception("m-th row is out of range!");
                }
                if (vec.GetVectorSize != nCols)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nCols; i++)
                {
                    matrix[m, i] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix ReplaceCol(RVector vec, int n)
            {
                if (n < 0 || n > nCols)
                {
                    throw new Exception("n-th col is out of range!");
                }
                if (vec.GetVectorSize != nRows)
                {
                    throw new Exception("Vector ndim is out of range!");
                }
                for (int i = 0; i < nRows; i++)
                {
                    matrix[i, n] = vec[i];
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapRMatrixow(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nCols; i++)
                {
                    temp = matrix[m, i];
                    matrix[m, i] = matrix[n, i];
                    matrix[n, i] = temp;
                }
                return new RMatrix(matrix);
            }

            public RMatrix SwapMatrixColumn(int m, int n)
            {
                double temp = 0.0;
                for (int i = 0; i < nRows; i++)
                {
                    temp = matrix[i, m];
                    matrix[i, m] = matrix[i, n];
                    matrix[i, n] = temp;
                }
                return new RMatrix(matrix);
            }

            public static RVector Transform(RMatrix mat, RVector vec)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnCols != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of cols of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += mat[i, j] * vec[j];
                    }
                }
                return result;
            }

            public static RVector Transform(RVector vec, RMatrix mat)
            {
                RVector result = new RVector(vec.GetVectorSize);
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows != vec.GetVectorSize)
                {
                    throw new Exception("The ndim of the vector must be equal "
                        + "to the number of rows of the matrix!");
                }
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    result[i] = 0.0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        result[i] += vec[j] * mat[j, i];
                    }
                }
                return result;
            }

            public static RMatrix Transform(RVector v1, RVector v2)
            {
                if (v1.GetVectorSize != v2.GetVectorSize)
                {
                    throw new Exception("The vectors must have the same ndim!");
                }
                RMatrix result = new RMatrix(v1.GetVectorSize, v1.GetVectorSize);
                for (int i = 0; i < v1.GetVectorSize; i++)
                {
                    for (int j = 0; j < v1.GetVectorSize; j++)
                    {
                        result[j, i] = v1[i] * v2[j];
                    }
                }
                return result;
            }

            public static double Determinant(RMatrix mat)
            {
                double result = 0.0;
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                if (mat.GetnRows == 1)
                    result = mat[0, 0];
                else
                {
                    for (int i = 0; i < mat.GetnRows; i++)
                    {
                        result += Math.Pow(-1, i) * mat[0, i] * Determinant(RMatrix.Minor(mat, 0, i));
                    }
                }
                return result;
            }

            public static RMatrix Minor(RMatrix mat, int row, int col)
            {
                RMatrix mm = new RMatrix(mat.GetnRows - 1, mat.GetnCols - 1);
                int ii = 0, jj = 0;
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    if (i == row)
                        continue;
                    jj = 0;
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        if (j == col)
                            continue;
                        mm[ii, jj] = mat[i, j];
                        jj++;
                    }
                    ii++;
                }
                return mm;
            }

            public static RMatrix Adjoint(RMatrix mat)
            {
                if (!mat.IsSquared())
                {
                    throw new Exception("The matrix must be squared!");
                }
                RMatrix ma = new RMatrix(mat.GetnRows, mat.GetnCols);
                for (int i = 0; i < mat.GetnRows; i++)
                {
                    for (int j = 0; j < mat.GetnCols; j++)
                    {
                        ma[i, j] = Math.Pow(-1, i + j) * (Determinant(Minor(mat, i, j)));
                    }
                }
                return ma.GetTranspose();
            }

            public static RMatrix Inverse(RMatrix mat)
            {
                if (Determinant(mat) == 0)
                {
                    throw new Exception("Cannot inverse a matrix with a zero determinant!");
                }
                return (Adjoint(mat) / Determinant(mat));
            }
        }
        #endregion

        #region Real Linear Equation Methods from Chapter 8
        const double epsilon = 1.0e-500;

        // Gauss-Jordan elimination to solve Ax = b for x
        public static RVector GaussJordan(RMatrix A, RVector b)
        {
            Triangulate(A, b);
            int bSize = b.GetVectorSize;
            RVector x = new RVector(bSize);
            for (int i = bSize - 1; i >= 0; i--)
            {
                double Aii = A[i, i];
                if (Math.Abs(Aii) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                x[i] = (b[i] - RVector.DotProduct(A.GetRowVector(i), x)) / Aii;
            }
            return x;
        }

        // Triangulate matrix A
        private static void Triangulate(RMatrix A, RVector b)
        {
            int nRows = A.GetnRows;
            for (int i = 0; i < nRows - 1; i++)
            {
                double diagonalElement = pivotGaussJordan(A, b, i);
                if (Math.Abs(diagonalElement) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                for (int j = i + 1; j < nRows; j++)
                {
                    double w = A[j, i] / diagonalElement;
                    for (int k = i + 1; k < nRows; k++)
                    {
                        A[j, k] -= w * A[i, k];
                    }
                    b[j] -= w * b[i];
                }
            }
        }

        private static double pivotGaussJordan(RMatrix A, RVector b, int q)
        {
            int bSize = b.GetVectorSize;
            int c = q;
            double d = 0.0;
            for (int j = q; j < bSize; j++)
            {
                double w = Math.Abs(A[j, q]);
                if (w > d)
                {
                    d = w;
                    c = j;
                }
            }
            if (c > q)
            {
                A.SwapRMatrixow(q, c);
                b.SwapVectorEntries(q, c);
            }
            return A[q, q];
        }

        // LU decomposition using the Crout algorithm with pivoting
        public static double LUCrout(RMatrix A, RVector b)
        {
            LUDecompose(A);
            return LUSubstitute(A, b);
        }

        private static void LUDecompose(RMatrix matrix)
        {
            int nRows = matrix.GetnRows;
            for (int i = 0; i < nRows; i++)
            {
                for (int j = 0; j < nRows; j++)
                {
                    double w = matrix[i, j];
                    for (int k = 0; k < Math.Min(i, j); k++)
                    {
                        w -= matrix[i, k] * matrix[k, j];
                    }
                    if (j > i)
                    {
                        double s = matrix[i, i];
                        if (Math.Abs(w) < epsilon)
                            throw new Exception("Diagonal element is too small!");
                        w /= s;
                    }
                    matrix[i, j] = w;
                }
            }
        }

        private static double LUSubstitute(RMatrix matrix, RVector vec)
        {
            int size = vec.GetVectorSize;
            double det = 1.0;
            for (int i = 0; i < size; i++)
            {
                double w = vec[i];
                for (int j = 0; j < i; j++)
                {
                    w -= matrix[i, j] * vec[j];
                }
                double p = matrix[i, i];
                if (Math.Abs(w) < epsilon)
                    throw new Exception("Diagonal element is too small!");
                w /= p;
                vec[i] = w;
                det *= matrix[i, i];
            }
            for (int i = size - 1; i >= 0; i--)
            {
                double s = vec[i];
                for (int j = i + 1; j < size; j++)
                {
                    s -= matrix[i, j] * vec[j];
                }
                vec[i] = s;
            }
            return det;
        }

        //LU Matrix Inverse
        public static RMatrix LUInverse(RMatrix matrix)
        {
            int nRows = matrix.GetnRows;
            RMatrix u = matrix.IdentityMatrix();
            LUDecompose(matrix);
            RVector uv = new RVector(nRows);
            for (int i = 0; i < nRows; i++)
            {
                uv = u.GetRowVector(i);
                LUSubstitute(matrix, uv);
                u.ReplaceRow(uv, i);
            }
            RMatrix inverse = u.GetTranspose();
            return inverse;
        }

        // Gauss-Jacobi method
        public static RVector GaussJacobi(RMatrix A, RVector b, int MaxIterations, double tolerance)
        {
            int bSize = b.GetVectorSize;
            RVector x = new RVector(bSize);
            for (int nIteration = 0; nIteration < MaxIterations; nIteration++)
            {
                RVector xOld = x.Clone();
                for (int i = 0; i < bSize; i++)
                {
                    double entry = b[i];
                    double diagonal = A[i, i];
                    if (Math.Abs(diagonal) < epsilon)
                        throw new Exception("Diagonal element is too small!");
                    for (int j = 0; j < bSize; j++)
                    {
                        if (j != i)
                        {
                            entry -= A[i, j] * xOld[j];
                        }
                    }
                    x[i] = entry / diagonal;
                }
                RVector dx = x - xOld;
                if (dx.GetNorm() < tolerance)
                {
                    return x;
                }
            }
            return x;
        }

        // Gauss-Seidel method
        public static RVector GaussSeidel(RMatrix A, RVector b, int MaxIterations, double tolerance)
        {
            int size = b.GetVectorSize;
            RVector x = new RVector(size);
            for (int nIteration = 0; nIteration < MaxIterations; nIteration++)
            {
                RVector xOld = x.Clone();
                for (int i = 0; i < size; i++)
                {
                    double entry = b[i];
                    double diagonal = A[i, i];
                    if (Math.Abs(diagonal) < epsilon)
                        throw new Exception("Diagonal element is too small!");
                    for (int j = 0; j < i; j++)
                    {
                        entry -= A[i, j] * x[j];
                    }
                    for (int j = i + 1; j < size; j++)
                    {
                        entry -= A[i, j] * xOld[j];
                    }
                    x[i] = entry / diagonal;
                }
                RVector dx = x - xOld;
                if (dx.GetNorm() < tolerance)
                {
                    return x;
                }
            }
            return x;
        }
        #endregion

        #region Binomial Distribution from Chapter 10

        private static Random rand = new Random();

        public static double GammaLn(double x)
        {
            if (x <= 0) throw
            new Exception("Input value must be > 0");

            double[] coef = new double[14]
            {57.1562356658629235,
            -59.5979603554754912,
            14.1360979747417471,
            -0.491913816097620199,
             0.339946499848118887E-4,
             0.465236289270485756E-4,
            -0.983744753048795646E-4,
             0.158088703224912494E-3,
            -0.210264441724104883E-3,
             0.217439618115212643E-3,
            -0.164318106536763890E-3,
             0.844182239838527433E-4,
            -0.261908384015814087E-4,
             0.368991826595316234E-5};

            double denominator = x;
            double series = 0.999999999999997092;
            double temp = x + 5.24218750000000000;
            temp = (x + 0.5) * Math.Log(temp) - temp;
            for (int j = 0; j < 14; j++)
                series += coef[j] / ++denominator;
            return (temp + Math.Log(2.5066282746310005 * series / x));
        }

        public static double Gamma(double x)
        {
            if (x <= 0) throw
                new Exception("Input value must be > 0");
            return Math.Exp(GammaLn(x));
        }

        public static double BinomialPDF(int x, int n, double p)
        {
            return Gamma(n + 1) * Math.Pow(p, x) * Math.Pow(1 - p, n - x) / Gamma(x + 1) / Gamma(n - x + 1);
        }

        public static double[] BinomialPDF(int[] x, int n, double p)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = BinomialPDF(x[i], n, p);
            }
            return y;
        }

        public static double NextBinomial(int n, double p)
        {
            double result = 0.0;
            for (int i = 0; i < n; i++)
            {
                if (rand.NextDouble() < p)
                {
                    result++;
                }
            }
            return result;
        }

        public static double[] NextBinomial(int n, double p, int nLen)
        {
            double[] array = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                array[i] = NextBinomial(n, p);
            }
            return array;
        }

        #endregion

        #region Histogram Utilities from Chapter 10

        public static double[] makeHistogram(double[] data, double min, double max, int nBins)
        {
            double[] Histogram = new double[nBins];
            double BinWidth = (max - min) / nBins;
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] >= min + (i) * BinWidth && data[j] < min + (i + 1) * BinWidth)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double[] makeHistogram(double[] data, int nBins)
        {
            double[] Histogram = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] == i)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double dataMax(double[] data)
        {
            double max = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                max = Math.Max(max, data[i]);
            }
            return max;
        }

        public static double dataMin(double[] data)
        {
            double min = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                min = Math.Min(min, data[i]);
            }
            return min;
        }

        #endregion

        // Material from Chapter 15 starts here

        #region Methods for Chapter 15

        public static double[] LeastSquaresBestFitLine1(double[] x, double[] y)
        {
            //Calculates equation of best-fit line using short cuts
            int n = x.Length;
            double xMean = 0.0;
            double yMean = 0.0;
            double numeratorSum = 0.0;
            double denominatorSum = 0.0;
            double bestfitYintercept = 0.0;
            double bestfitSlope = 0.0;
            double sigma = 0.0;
            double sumOfResidualsSquared = 0.0;

            //Calculates the mean values for x and y arrays
            for (int i = 0; i < n; i++)
            {
                xMean += x[i] / n;
                yMean += y[i] / n;
            }

            //Calculates the numerator and denominator for best-fit slope
            for (int i = 0; i < n; i++)
            {
                numeratorSum += y[i] * (x[i] - xMean);
                denominatorSum += x[i] * (x[i] - xMean);
            }

            //Calculate the best-fit slope and y-intercept
            bestfitSlope = numeratorSum / denominatorSum;
            bestfitYintercept = yMean - xMean * bestfitSlope;

            //Calculate the best-fit standard deviation
            for (int i = 0; i < n; i++)
            {
                sumOfResidualsSquared += (y[i] - bestfitYintercept - bestfitSlope * x[i]) * (y[i] - bestfitYintercept - bestfitSlope * x[i]);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - 2));

            return new double[] { bestfitYintercept, bestfitSlope, sigma };
        }

        public static double[] LeastSquaresBestFitLine2(double[] x, double[] y)
        {
            //Calculates equation of best-fit line using sums
            int n = x.Length;
            double xSum = 0.0;
            double ySum = 0.0;
            double xySum = 0.0;
            double xSqrSum = 0.0;
            double denominator = 0.0;
            double bNumerator = 0.0;
            double mNumerator = 0.0;
            double bestfitYintercept = 0.0;
            double bestfitSlope = 0.0;
            double sigma = 0.0;
            double sumOfResidualsSquared = 0.0;

            //calculate sums
            for (int i = 0; i < n; i++)
            {
                xSum += x[i];
                ySum += y[i];
                xySum += x[i] * y[i];
                xSqrSum += x[i] * x[i];
            }

            denominator = n * xSqrSum - xSum * xSum;
            bNumerator = xSqrSum * ySum - xSum * xySum;
            mNumerator = n * xySum - xSum * ySum;

            //calculate best-fit y-intercept
            bestfitYintercept = bNumerator / denominator;

            //calculate best-fit slope
            bestfitSlope = mNumerator / denominator;

            //calculate best-fit standard deviation
            for (int i = 0; i < n; i++)
            {
                sumOfResidualsSquared += (y[i] - bestfitYintercept - bestfitSlope * x[i]) * (y[i] - bestfitYintercept - bestfitSlope * x[i]);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - 2));

            return new double[] { bestfitYintercept, bestfitSlope, sigma };
        }

        public static double[] LeastSquaresWeightedBestFitLine1(double[] x, double[] y, double[] w)
        {
            //Calculates equation of best-fit line using short cuts
            int n = x.Length;
            double wxMean = 0.0;
            double wyMean = 0.0;
            double wSum = 0.0;
            double wnumeratorSum = 0.0;
            double wdenominatorSum = 0.0;
            double bestfitYintercept = 0.0;
            double bestfitSlope = 0.0;
            double sigma = 0.0;
            double sumOfResidualsSquared = 0.0;

            //Calculates the sum of the weights w[i]
            for (int i = 0; i < n; i++)
            {
                wSum += w[i];
            }

            //Calculates the mean values for x and y arrays
            for (int i = 0; i < n; i++)
            {
                wxMean += w[i] * x[i] / wSum;
                wyMean += w[i] * y[i] / wSum;
            }

            //Calculates the numerator and denominator for best-fit slope
            for (int i = 0; i < n; i++)
            {
                wnumeratorSum += w[i] * y[i] * (x[i] - wxMean);
                wdenominatorSum += w[i] * x[i] * (x[i] - wxMean);
            }

            //Calculate the best-fit slope and y-intercept
            bestfitSlope = wnumeratorSum / wdenominatorSum;
            bestfitYintercept = wyMean - wxMean * bestfitSlope;

            //Calculate the best-fit standard deviation
            for (int i = 0; i < n; i++)
            {
                sumOfResidualsSquared += w[i]*(y[i] - bestfitYintercept - bestfitSlope * x[i]) * (y[i] - bestfitYintercept - bestfitSlope * x[i]);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - 2));

            return new double[] { bestfitYintercept, bestfitSlope, sigma };
        }

        public static double[] LeastSquaresWeightedBestFitLine2(double[] x, double[] y, double[] w)
        {
            //Calculates equation of best-fit line using sums
            int n = x.Length;
            double wSum = 0.0;
            double wxSum = 0.0;
            double wySum = 0.0;
            double wxySum = 0.0;
            double wxSqrSum = 0.0;
            double denominator = 0.0;
            double bNumerator = 0.0;
            double mNumerator = 0.0;
            double bestfitYintercept = 0.0;
            double bestfitSlope = 0.0;
            double sigma = 0.0;
            double sumOfResidualsSquared = 0.0;

            //calculate sums
            for (int i = 0; i < n; i++)
            {
                wSum += w[i];
                wxSum += w[i]*x[i];
                wySum += w[i]*y[i];
                wxySum += w[i]*x[i] * y[i];
                wxSqrSum += w[i]* x[i] * x[i];
            }

            denominator = wSum * wxSqrSum - wxSum * wxSum;
            bNumerator = wxSqrSum * wySum - wxSum * wxySum;
            mNumerator = wSum * wxySum - wxSum * wySum;

            //calculate best-fit y-intercept
            bestfitYintercept = bNumerator / denominator;

            //calculate best-fit slope
            bestfitSlope = mNumerator / denominator;

            //calculate best-fit standard deviation
            for (int i = 0; i < n; i++)
            {
                sumOfResidualsSquared += w[i] * (y[i] - bestfitYintercept - bestfitSlope * x[i]) * (y[i] - bestfitYintercept - bestfitSlope * x[i]);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - 2));            

            return new double[] { bestfitYintercept, bestfitSlope, sigma };
        }

        public delegate double ModelFunction(double x);

        public static RVector LinearRegression(double[] x, double[] y, ModelFunction[] f, out double sigma)
        {
            //m = number of data points
            int m = f.Length;
            RMatrix Fmatrix = new RMatrix(m, m);
            RVector Bvector = new RVector(m);
            // n = number of linear terms in the regression equation
            int n = x.Length;

            //Calculate the B vector entries
            for (int k = 0; k < m; k++)
            {
                Bvector[k] = 0.0;
                for (int i = 0; i < n; i++)
                {
                    Bvector[k] += f[k](x[i]) * y[i];
                }
            }

            //Calculate the F matrix entries
            for (int j = 0; j < m; j++)
            {
                for (int k = 0; k < m; k++)
                {
                    Fmatrix[j, k] = 0.0;
                    for (int i = 0; i < n; i++)
                    {
                        Fmatrix[j, k] += f[j](x[i]) * f[k](x[i]);
                    }
                }
            }

            // FA = B so A = F^{-1}B
            RVector Avector = GaussJordan(Fmatrix, Bvector);

            // Calculate the standard deviation to estimate error
            double sumOfResidualsSquared = 0.0;
            for (int i = 0; i < n; i++)
            {
                double sum = 0.0;
                for (int j = 0; j < m; j++)
                {
                    sum += Avector[j] * f[j](x[i]);
                }
                sumOfResidualsSquared += (y[i] - sum) * (y[i] - sum);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - m));
            return Avector;
        }

        public static RVector PolynomialFit(double[] x, double[] y, int m, out double sigma)
        {
            //m = number of data points which in this case
            //for polynomials = order or degree of polynomial P_m(x)
            m++; //minor adjust
            RMatrix Fmatrix = new RMatrix(m, m);
            RVector Bvector = new RVector(m);
            // n = number of linear terms in the regression equation
            int n = x.Length;

            //Calculate the B vector entries
            for (int k = 0; k < m; k++)
            {
                Bvector[k] = 0.0;
                for (int i = 0; i < n; i++)
                {
                    Bvector[k] += Math.Pow(x[i], k) * y[i];
                }
            }

            //Calculate the F matrix entries
            for (int j = 0; j < m; j++)
            {
                for (int k = 0; k < m; k++)
                {
                    Fmatrix[j, k] = 0.0;
                    for (int i = 0; i < n; i++)
                    {
                        Fmatrix[j, k] += Math.Pow(x[i], j + k);
                    }
                }
            }

            // FA = B so A = F^{-1}B
            RVector Avector = GaussJordan(Fmatrix, Bvector);

            // Calculate the standard deviation to estimate error
            double sumOfResidualsSquared = 0.0;
            for (int i = 0; i < n; i++)
            {
                double sum = 0.0;
                for (int j = 0; j < m; j++)
                {
                    sum += Avector[j] * Math.Pow(x[i], j);
                }
                sumOfResidualsSquared += (y[i] - sum) * (y[i] - sum);
            }
            sigma = Math.Sqrt(sumOfResidualsSquared / (n - m));
            return Avector;
        }
        
        #endregion

        #region Tests for Chapter 15

        private static void TestLeastSquaresBestFitLine()
        {
            double[] xdata = new double[] { 2.0, 3.0, 4.0, 5.0, 6.0, 7.0 };
            double[] ydata = new double[] { 2.9, 4.7, 6.3, 8.4, 10.5, 12.5 };
            double[] results1 = LeastSquaresBestFitLine1(xdata, ydata);
            double[] results2 = LeastSquaresBestFitLine2(xdata, ydata);

            Console.WriteLine("Testing straight line best-fit method \n");
            Console.WriteLine("Results1: (compact version)");
            Console.WriteLine("Best-fit y-intercept = {0}", results1[0]);
            Console.WriteLine("Best-fit slope = {0}", results1[1]);
            Console.WriteLine("Best-fit sigma = {0}", results1[2]);
            Console.WriteLine("Equation for best-fit line: y = mx + b = {0}x + {1}\n", results1[1], results1[0]);

            Console.WriteLine("Results2: (traditional version)");
            Console.WriteLine("Best-fit y-intercept = {0}", results2[0]);
            Console.WriteLine("Best-fit slope = {0}", results2[1]);
            Console.WriteLine("Best-fit sigma = {0}", results2[2]);
            Console.WriteLine("Equation for best-fit line: y = mx + b = {0}x + {1}\n", results2[1], results2[0]);
            Console.ReadLine();
        }

        private static void TestLeastSquaresWeightedBestFitLine()
        {
            double[] xdata = new double[] { 0.0, 2.0, 4.0, 6.0 };
            double[] ydata = new double[] { 10.0, 15.0, 18.0, 25.0 };
            double[] weights = new double[] { 1.0, 5.0, 10.0, 1.0 };
            double[] results1 = LeastSquaresWeightedBestFitLine1(xdata, ydata, weights);
            double[] results2 = LeastSquaresWeightedBestFitLine2(xdata, ydata, weights);

            Console.WriteLine("Testing weighted straight line best-fit method\n");
            Console.WriteLine("Results1: (compact version)");
            Console.WriteLine("Best-fit y-intercept = {0}", results1[0]);
            Console.WriteLine("Best-fit slope = {0}", results1[1]);
            Console.WriteLine("Best-fit sigma = {0}", results1[2]);
            Console.WriteLine("Equation for weighted best-fit line: y = mx + b = {0}x + {1}\n", results1[1], results1[0]);

            Console.WriteLine("Results2: (traditional version)");
            Console.WriteLine("Best-fit y-intercept = {0}", results2[0]);
            Console.WriteLine("Best-fit slope = {0}", results2[1]);
            Console.WriteLine("Best-fit sigma = {0}", results2[2]);
            Console.WriteLine("Equation for weighted best-fit line: y = mx + b = {0}x + {1}\n", results2[1], results2[0]);
            Console.ReadLine();
        }

        private static double f0(double x)
        {
            return 1.0;
        }

        private static double f1(double x)
        {
            return x;
        }

        private static double f2(double x)
        {
            return x * x;
        }

        private static double f3(double x)
        {
            return x * x * x;
        }

        private static void TestLinearRegression()
        {
            Console.WriteLine("Testing Linear Regression Fit for a set of basis");
            Console.WriteLine("functions consisting of a polynomial of order 3");
            double[] x = new double[] { 0, 1, 2, 3, 4, 5 };
            double[] y = new double[] { 4, 2, 8, 8, 6, 4 };
            double sigma = 0.0;
            ModelFunction[] f = new ModelFunction[] { f0, f1, f2, f3 };
            RVector results = LinearRegression(x, y, f, out sigma);
            Console.WriteLine("Order of polynomial m = 3: A0 + A1*x + A2*x^2 + A3*x^3 where");
            Console.WriteLine("A0 = {0}", results[0]);
            Console.WriteLine("A1 = {0}", results[1]);
            Console.WriteLine("A2 = {0}", results[2]);
            Console.WriteLine("A3 = {0}", results[3]);
            Console.WriteLine("Standard deviation = {0}\n\n", sigma);
            Console.ReadLine();
        }

        private static void TestPolynomialFit()
        {
            Console.WriteLine("Testing Polynomial Fit");
            Console.WriteLine("using an actual polynomial of order 3");
            double[] x = new double[] { 0, 1, 2, 3, 4, 5 };
            double[] y = new double[] { 4, 2, 8, 8, 6, 4 };
            int polynomialOrder = 3;
            double sigma = 0.0;
            RVector results = PolynomialFit(x, y, polynomialOrder, out sigma);
            Console.WriteLine("Order of polynomial m = 3: A0 + A1*x + A2*x^2 + A3*x^3 where");
            Console.WriteLine("A0 = {0}", results[0]);
            Console.WriteLine("A1 = {0}", results[1]);
            Console.WriteLine("A2 = {0}", results[2]);
            Console.WriteLine("A3 = {0}", results[3]);
            Console.WriteLine("Standard deviation = {0}\n\n", sigma);
            Console.ReadLine();
        }
        
        private static void TestExponentialFit()
        {
            Console.WriteLine("Testing Exponential Fit");
            Console.WriteLine("  Original equation: y  = a e^{cx}");
            Console.WriteLine("Linearized equation: ln(y) = cx + ln(a)");
            Console.WriteLine("which is in the form: Y = mx + b for a straight line\n");
            double[] x = new double[] { 1, 2, 3, 4, 5 };
            double[] y = new double[] { 3.5, 6.2, 9.5, 15.3, 20.4 };
            double[] logy = new double[] { 1.25276, 1.82455, 2.25129, 2.72785, 3.01553};

            double[] results = LeastSquaresBestFitLine1(x, logy);
            Console.WriteLine("Results without weights, just (x,ln(y)): ");
            Console.WriteLine("Best-fit y-intercept b = ln(a) = {0}", results[0]);
            Console.WriteLine("Best-fit slope = m = c ={0}", results[1]);
            Console.WriteLine("Best-fit sigma = {0}", results[2]);
            Console.WriteLine("Equation for weighted best-fit exponential:");
            Console.WriteLine("y = a e^(cx) = {0}exp({1}x)\n", Math.Exp(results[0]), results[1]);

            double[] resultsWt = LeastSquaresWeightedBestFitLine1(x, logy, y);
            Console.WriteLine("Results with weights w = y. Calc (x,ln(y),w)):");
            Console.WriteLine("Best-fit y-intercept b = ln(a)= {0}", resultsWt[0]);
            Console.WriteLine("Best-fit slope = m = c = {0}", resultsWt[1]);
            Console.WriteLine("Best-fit sigma = {0}", resultsWt[2]);
            Console.WriteLine("Equation for weighted best-fit exponential:");
            Console.WriteLine("y = a e^(cx) = {0}exp({1}x)\n", Math.Exp(resultsWt[0]), resultsWt[1]);
            Console.ReadLine();
        }

        private static void TestChiSquareStatistic()
        {
            int nBins = 20;     //number of bins for histogram
            int nPoints = 2000; //number of data points

            //Create an array of 2000 random data points 
            //that supposedly follow a binomial distribution
            double[] RandomData = NextBinomial(20, 0.5, nPoints);

            //Distribute the 2000 random data points into a
            //histogram consisting of 20 bins
            double[] RandomDistributionValues = makeHistogram(RandomData, nBins);

            //Setup arrays to hold data
            double[] xdata = new double[nBins];
            double[] ydata = new double[nBins];

            double[] Observed = new double[nBins];
            double[] Expected = new double[nBins];

            double[] ydistribution = new double[nBins];

            //Assign random data and binomial distribution data
            //to their proper arrays
            for (int i = 0; i < nBins; i++)
            {
                xdata[i] = i;
                ydata[i] = (double)RandomDistributionValues[i];
                ydistribution[i] = BinomialPDF(i, 20, 0.5);
            }

            //Calculate the normalization factor
            double normalizationFactor = dataMax(ydata) / dataMax(ydistribution);

            //Assign observed and expected data arrays and adjust
            //the expected data array by the normalization factor
            for (int i = 0; i < nBins; i++)
            {
                Observed[i] = ydata[i];
                Expected[i] = ydistribution[i] * normalizationFactor;
            }

            //Calculate the number of degrees of freedom
            int NDegreesOfFreedom = nBins - 1;

            //Calculate chi-square value
            double chi_square = 0.0;
            for (int i = 0; i < nBins; i++)
            {
                chi_square += ((Observed[i] - Expected[i]) * (Observed[i] - Expected[i])) / (Expected[i]);
            }

            Console.WriteLine("\nChi-Square Statistic Test\n");
            Console.WriteLine("A random sample of 2000 points supposedly following a Binomial Distribution");
            Console.WriteLine("is compared against a set of calculated binomial distribution values");
            Console.WriteLine("using the chi-square statistic");
            Console.WriteLine("Results:"); 
            Console.WriteLine("        chi-square = {0}", chi_square);
            Console.WriteLine("Reduced chi-square = {0}", chi_square/ NDegreesOfFreedom);
            Console.ReadLine();
        }
        
        #endregion            

        static void Main(string[] args)
        {
            TestLeastSquaresBestFitLine();
            TestLeastSquaresWeightedBestFitLine();
            TestLinearRegression();
            TestPolynomialFit();
            TestExponentialFit();
            TestChiSquareStatistic();
        }            
    }
}
