﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Chapter17_CodeExamples
{
    class Program
    {

        #region Parabolic (Heat) PDE by Finite Differences Method
        static double functionX(double x)
        { return Math.Sin(Math.PI * x); }

        static void Solving_Heat_PDE_byFiniteDiff()
        {
            /*
             *  Solution of Heat PDE by Backward-Difference Algorithm  
             *
             *  Goal: To approximate the solution to the parabolic 
             *  PDE subject to the boundary conditions
             *  u(0,t) = u(l,t) = 0, 0 < t < T = max t,
             *  and the initial conditions u(x,0) = F(x), 0 <= x <= l:
             *
             *  Input: endpoint l; maximum time T; constant alpha; integers m, N.
             *
             *  Output: approximations W(I,J) to u(x(I),t(J)) for each
             *  I = 1, ..., m-1 and J = 1, ..., N.
             */

            double[] W, L, U, Z, ActualSolution;
            double FT, FX, alpha, H, K, lambda, T, X;
            int N, M, M1, M2, N1, I1, I, J;

            W = new double[25]; L = new double[25];
            U = new double[25]; Z = new double[25];
            ActualSolution = new double[25];

            FX = 1;     //These input parameters values 
            FT = 0.5;   //are set by the user and may 
            alpha = 1;  //be varied accordingly
            M = 10;
            N = 50;

            //Initialize variables with input parameter values
            M1 = M - 1;
            M2 = M - 2;
            N1 = N - 1;
            H = FX / M;
            K = FT / N;
            lambda = alpha * alpha * K / (H * H);
            for (I = 1; I <= M1; I++) W[I - 1] = functionX(I * H);
            /* Solve the tridiagonal linear system */
            L[0] = 1.0 + 2.0 * lambda;
            U[0] = -lambda / L[0];
            for (I = 2; I <= M2; I++)
            {
                L[I - 1] = 1.0 + 2.0 * lambda + lambda * U[I - 2];
                U[I - 1] = -lambda / L[I - 1];
            }
            L[M1 - 1] = 1.0 + 2.0 * lambda + lambda * U[M2 - 1];
            for (J = 1; J <= N; J++)
            {
                T = J * K;
                Z[0] = W[0] / L[0];
                for (I = 2; I <= M1; I++)
                    Z[I - 1] = (W[I - 1] + lambda * Z[I - 2]) / L[I - 1];
                W[M1 - 1] = Z[M1 - 1];
                for (I1 = 1; I1 <= M2; I1++)
                {
                    I = M2 - I1 + 1;
                    W[I - 1] = Z[I - 1] - U[I - 1] * W[I];
                }
            }
            Console.WriteLine("Solution of a Parabolic (Heat) Partial Differential Equation");
            Console.WriteLine("using the backward-difference algorithm and user defined parameters\n");

            Console.WriteLine("I\tX(I)\t\tW(X(I)," + FT + ")\tActual Solution\n");
            for (I = 1; I <= M1; I++)
            {
                X = I * H;
                ActualSolution[I] = Math.Exp(-Math.PI * Math.PI * 0.5) * Math.Sin(Math.PI * X);
                Console.WriteLine(I.ToString() + "\t" + X.ToString("0.000000000") + "\t" + W[I - 1].ToString("0.000000000") + "\t" + ActualSolution[I].ToString("0.000000000"));
            }
            Console.ReadLine();
        }
        #endregion

        #region Parabolic (Heat) PDE by Crank-Nicolson Method

        static void Solving_Heat_PDE_by_CrankNicolson_Algorithm()
        {        
           /*
            *   Solution to Heat PDE by Crank-Nicolson Algorithm 
            *
            *   Goal: To approximate the solution of the parabolic 
            *   PDE subject to the boundary conditions
            *   u(0,t) = u(l,t) = 0, 0 < t < T = max t
            *   and the initial conditions u(x,0) = F(x), 0 <= x <= l:         
            *   Input: endpoint l; maximum time T; constant alpha; integers m,N:
            *   Output: approximations W(I,J) to u(x(I),t(J)) for each
            *   I = 1,..., m-1 and J = 1,..., N.
            */

            double[] V, L, U, ActualSolution, Z;
            double FT, FX, alpha, H, K, lambda, T, X;
            int N, M, M1, M2, I1, I, J;

            V = new double[25]; L = new double[25];
            U = new double[25]; Z = new double[25];
            ActualSolution = new double[25];

            FX = 1;     //These input parameters values 
            FT = 0.5;   //are set by the user and may 
            alpha = 1;  //be varied accordingly
            M = 10; 
            N = 50;

            //Initialize variables with input parameter values
            M1 = M - 1;
            M2 = M - 2;
            H = FX / M;
            K = FT / N;
            lambda = alpha * alpha * K / (H * H);
            V[M - 1] = 0.0;
            for (I = 1; I <= M1; I++) V[I - 1] = functionX(I * H);

            /* Solve the tridiagonal linear system */
            L[0] = 1.0 + lambda;
            U[0] = -lambda / (2.0 * L[0]);
            for (I = 2; I <= M2; I++)
            {
                L[I - 1] = 1.0 + lambda + lambda * U[I - 2] / 2.0;
                U[I - 1] = -lambda / (2.0 * L[I - 1]);
            }
            L[M1 - 1] = 1.0 + lambda + 0.5 * lambda * U[M2 - 1];
            for (J = 1; J <= N; J++)
            {
                T = J * K;
                Z[0] = ((1.0 - lambda) * V[0] + lambda * V[1] / 2.0) / L[0];
                for (I = 2; I <= M1; I++)
                    Z[I - 1] = ((1.0 - lambda) * V[I - 1] + 0.5 * lambda * (V[I] + V[I - 2] + Z[I - 2])) / L[I - 1];
                V[M1 - 1] = Z[M1 - 1];
                for (I1 = 1; I1 <= M2; I1++)
                {
                    I = M2 - I1 + 1;
                    V[I - 1] = Z[I - 1] - U[I - 1] * V[I];
                }
            }
            Console.WriteLine("Solution of a Parabolic (Heat) Partial Differential Equation");
            Console.WriteLine("using the Crank-Nicolson Method and user defined parameters\n");
            Console.WriteLine("I\tX(I)\t\tW(X(I)," + FT + ")\tActual Solution\n");
            for (I = 1; I <= M1; I++)
            {
                X = I * H;
                ActualSolution[I] = Math.Exp(-Math.PI * Math.PI * 0.5) * Math.Sin(Math.PI * X);
                Console.WriteLine(I.ToString() + "\t" + X.ToString("0.000000000") + "\t" + V[I - 1].ToString("0.000000000") + "\t" + ActualSolution[I].ToString("0.000000000"));
            }
            Console.ReadLine();
        }

        #endregion

        #region Hyperbolic (Wave) PDE by Finite Differences Method

        static double fx(double x)
        { return Math.Sin(Math.PI * x); }

        static double gx(double x)
        { return 0.0; }

        static void Solving_Wave_PDE_byBackward_Diff()
        {
            /*
             *  Solution to Wave PDE by finite difference method 
             *  Goal: To approximate the solution of the hyperbolic 
             *  PDE subject to the boundary conditions
             *  u(0,t) = u(l,t) = 0, 0 < t < T = max t
             *  and the initial conditions
             *  u(x,0) = F(x) and Du(x,0)/Dt = G(x), 0 <= x <= l:
             *  Input:  endpoint l; maximum time T; constant alpha; integers m,N.
             *  Output: approximations W(I,J) to u(x(I),t(J)) for each 
             *  I = 0, ..., m and J=0,...,N.
            */

            double[,] W; 
            double[] ActualSolution;
            double FT, FX, alpha, H, K, lambda, X;
            int M1, M2, N1, N2, I, J;
            W = new double[51, 51];
            ActualSolution = new double[51];

            FX = 1;     //These input parameters values
            FT = 1;     //are set by the user and may 
            alpha = 2;  //be varied accordingly
            int M = 10; 
            int N = 20; 
            
            //Initialize variables with input parameter values
            M1 = M + 1;
            M2 = M - 1;
            N1 = N + 1;
            N2 = N - 1;
            H = FX / M;
            K = FT / N;
            lambda = alpha * K / H;
            for (J = 2; J <= N1; J++)
            {
                W[0, J - 1] = 0.0;
                W[M1 - 1, J - 1] = 0.0;
            }
            W[0, 0] = fx(0.0);
            W[M1 - 1, 0] = fx(FX);
            for (I = 2; I <= M; I++)
            {
                W[I - 1, 0] = fx(H * (I - 1.0));
                W[I - 1, 1] = (1.0 - lambda * lambda) * fx(H * (I - 1.0)) + lambda * lambda * (fx(I * H) + fx(H * (I - 2.0))) / 2.0 + K * gx(H * (I - 1.0));
            }
            for (J = 2; J <= N; J++)
            {
                for (I = 2; I <= M; I++)
                {
                    W[I - 1, J] = 2.0 * (1.0 - lambda * lambda) * W[I - 1, J - 1] + lambda * lambda * (W[I, J - 1] + W[I - 2, J - 1]) - W[I - 1, J - 2];
                }
            }
            Console.WriteLine("Solution of a Hyperbolic (Wave) Partial Differential Equation");
            Console.WriteLine("using the Finite-Difference Method and user defined parameters\n");
            Console.WriteLine("I\tX(I)\t\tW(X(I)," + FT + ")\tActual Solution\n");
            for (I = 1; I <= M1; I++)
            {
                X = (I - 1) * H;
                //ActualSolution[I] = Math.Sin(Math.PI * X)*Math.Cos(2.0*Math.PI*1.0);
                ActualSolution[I] = Math.Sin(Math.PI * X) * Math.Cos(2.0 * Math.PI * 1.0 );
                Console.WriteLine(I.ToString() + "\t" + X.ToString("0.000000000") + "\t" + W[I - 1, N1 - 1].ToString("0.000000000") + "\t" + ActualSolution[I].ToString("0.000000000"));
            }
            Console.ReadLine();
        }
        #endregion

        #region Elliptic (Poisson) PDE by Finite Differences Method

        static double funcx(double x, double y)
        { return x * Math.Exp(y); }

        static double gx(double x, double y)
        { return x * Math.Exp(y); }

        static void Solving_Poisson_PDE_byFiniteDiff()
        {
            /*  Solution to the Poisson PDE by finite difference method 
             *  Goal: To approximate the solution of the Elliptic PDE
             *  DEL(u) = F(x,y), a <= x <= b, c <= y <= d,
             *  subject to the boundary conditions 
             *  u(x,y) = G(x,y)
             *  if x = a or x = b for c <= y <= d,
             *  if y = c or y = d for a <= x <= b 
             *  Input: endpoints a, b, c, d; integers m, n; tolerance TOL;
             *         maximum number of iterations M
             *  Output: approximations W(I,J) to u(X(I),Y(J)) for each 
             *  I = 1,..., n-1 and J=1,..., m-1 or a message that the
             *  maximum number of iterations was exceeded.
             */
            double[,] W, ActualSolution;
            double[] X, Y;
            double TOL, A, B, C, D, H, K, lambda, mu, Z, E;
            int I, J, M, N, NN, M1, M2, N1, N2, L, LL;
            bool OK;
            W = new double[26, 26];
            ActualSolution = new double[26, 26];
            X = new double[26];
            Y = new double[26];            

            A = 0.0;    //These input parameters values
            B = 2.0;    //are set by the user and may
            C = 0.0;    //be varied accordingly
            D = 1.0;
            TOL = 1.0E-10;
            NN = 100;
            M = 5;
            N = 6;

            //Initialize variables with input parameter values
            M1 = M - 1;
            M2 = M - 2;
            N1 = N - 1;
            N2 = N - 2;
            H = (B - A) / N;
            K = (D - C) / M;

            // Construct mesh points 
            for (I = 0; I <= N; I++) X[I] = A + I * H;
            for (J = 0; J <= M; J++) Y[J] = C + J * K;
            for (I = 1; I <= N1; I++)
            {
                W[I, 0] = gx(X[I], Y[0]);
                W[I, M] = gx(X[I], Y[M]);
            }
            for (J = 0; J <= M; J++)
            {
                W[0, J] = gx(X[0], Y[J]);
                W[N, J] = gx(X[N], Y[J]);
            }
            for (I = 1; I <= N1; I++)
                for (J = 1; J <= M1; J++) W[I, J] = 0.0;
            lambda = H * H / (K * K);
            mu = 2.0 * (1.0 + lambda);
            L = 1;
            OK = false;
            /* Z is a new value of W(I,J) to be used in computing
             the norm of the error E used in place of NORM */

            /* Perform Gauss-Seidel iterations */
            while ((L <= NN) && (!OK))
            {                
                Z = (-H * H * funcx(X[1], Y[M1]) + gx(A, Y[M1]) + lambda * gx(X[1], D) + lambda * W[1, M2] + W[2, M1]) / mu;
                E = Math.Abs(W[1, M1] - Z);
                W[1, M1] = Z;
                for (I = 2; I <= N2; I++)
                {
                    Z = (-H * H * funcx(X[I], Y[M1]) + lambda * gx(X[I], D) + W[I - 1, M1] + W[I + 1, M1] + lambda * W[I, M2]) / mu;
                    if (Math.Abs(W[I, M1] - Z) > E) E = Math.Abs(W[I, M1] - Z);
                    W[I, M1] = Z;
                }
                Z = (-H * H * funcx(X[N1], Y[M1]) + gx(B, Y[M1]) + lambda * gx(X[N1], D) + W[N2, M1] + lambda * W[N1, M2]) / mu;
                if (Math.Abs(W[N1, M1] - Z) > E) E = Math.Abs(W[N1, M1] - Z);
                W[N1, M1] = Z;
                for (LL = 2; LL <= M2; LL++)
                {
                    J = M2 - LL + 2;
                    Z = (-H * H * funcx(X[1], Y[J]) + gx(A, Y[J]) + lambda * W[1, J + 1] + lambda * W[1, J - 1] + W[2, J]) / mu;
                    if (Math.Abs(W[1, J] - Z) > E) E = Math.Abs(W[1, J] - Z);
                    W[1, J] = Z;
                    for (I = 2; I <= N2; I++)
                    {
                        Z = (-H * H * funcx(X[I], Y[J]) + W[I - 1, J] + lambda * W[I, J + 1] + lambda * W[I, J - 1] + W[I + 1, J]) / mu;
                        if (Math.Abs(W[I, J] - Z) > E) E = Math.Abs(W[I, J] - Z);
                        W[I, J] = Z;
                    }
                    Z = (-H * H * funcx(X[N1], Y[J]) + gx(B, Y[J]) + W[N2, J] + lambda * W[N1, J + 1] + lambda * W[N1, J - 1]) / mu;
                    if (Math.Abs(W[N1, J] - Z) > E) E = Math.Abs(W[N1, J] - Z);
                    W[N1, J] = Z;
                }
                Z = (-H * H * funcx(X[1], Y[1]) + lambda * gx(X[1], C) + gx(A, Y[1]) + lambda * W[1, 2] + W[2, 1]) / mu;
                if (Math.Abs(W[1, 1] - Z) > E) E = Math.Abs(W[1, 1] - Z);
                W[1, 1] = Z;
                for (I = 2; I <= N2; I++)
                {
                    Z = (-H * H * funcx(X[I], Y[1]) + lambda * gx(X[I], C) + W[I + 1, 1] + W[I - 1, 1] + lambda * W[I, 2]) / mu;
                    if (Math.Abs(W[I, 1] - Z) > E) E = Math.Abs(W[I, 1] - Z);
                    W[I, 1] = Z;
                }
                Z = (-H * H * funcx(X[N1], Y[1]) + lambda * gx(X[N1], C) + gx(B, Y[1]) + W[N2, 1] + lambda * W[N1, 2]) / mu;
                if (Math.Abs(W[N1, 1] - Z) > E) E = Math.Abs(W[N1, 1] - Z);
                W[N1, 1] = Z;
                if (E <= TOL)
                {
                    Console.WriteLine("Solution of a Elliptic (Poisson) Partial Differential Equation");
                    Console.WriteLine("using the Finite-Difference Method and user defined parameters\n");
                    Console.WriteLine(" I\tJ\tX(I)\t\t\tY(I)\tW(I,J)\t\tActualSolution(I,J)\n");
                    for (I = 1; I <= N1; I++)
                    {
                        for (J = 1; J <= M1; J++)
                        {
                            ActualSolution[I, J] = X[I] * Math.Exp(Y[J]);
                            Console.WriteLine(I + "\t" + J + "\t" + X[I].ToString("0.000000000") + "\t" + Y[J].ToString("0.000000000") + "\t" + W[I, J].ToString("0.000000000") + "\t" + ActualSolution[I, J].ToString("0.000000000"));
                        }
                        Console.WriteLine("Convergence occurred on iteration number " + L + "\n");
                    }
                    OK = true;
                }
                else
                    L++;
            }
            if (L > NN)
                Console.WriteLine("Method fails after iteration number " + NN + "\n");
            Console.ReadLine();
        }
        #endregion

        static void Main(string[] args)
        {
            Solving_Heat_PDE_byFiniteDiff();
            Solving_Heat_PDE_by_CrankNicolson_Algorithm();
            Solving_Wave_PDE_byBackward_Diff();
            Solving_Poisson_PDE_byFiniteDiff();
        }
    }
}
