﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.IO;

namespace Chapter18_CodeExamples
{
    class Program
    {
        #region Steepest Descent Method

        public static void steepestDescent(double[] x, double alpha, double tolerance)
        {
            int n = x.Length; //Size of input array
            double h = 1e-6;  //Tolerance factor
            double g0 = g(x); //Initial estimate of result

            //Calculate initial gradient
            double[] fi = new double[n];
            fi = GradG(x, h);

            //Calculate initial norm
            double DelG = 0;
            for (int i = 0; i < n; ++i)
                DelG += fi[i] * fi[i];
            DelG = Math.Sqrt(DelG);

            double b = alpha / DelG;

            //Iterate until value is <= tolerance limit
            while (DelG > tolerance)
            {
                //Calculate next value
                for (int i = 0; i < n; ++i)
                    x[i] -= b * fi[i];
                h /= 2;

                //Calculate next gradient
                fi = GradG(x, h);

                //Calculate next norm 
                DelG = 0;
                for (int i = 0; i < n; ++i)
                    DelG += fi[i] * fi[i];
                DelG = Math.Sqrt(DelG);

                b = alpha / DelG;

                //Check value of given function
                //with current values
                double g1 = g(x);

                //Adjust parameter
                if (g1 > g0) alpha /= 2;
                else g0 = g1;
            }
        }

        // Provides a rough calculation of gradient g(x).
        public static double[] GradG(double[] x, double h)
        {
            int n = x.Length;
            double[] z = new double[n];
            double[] y = (double[])x.Clone();
            double g0 = g(x);
            for (int i = 0; i < n; ++i)
            {
                y[i] += h;
                z[i] = (g(y) - g0) / h;
            }
            return z;
        }

        // Method to provide function g(x).
        public static double g(double[] x)
        {
            return (x[0] - 1) * (x[0] - 1) * Math.Exp(-x[1] * x[1]) + x[1] * (x[1] + 2) * Math.Exp(-2 * x[0] * x[0]);
        }

        public static void SteepestDescentTest()
        {            
            double tolerance = 1e-6;
            double alpha = 0.1;
            double[] x = new double[2];
            x[0] = 0.1; //Initial guesses
            x[1] = -1;  //of location of minimums
            steepestDescent(x, alpha, tolerance);
            Console.WriteLine("Testing STEEPEST DESCENT method\n");
            Console.WriteLine("The minimum is at x[0] = " + x[0] + ", x[1] = " + x[1]);
            Console.ReadLine();
        }

        #endregion

        #region Revised Simplex Method

        public static void revisedSimplex(bool maximize, int n, int m, double[,] a, double epsilon, int[] basicvar)
        {
            int i, j, k, m2, p, idx = 0;
            double[] objcoeff = new double[n + 1];
            double[] varsum = new double[n + 1];
            double[] optbasicval = new double[m + 3];
            double[] aux = new double[m + 3];

            double[,] work = new double[m + 3, m + 3];
            double part, sum;
            bool infeasible, unbound, abort, outres, iterate;

            if (maximize)
                for (j = 1; j <= n; j++)
                    a[0, j] = -a[0, j];
            infeasible = false;
            unbound = false;
            m2 = m + 2;
            p = m + 2;
            outres = true;
            k = m + 1;
            for (j = 1; j <= n; j++)
            {
                objcoeff[j] = a[0, j];
                sum = 0.0;
                for (i = 1; i <= m; i++)
                    sum -= a[i, j];
                varsum[j] = sum;
            }
            sum = 0.0;
            for (i = 1; i <= m; i++)
            {
                basicvar[i] = n + i;
                optbasicval[i] = a[i, 0];
                sum -= a[i, 0];
            }
            optbasicval[k] = 0.0;
            optbasicval[m2] = sum;
            for (i = 1; i <= m2; i++)
            {
                for (j = 1; j <= m2; j++)
                    work[i, j] = 0.0;
                work[i, i] = 1.0;
            }
            iterate = true;
            do
            {
                //  phase 1
                if ((optbasicval[m2] >= -epsilon) && outres)
                {
                    outres = false;
                    p = m + 1;
                }
                part = 0.0;
                //  phase 2
                for (j = 1; j <= n; j++)
                {
                    sum = work[p, m + 1] * objcoeff[j] + work[p, m + 2] * varsum[j];
                    for (i = 1; i <= m; i++)
                        sum += work[p, i] * a[i, j];
                    if (part > sum)
                    {
                        part = sum;
                        k = j;
                    }
                }
                if (part > -epsilon)
                {
                    iterate = false;
                    if (outres)
                        infeasible = true;
                    else
                        a[0, 0] = -optbasicval[p];
                }
                else
                {
                    for (i = 1; i <= p; i++)
                    {
                        sum = work[i, m + 1] * objcoeff[k] + work[i, m + 2] * varsum[k];
                        for (j = 1; j <= m; j++)
                            sum += work[i, j] * a[j, k];
                        aux[i] = sum;
                    }
                    abort = true;
                    for (i = 1; i <= m; i++)
                        if (aux[i] >= epsilon)
                        {
                            sum = optbasicval[i] / aux[i];
                            if (abort || (sum < part))
                            {
                                part = sum;
                                idx = i;
                            }
                            abort = false;
                        }
                    if (abort)
                    {
                        unbound = true;
                        iterate = false;
                    }
                    else
                    {
                        basicvar[idx] = k;
                        sum = 1.0 / aux[idx];
                        for (j = 1; j <= m; j++)
                            work[idx, j] *= sum;
                        i = ((idx == 1) ? 2 : 1);
                        do
                        {
                            sum = aux[i];
                            optbasicval[i] -= part * sum;
                            for (j = 1; j <= m; j++)
                                work[i, j] -= work[idx, j] * sum;
                            i += ((i == idx - 1) ? 2 : 1);
                        } while (i <= p);
                        optbasicval[idx] = part;
                    }
                }
            } while (iterate);
            // return results
            basicvar[m + 1] = (infeasible ? 1 : 0);
            basicvar[m + 2] = (unbound ? 1 : 0);
            for (i = 1; i <= m; i++)
                a[i, 0] = optbasicval[i];
            if (maximize)
            {
                for (j = 1; j <= n; j++)
                    a[0, j] = -a[0, j];
                a[0, 0] = -a[0, 0];
            }
        }

        public static void revisedSimplexTest()
        {
            int n = 7;
            int m = 4;
            double eps = 1.0e-5;

            double[,] a = {{0,  5,  5,  3,  0,  0,  0, 0},
                           {3,  1,  3,  1,  1,  0,  0, 0},
                           {2, -1,  0,  3,  0,  1,  0, 0}, 
                           {4,  2, -1,  2,  0,  0,  1, 0},
                           {2,  2,  3, -1,  0,  0,  0, 1}};
            int[] basicvar = new int[m + 3];

            revisedSimplex(true, n, m, a, eps, basicvar);

            Console.WriteLine("Testing the revised simplex algorithm\n");
            if (basicvar[m + 1] > 0)
                Console.WriteLine("No feasible solution.");
            else
            {
                if (basicvar[m + 2] > 0)
                    Console.WriteLine("Objective function is unbound.");
                else
                {
                    Console.WriteLine("Optimal solution found. \n\nBasic variable   Value");
                    for (int i = 1; i <= m; i++)
                        Console.WriteLine("{0,10}\t {1,-10}", basicvar[i], a[i, 0]);
                    Console.WriteLine("\nOptimal value of the objective function = {0}", a[0, 0]);
                }
                Console.ReadLine();
            }
        }

        #endregion        

        #region Vector Functions from Chapter 3
        
        /*
         *  real zero-based n-dimensional vectors
         *   v = (v[0],v[1],v[2], ... ,v[n-1])
         */
        public struct RVector : ICloneable
        {
            private int ndim;
            private double[] vector;

            public RVector(int ndim)
            {
                this.ndim = ndim;
                this.vector = new double[ndim];
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] = 0.0;
                }
            }

            public RVector(double[] vector)
            {
                this.ndim = vector.Length;
                this.vector = vector;
            }

            public double this[int i]
            {
                get
                {
                    if (i < 0 || i > ndim)
                    {
                        throw new Exception("Requested vector index is out of range!");
                    }
                    return vector[i];
                }
                set { vector[i] = value; }
            }

            public int GetVectorSize
            {
                get { return ndim; }
            }

            public RVector Clone()
            {
                RVector v = new RVector(vector);
                v.vector = (double[])vector.Clone();
                return v;
            }

            object ICloneable.Clone()
            {
                return Clone();
            }

            public RVector SwapVectorEntries(int m, int n)
            {
                double temp = vector[m];
                vector[m] = vector[n];
                vector[n] = temp;
                return new RVector(vector);
            }

            public override string ToString()
            {
                string str = "(";
                for (int i = 0; i < ndim - 1; i++)
                {
                    str += vector[i].ToString() + ", ";
                }
                str += vector[ndim - 1].ToString() + ")";
                return str;
            }

            public override bool Equals(object obj)
            {
                return (obj is RVector) && this.Equals((RVector)obj);
            }

            public bool Equals(RVector v)
            {
                return vector == v.vector;
            }

            public override int GetHashCode()
            {
                return vector.GetHashCode();
            }

            public static bool operator ==(RVector v1, RVector v2)
            {
                return v1.Equals(v2);
            }

            public static bool operator !=(RVector v1, RVector v2)
            {
                return !v1.Equals(v2);
            }

            public static RVector operator +(RVector v)
            {
                return v;
            }

            public static RVector operator +(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] + v2[i];
                }
                return result;
            }

            public static RVector operator -(RVector v)
            {
                double[] result = new double[v.ndim];
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = -v[i];
                }
                return new RVector(result);
            }

            public static RVector operator -(RVector v1, RVector v2)
            {
                RVector result = new RVector(v1.ndim);
                for (int i = 0; i < v1.ndim; i++)
                {
                    result[i] = v1[i] - v2[i];
                }
                return result;
            }

            public static RVector operator *(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] * d;
                }
                return result;
            }

            public static RVector operator *(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = d * v[i];
                }
                return result;
            }

            public static RVector operator /(RVector v, double d)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static RVector operator /(double d, RVector v)
            {
                RVector result = new RVector(v.ndim);
                for (int i = 0; i < v.ndim; i++)
                {
                    result[i] = v[i] / d;
                }
                return result;
            }

            public static double DotProduct(RVector v1, RVector v2)
            {
                double result = 0.0;
                for (int i = 0; i < v1.ndim; i++)
                {
                    result += v1[i] * v2[i];
                }
                return result;
            }

            public double GetNorm()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return Math.Sqrt(result);
            }

            public double GetNormSquare()
            {
                double result = 0.0;
                for (int i = 0; i < ndim; i++)
                {
                    result += vector[i] * vector[i];
                }
                return result;
            }

            public void Normalize()
            {
                double norm = GetNorm();
                if (norm == 0)
                {
                    throw new Exception("Tried to normalize a vector with norm of zero!");
                }
                for (int i = 0; i < ndim; i++)
                {
                    vector[i] /= norm;
                }
            }

            public RVector GetUnitVector()
            {
                RVector result = new RVector(vector);
                result.Normalize();
                return result;
            }

            public static RVector CrossProduct(RVector v1, RVector v2)
            {
                if (v1.ndim != 3)
                {
                    throw new Exception("Vector v1 must be 3 dimensional!");
                }
                if (v2.ndim != 3)
                {
                    throw new Exception("Vector v2 must be 3 dimensional!");
                }
                RVector result = new RVector(3);
                result[0] = v1[1] * v2[2] - v1[2] * v2[1];
                result[1] = v1[2] * v2[0] - v1[0] * v2[2];
                result[2] = v1[0] * v2[1] - v1[1] * v2[0];
                return result;
            }
        }

        #endregion

        #region Random Probability Distribution Functions From Chapter 10

        #region Instantiation of random number generator object

        //Use the Visual C\# built-in random number generator
        //generator or one of your own choosing here.
        private static Random rand = new Random();

        #endregion

        #region Special Functions from Chapter 14

        public static double GammaLn(double x)
        {
            if (x <= 0) throw
            new Exception("Input value must be > 0");

            double[] coef = new double[14]
            {57.1562356658629235,
            -59.5979603554754912,
            14.1360979747417471,
            -0.491913816097620199,
             0.339946499848118887E-4,
             0.465236289270485756E-4,
            -0.983744753048795646E-4,
             0.158088703224912494E-3,
            -0.210264441724104883E-3,
             0.217439618115212643E-3,
            -0.164318106536763890E-3,
             0.844182239838527433E-4,
            -0.261908384015814087E-4,
             0.368991826595316234E-5};

            double denominator = x;
            double series = 0.999999999999997092;
            double temp = x + 5.24218750000000000;
            temp = (x + 0.5) * Math.Log(temp) - temp;
            for (int j = 0; j < 14; j++)
                series += coef[j] / ++denominator;
            return (temp + Math.Log(2.5066282746310005 * series / x));
        }

        public static double Gamma(double x)
        {
            if (x <= 0) throw
                new Exception("Input value must be > 0");
            return Math.Exp(GammaLn(x));
        }

        public static double Beta(double x, double y)
        {
            if ((x <= 0) || (y <= 0)) throw
            new Exception("Input values must both be > 0");

            return Math.Exp(GammaLn(x) + GammaLn(y) - GammaLn(x + y));
        }

        #endregion

        #region Histogram Utilities

        public static double[] makeHistogram(double[] data, double min, double max, int nBins)
        {
            double[] Histogram = new double[nBins];
            double BinWidth = (max - min) / nBins;
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] >= min + (i) * BinWidth && data[j] < min + (i + 1) * BinWidth)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double[] makeHistogram(double[] data, int nBins)
        {
            double[] Histogram = new double[nBins];
            for (int i = 0; i < nBins; i++)
            {
                int nCounts = 0;
                for (int j = 0; j < data.Length; j++)
                {
                    if (data[j] == i)
                    {
                        nCounts++;
                    }
                }
                Histogram[i] = nCounts;
            }
            return Histogram;
        }

        public static double dataMax(double[] data)
        {
            double max = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                max = Math.Max(max, data[i]);
            }
            return max;
        }

        public static double dataMin(double[] data)
        {
            double min = data[0];
            for (int i = 1; i < data.Length; i++)
            {
                min = Math.Min(min, data[i]);
            }
            return min;
        }

        #endregion

        #region Binomial Distribution

        public static double BinomialPDF(int x, int n, double p)
        {
            return Gamma(n + 1) * Math.Pow(p, x) * Math.Pow(1 - p, n - x) / Gamma(x + 1) / Gamma(n - x + 1);
        }

        public static double[] BinomialPDF(int[] x, int n, double p)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = BinomialPDF(x[i], n, p);
            }
            return y;
        }

        public static double NextBinomial(int n, double p)
        {
            double result = 0.0;
            for (int i = 0; i < n; i++)
            {
                if (rand.NextDouble() < p)
                {
                    result++;
                }
            }
            return result;
        }

        public static double[] NextBinomial(int n, double p, int nLen)
        {
            double[] array = new double[nLen];
            for (int i = 0; i < nLen; i++)
            {
                array[i] = NextBinomial(n, p);
            }
            return array;
        }

        #endregion

        #region Poisson Distribution

        public static double PoissonPDF(int x, double lambda)
        {
            double y = Math.Exp(-lambda) * Math.Pow(lambda, x);
            return y / Gamma(x + 1);
        }

        public static double[] PoissonPDF(int[] x, double lambda)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = PoissonPDF(x[i], lambda);
            }
            return y;
        }

        public static double NextPoisson(double lambda)
        {
            if (lambda < 0.0) throw new Exception("Lambda must be positive!");
            int i = 0;
            double flambda = Math.Exp(-lambda);
            double p = flambda;
            double u = rand.NextDouble();
            while (flambda <= u)
            {
                p *= (lambda / (i + 1.0));
                flambda += p;
                i++;
            }
            return i;
        }

        public static double[] NextPoisson(double lambda, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextPoisson(lambda);
            }
            return array;
        }

        #endregion

        #region Gamma Distribution

        public static double GammaPDF(double x, int r, double alpha)
        {
            return Math.Pow(alpha, r) * Math.Pow(x, r - 1) * Math.Exp(-alpha * x) / Gamma(r);
        }

        public static double[] GammaPDF(double[] x, int r, double alpha)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = GammaPDF(x[i], r, alpha);
            }
            return y;
        }

        public static double NextGamma(int r, double alpha)
        {
            if (r <= 0) throw new Exception("r must be > 0!");
            if (alpha <= 0.0) throw new Exception("alpha must be > 0!");

            double result = 0.0;
            for (int i = 0; i < r; i++)
            {
                result += -Math.Log(rand.NextDouble()) / alpha;
            }
            return result;
        }

        public static double[] NextGamma(int r, double alpha, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextGamma(r, alpha);
            }
            return array;
        }

        #endregion

        #region Beta Distribution

        public static double BetaPDF(double x, double alpha, double beta)
        {
            double b = Beta(alpha, beta);
            return Math.Pow(x, alpha - 1) * Math.Pow(1 - x, beta - 1) / b;
        }

        public static double[] BetaPDF(double[] x, double alpha, double beta)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = BetaPDF(x[i], alpha, beta);
            }
            return y;
        }

        public static double NextBeta(int a, int b)
        {
            double gamma1 = NextGamma(a, 1);
            double gamma2 = NextGamma(b, 1);
            return gamma1 / (gamma1 + gamma2);
        }

        public static double[] NextBeta(int a, int b, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextBeta(a, b);
            }
            return array;
        }

        #endregion

        #region Normal Distribution

        public static double NormalPDF(double x, double mu, double sigma)
        {
            double x1, x2;
            x1 = 1 / sigma / Math.Sqrt(2 * Math.PI);
            x2 = (x - mu) * (x - mu) / (2 * sigma * sigma);
            return x1 * Math.Exp(-x2);
        }

        public static double[] NormalPDF(double[] x, double mu, double sigma)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = NormalPDF(x[i], mu, sigma);
            }
            return y;
        }

        public static double NextNormal(double mu, double sigma)
        {
            double v1 = 0.0, v2 = 0.0, v12 = 0.0, y = 0.0;
            while (v12 >= 1.0 || v12 == 0.0)
            {
                v1 = 2.0 * rand.NextDouble() - 1.0;
                v2 = 2.0 * rand.NextDouble() - 1.0;
                v12 = v1 * v1 + v2 * v2;
            }
            y = Math.Sqrt(-2.0 * Math.Log(v12) / v12);

            return v1 * y * sigma + mu;
        }

        public static double[] NextNormal(double mu, double sigma, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextNormal(mu, sigma);
            }
            return array;
        }

        #endregion

        #region Chi and ChiSquare Distribution

        public static double ChiPDF(double x, int n)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-n * x * x / 2.0);
            return 2.0 * Math.Pow(n / 2.0, n / 2) * Math.Pow(x, n - 1) * exp / Math.Pow(2, n / 2) / gamma;
        }

        public static double ChiPDF(double x, int n, double sigma)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-n * x * x / 2.0 / sigma / sigma);
            return 2.0 * Math.Pow(n / 2.0, n / 2) * Math.Pow(x, n - 1) * exp / Math.Pow(2, n / 2) / gamma / Math.Pow(sigma, n);
        }

        public static double[] ChiPDF(double[] x, int n, double sigma)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = ChiPDF(x[i], n, sigma);
            }
            return y;
        }

        public static double NextChi(int n)
        {
            return Math.Sqrt(NextChiSquare(n) / n);
        }

        public static double NextChi(int n, double sigma)
        {
            return Math.Sqrt(NextChiSquare(n, sigma) / n);
        }

        public static double[] NextChi(int n, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextChi(n);
            }
            return array;
        }

        public static double[] NextChi(int n, double sigma, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextChi(n, sigma);
            }
            return array;
        }

        public static double ChiSquarePDF(double x, int n)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-x / 2.0);
            return Math.Pow(x, n / 2 - 1) * exp / Math.Pow(2, n / 2) / gamma;
        }

        public static double ChiSquarePDF(double x, int n, double sigma)
        {
            double gamma = Gamma(n / 2.0);
            double exp = Math.Exp(-x / 2.0 / sigma / sigma);
            return Math.Pow(x, n / 2 - 1) * exp / Math.Pow(2, n / 2) / gamma / Math.Pow(sigma, n);
        }

        public static double[] ChiSquarePDF(double[] x, int n)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = ChiSquarePDF(x[i], n);
            }
            return y;
        }

        public static double[] ChiSquarePDF(double[] x, int n, double sigma)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = ChiSquarePDF(x[i], n, sigma);
            }
            return y;
        }

        public static double NextChiSquare(int n)
        {
            if (n <= 0) throw new Exception("n must be positive!");

            double result = 0.0;
            for (int i = 0; i < n; i++)
            {
                result += Math.Pow(NextNormal(0, 1), 2);
            }
            return result;
        }

        public static double NextChiSquare(int n, double sigma)
        {
            if (n <= 0) throw new Exception("n must be positive!");
            if (sigma <= 0.0) throw new Exception("sigma must be positive!");

            double result = 0.0;
            for (int i = 0; i < n; i++)
            {
                result += Math.Pow(NextNormal(0, sigma * sigma), 2);
            }
            return result;
        }

        public static double[] NextChiSquare(int n, double sigma, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextChiSquare(n, sigma);
            }
            return array;
        }

        public static double[] NextChiSquare(int n, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextChiSquare(n);
            }
            return array;
        }
      
        #endregion

        #region Exponential Distribution

        public static double ExponentialPDF(double x, double alpha)
        {
            return alpha * Math.Exp(-alpha * x);
        }

        public static double[] ExponentialPDF(double[] x, double alpha)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = ExponentialPDF(x[i], alpha);
            }
            return y;
        }

        public static double NextExponential(double alpha)
        {
            if (alpha <= 0.0) throw new Exception("alpha must be > zero!");
            return -Math.Log(rand.NextDouble()) / alpha;
        }

        public static double[] NextExponential(double alpha, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextExponential(alpha);
            }
            return array;
        }

        #endregion

        #region Student-T Distribution

        public static double StudentTPDF(double x, int n)
        {
            double gamma1 = Gamma((n + 1.0) / 2.0);
            double gamma2 = Gamma(1.0 / 2.0);
            double gamma3 = Gamma(n / 2.0);
            return Math.Pow(n, -0.5) * Math.Pow(1 + x * x / n, -(n + 1) / 2) * gamma1 / gamma2 / gamma3;
        }

        public static double[] StudentTPDF(double[] x, int n)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = StudentTPDF(x[i], n);
            }
            return y;
        }

        public static double NextStudentT(int n)
        {
            if (n <= 0) throw new Exception("n must be positive!");
            return NextNormal(0, 1) / NextChi(n);
        }

        public static double[] NextStudentT(int n, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextStudentT(n);
            }
            return array;
        }

        #endregion

        #region Cauchy Distribution

        public static double CauchyPDF(double x)
        {
            return 1 / (Math.PI * (1.0 + x * x));
        }

        public static double[] CauchyPDF(double[] x)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = CauchyPDF(x[i]);
            }
            return y;
        }

        public static double CauchyPDF(double x, double a, double b)
        {
            return b / (Math.PI * (b * b + (x - a) * (x - a)));
        }

        public static double[] CauchyPDF(double[] x, double a, double b)
        {
            double[] y = new double[x.Length];
            for (int i = 0; i < x.Length; i++)
            {
                y[i] = CauchyPDF(x[i], a, b);
            }
            return y;
        }

        public static double NextCauchy(double a, double b)
        {
            if (b <= 0.0) throw new Exception("b must be positive!");
            return a + b * (Math.Tan(Math.PI * rand.NextDouble() - 0.5));
        }

        public static double[] NextCauchy(double a, double b, int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextCauchy(a, b);
            }
            return array;
        }

        public static double NextCauchy()
        {
            return Math.Tan(Math.PI * rand.NextDouble() - 0.5);
        }

        public static double[] NextCauchy(int nLength)
        {
            double[] array = new double[nLength];
            for (int i = 0; i < nLength; i++)
            {
                array[i] = NextCauchy();
            }
            return array;
        }

        
        #endregion

        #region Random Permutations and Perturbations

        public static int[] RandomPermutation(int n)
        {
            //Create an array of size specified by nSize
            int[] numbers = new int[n];
            //to hold the numbers 0,1,2,...,n
            for (int i = 0; i < numbers.Length; i++)
            {
                numbers[i] = i;
            }
            //Exchange each entry of the array with another 
            //entry located at a random position in the array.
            for (int i = numbers.Length - 1; i > 0; i--)
            {
                int randomPosition = rand.Next(i + 1);
                int temp = numbers[i];
                numbers[i] = numbers[randomPosition];
                numbers[randomPosition] = temp;
            }
            return numbers;
        }

        public static double[] RandomPerturbation(double[] data)
        {
            //Randomize the positions of all data array indices
            int[] randomizedIndicesArray = RandomPermutation(data.Length);
            //then just pick one position and add a little 
            //perturbation value to it.
            data[randomizedIndicesArray[0]] += NextNormal(0, 1) / 100.0;
            return data;
        }

        #endregion

        #endregion

        #region Simulated Annealing Method        

        public delegate double MultiFunction(RVector x);

        public static RVector Anneal(MultiFunction f, double[] x, double TMin, int nMaxIterations)
        {
            //Set the initial "temperature" value of the system
            double T0 = 1.0; 

            //Set the temperature variable equal to the given
            //initial temperature value of the system
            double T = T0;   

            //Calculate the initial function value with the
            //given input configuration array
            double f0 = f(new RVector(x));
            double f1 = 0.0;
            double Deltaf = 0.0;
            double[] xCurrentState;

            int j = 0;

            //Loop until the current "temperature" is less than some
            //user supplied minimum threshold temperature.
            do
            {
                j++;
                int i = 0; //Iteration counter
                //Loop until the iteration counter is < than some
                //minimum tolerance value specified by user
                while (i < nMaxIterations)
                {
                    i++;

                    //Generate a random perturbation of the configuration array
                    //This method is given in chapter 10 which covers the topic 
                    //of random numbers and random distributions
                    xCurrentState = RandomPerturbation((double[])x.Clone());

                    //and update the target function with this new perturbed value
                    //of the configuration array
                    f1 = f(new RVector(xCurrentState));

                    //Calculate the difference between the new and old function
                    //values using the current and the previous perturbed
                    //configuration array values.
                    Deltaf = f1 - f0;

                    //If this difference is < 0 update the previous configuration
                    //and function values with the current ones otherwise anneal
                    //the current difference in function values with the current
                    //temperature, compare the result with a random value and
                    //update the previous configuration and function values with 
                    //the current ones only if the random value is < than the
                    //annealed value.
                    if (Deltaf < 0)
                    {
                        x = xCurrentState;
                        f0 = f1;
                    }
                    else
                    {
                        if (Math.Exp(-Deltaf / T) > rand.NextDouble())
                        {
                            x = xCurrentState;
                            f0 = f1;
                        }
                    }
                }
                //Decrease the system temperature value
                T *= Math.Pow(0.9, j);
            }
            while (TMin < T);
            //Return the latest configuration value
            return new RVector(x);
        }

        static double f1(RVector x)
        {
            //f(x,y) = (1-x)^2 + 100(y-x^2)^2
            return (1 - x[0]) * (1 - x[0]) + 100.0 * (x[1] - x[0] * x[0]) * (x[1] - x[0] * x[0]);
        }

        static double f2(RVector x)
        {
            //f(x) = |x| + sin(x)
            return (Math.Abs(x[0]) + Math.Sin(x[0]));
        }

        static void SimulatedAnnealingTest()
        {
            Console.WriteLine("\nSimulated Annealing Test\n");

            Console.WriteLine("TEST 1: Rosebrock function f(x,y) = (1-x)^2 + 100(y-x^2)^2");
            Console.WriteLine("with known global minimums at f(1,1) = 0");
            //Create input array with initial guesses
            double[] xydata1 = new double[] {1.5, 0.5};
            Console.WriteLine("\nInitial guess (x_0,y_0) = ({0},{1})", xydata1[0], xydata1[1]);
            Console.WriteLine("Running simulated annealing algorithm...please wait");
            //Run the simulated annealing algorithm on initial guess 
            RVector result1 = Anneal(f1, xydata1, 1e-15, 20000);
            //Print out result and function value
            Console.WriteLine("Results obtained from simulated annealing algorithm:");
            Console.WriteLine("\nf(x_min,y_min) = f({0},{1}) = {2}", result1[0], result1[1], f1(result1).ToString());

            Console.WriteLine("\n\nTEST 2: f(x) = |x| + sin(x)");
            Console.WriteLine("with known global minimum at f(0) = 0");
            //Create input array with initial guesses
            double[] xydata2 = new double[] { 8.0 };
            Console.WriteLine("\nInitial guess (x_0) = ({0})", xydata2[0]);
            Console.WriteLine("Running simulated annealing algorithm...please wait");
            //Run the simulated annealing algorithm on initial guess 
            RVector result2 = Anneal(f2, xydata2, 1e-15, 20000);
            //Print out result and function value
            Console.WriteLine("Results obtained from simulated annealing algorithm:");
            Console.WriteLine("\nf(x_min) = f({0}) = {1}", result2[0], f2(result2).ToString());
            
            Console.ReadLine();
        }             

        #endregion

        #region Genetic Algorithm Search Example for Optimum Values

        public delegate double GAFunction(double[] values);

        //Use the Visual C\# built-in random number generator
        //generator or one of your own choosing here. This section
        //was commented out because there is already a global
        //instantiation of a random class object that was
        //initialized ealier in this demonstration file.
        //private static Random rand = new Random();

        //A Genetic Algorithm class
        public class GA
        {
            public double MutationRate;
            public double CrossoverRate;
            public int ChromosomeLength;
            public int PopulationSize;
            public int GenerationSize;
            public double TotalFitness;
            public bool Elitism;
            private ArrayList CurrentGenerationList;
            private ArrayList NextGenerationList;
            private ArrayList FitnessList;
            static private GAFunction getFitness;
            public GAFunction FitnessFunction
            {
                get { return getFitness; }
                set { getFitness = value; }
            }

            //Constructor with user specified crossover rate,
            //mutation rate, population size, generation size
            //and chromosome length.
            public GA(double XoverRate, double mutRate, int popSize, int genSize, int ChromLength)
            {
                Elitism = false;
                MutationRate = mutRate;
                CrossoverRate = XoverRate;
                PopulationSize = popSize;
                GenerationSize = genSize;
                ChromosomeLength = ChromLength;
            }

            // Method which starts the GA executing.
            public void LaunchGA()
            {
                //Create the arrays to hold the fitness, 
                //current and next generation lists
                FitnessList = new ArrayList();
                CurrentGenerationList = new ArrayList(GenerationSize);
                NextGenerationList = new ArrayList(GenerationSize);
                //and initilize the mutation rate.
                Chromosome.ChromosomeMutationRate = MutationRate;

                //Create the initial chromosome population by repeatedly 
                //calling the user supplied fitness function
                for (int i = 0; i < PopulationSize; i++)
                {
                    Chromosome g = new Chromosome(ChromosomeLength, true);
                    CurrentGenerationList.Add(g);
                }

                //Rank the initial chromosome population
                RankPopulation();

                //Loop through the entire generation size creating
                //and evaluating generations of new chromosomes.
                for (int i = 0; i < GenerationSize; i++)
                {
                    CreateNextGeneration();
                    RankPopulation();
                }
            }

            //After ranking all the chromosomes by fitness, use a
            //"roulette wheel" selection method that allocates a large
            //probability of being selected to those chromosomes with the
            //highest fitness. That is, preference in the selection process 
            //is biased towards those chromosomes exhibiting highest fitness.
            private int RouletteSelection()
            {
                double randomFitness = rand.NextDouble() * TotalFitness;
                int idx = -1;
                int mid;
                int first = 0;
                int last = PopulationSize - 1;
                mid = (last - first) / 2;
                while (idx == -1 && first <= last)
                {
                    if (randomFitness < (double)FitnessList[mid])
                    { last = mid; }
                    else if (randomFitness > (double)FitnessList[mid])
                    { first = mid; }
                    mid = (first + last) / 2;
                    if ((last - first) == 1) idx = last;
                }
                return idx;
            }

            // Rank population and then sort it in order of fitness.
            private void RankPopulation()
            {
                TotalFitness = 0;
                for (int i = 0; i < PopulationSize; i++)
                {
                    Chromosome g = ((Chromosome)CurrentGenerationList[i]);
                    g.ChromosomeFitness = FitnessFunction(g.ChromosomeGenes);
                    TotalFitness += g.ChromosomeFitness;
                }
                CurrentGenerationList.Sort(new ChromosomeComparer());
                double fitness = 0.0;
                FitnessList.Clear();
                for (int i = 0; i < PopulationSize; i++)
                {
                    fitness += ((Chromosome)CurrentGenerationList[i]).ChromosomeFitness;
                    FitnessList.Add((double)fitness);
                }
            }

            //Create a new generation of chromosomes. There are many 
            //different ways to do this. The basic idea used here is
            //to first check to see if the elitist flag has been set.
            //If so, then copy the chromosomes from this generation 
            //to the next before looping through the entire chromosome
            //population spawning and mutating children. Finally, if the
            //elitism flag has been set, then copy the best chromosomes 
            //to the new population.
            private void CreateNextGeneration()
            {
                NextGenerationList.Clear();
                Chromosome g = null;
                if (Elitism)
                    g = (Chromosome)CurrentGenerationList[PopulationSize - 1];
                for (int i = 0; i < PopulationSize; i += 2)
                {
                    int pidx1 = RouletteSelection();
                    int pidx2 = RouletteSelection();
                    Chromosome parent1, parent2, child1, child2;
                    parent1 = ((Chromosome)CurrentGenerationList[pidx1]);
                    parent2 = ((Chromosome)CurrentGenerationList[pidx2]);

                    if (rand.NextDouble() < CrossoverRate)
                    { parent1.Crossover(ref parent2, out child1, out child2); }
                    else
                    {
                        child1 = parent1;
                        child2 = parent2;
                    }
                    child1.Mutate();
                    child2.Mutate();
                    NextGenerationList.Add(child1);
                    NextGenerationList.Add(child2);
                }
                if (Elitism && g != null) NextGenerationList[0] = g;
                CurrentGenerationList.Clear();
                for (int i = 0; i < PopulationSize; i++)
                    CurrentGenerationList.Add(NextGenerationList[i]);
            }

            //Extract the best values based on fitness from the current generation.
            //Since the ranking process already sorted the latest current generation 
            //list, just pluck out the best values from the current generation list.
            public void GetBestValues(out double[] values, out double fitness)
            {
                Chromosome g = ((Chromosome)CurrentGenerationList[PopulationSize - 1]);
                values = new double[g.ChromosomeLength];
                g.ExtractChromosomeValues(ref values);
                fitness = (double)g.ChromosomeFitness;
            }
        }

        public class Chromosome
        {
            public double[] ChromosomeGenes;
            public int ChromosomeLength;
            public double ChromosomeFitness;
            public static double ChromosomeMutationRate;

            //Chromosome class constructor
            //Actual functionality is to set up an array 
            //called ChromosomeGenes and depending on the 
            //boolean flag createGenes, it may or may not
            //fill this array with random values from 0 to 1
            //up to some specified ChromosomeLength
            public Chromosome(int length, bool createGenes)
            {
                ChromosomeLength = length;
                ChromosomeGenes = new double[length];
                if (createGenes)
                {
                    for (int i = 0; i < ChromosomeLength; i++)
                        ChromosomeGenes[i] = rand.NextDouble();
                }
            }

            //Creates two offspring children using a single crossover point.
            //The basic idea is to first pick a random position, create two 
            //children and then swap their genes starting from the randomly 
            //picked position point.
            public void Crossover(ref Chromosome Chromosome2, out Chromosome child1, out Chromosome child2)
            {
                int position = (int)(rand.NextDouble() * (double)ChromosomeLength);
                child1 = new Chromosome(ChromosomeLength, false);
                child2 = new Chromosome(ChromosomeLength, false);
                for (int i = 0; i < ChromosomeLength; i++)
                {
                    if (i < position)
                    {
                        child1.ChromosomeGenes[i] = ChromosomeGenes[i];
                        child2.ChromosomeGenes[i] = Chromosome2.ChromosomeGenes[i];
                    }
                    else
                    {
                        child1.ChromosomeGenes[i] = Chromosome2.ChromosomeGenes[i];
                        child2.ChromosomeGenes[i] = ChromosomeGenes[i];
                    }
                }
            }

            //Mutates the chromosome genes by randomly switching them around
            public void Mutate()
            {
                for (int position = 0; position < ChromosomeLength; position++)
                {
                    if (rand.NextDouble() < ChromosomeMutationRate)
                        ChromosomeGenes[position] = (ChromosomeGenes[position] + rand.NextDouble()) / 2.0;
                }
            }

            //Extracts the chromosome values
            public void ExtractChromosomeValues(ref double[] values)
            {
                for (int i = 0; i < ChromosomeLength; i++)
                    values[i] = ChromosomeGenes[i];
            }
        }

        //Compares two chromosomes by their fitness values
        public sealed class ChromosomeComparer : IComparer
        {
            public int Compare(object x, object y)
            {
                if (!(x is Chromosome) || !(y is Chromosome))
                    throw new ArgumentException("Not of type Chromosome");
                if (((Chromosome)x).ChromosomeFitness > ((Chromosome)y).ChromosomeFitness)
                    return 1;
                else if (((Chromosome)x).ChromosomeFitness == ((Chromosome)y).ChromosomeFitness)
                    return 0;
                else
                    return -1;
            }
        }

        public static double GenAlgTestFcn(double[] values)
        {
            if (values.GetLength(0) != 2)
                throw new Exception("should only have 2 args");

            double x = values[0]; double y = values[1];

            return (15 * x * y * (1 - x) * (1 - y) * Math.Sin(Math.PI * x) * Math.Sin(Math.PI * y));
        }

        public static void GeneticAlgorithmSearchForOptimumValueTest()
        {
            Console.WriteLine("\nFinding global optimum values to the function:\n");
            Console.WriteLine("f(x,y) = 15xy(1-x)(1-y)sin(pi*x)sin(pi*y)\n");
            Console.WriteLine("by using a genetic algorithm with initial parameters: \n");
            Console.WriteLine("Crossover\t=80%");
            Console.WriteLine("Mutation\t=5%");
            Console.WriteLine("Population size\t=100");
            Console.WriteLine("Generations\t=2000");
            Console.WriteLine("Chromosome size\t=2\n");
            Console.WriteLine("Actual max values are: x_max = 0.5 and y_max = 0.5\n");

            GA ga = new GA(0.8, 0.05, 100, 2000, 2);
            ga.FitnessFunction = new GAFunction(GenAlgTestFcn);
            ga.Elitism = true;
            ga.LaunchGA();

            double[] values; double fitness;
            ga.GetBestValues(out values, out fitness);

            Console.WriteLine("Calculated max values are: \nx_max = {0} \ny_max = {1}\n",values[0],values[1]);
            Console.WriteLine("f(x_max,y_max) = f({0},{1}) = {2}", values[0], values[1], fitness);
            Console.WriteLine("\nPress ENTER to continue program");
            Console.ReadLine();
        }

        #endregion

        #region Genetic Algorithm Search Example for Strings

        public class strChromosome
        {
            public char[] Value;
            public int Fitness;

            //Constructor
            public strChromosome(int size)
            {
                Value = new char[size];
                Fitness = 0;
            }

            //Gets a random character from the available  
            //range of ASCII character values
            public void RandomValue(Random rand)
            {
                for (int j = 0; j < Value.Length; j++)
                {
                    Value[j] = (char)((126 * rand.NextDouble()) + 32);
                }
            }

            //Calcultes the fitness value by comparing the difference
            //between target and current string values using their
            //ASCII character values one by one.
            public void CalculateFitness(char[] target)
            {
                int fitness = 0;
                for (int j = 0; j < Value.Length; j++)
                {
                    fitness += Math.Abs(Value[j] - target[j]);
                }
                Fitness = fitness;
            }
        }

        public class strGA
        {
            public double elitismRate = 0.10;
            public double mutationRate = 0.25;
            public string targetWord;
            public int maxGenerations;
            public int maxPopulationPerGeneration;
            public List<strChromosome> Population;
            public List<strChromosome> NextGeneration;

            public void LaunchGA()
            {
                //Target string to find
                char[] target = targetWord.ToCharArray();
                //Size of string to find
                int targetSize = target.Length;

                //Instantiate result object to hold results
                GASearchResult result = new GASearchResult();

                //Initialize Population and nextGeneration lists
                Population = new List<strChromosome>(maxPopulationPerGeneration);
                NextGeneration = new List<strChromosome>(maxPopulationPerGeneration);

                //Generate Initial population
                GenerateInitialPopulation(targetSize, maxPopulationPerGeneration);

                //Print headings
                Console.WriteLine("\nString Value\tFitness Value\tGeneration Number\n");
                //Loop to process each generation                
                for (int generationCounter = 0; generationCounter < maxGenerations; generationCounter++)
                {
                    //Calculate fitness
                    CalculateFitness(target);

                    //Extract best values so far
                    result.Best = GetBest();
                    result.GenerationNumber = generationCounter;
                    Console.WriteLine("{0}\t\t{1}\t\t{2}", new String(result.Best.Value), result.Best.Fitness, result.GenerationNumber);

                    //Get current best results and check them against  
                    //target value. If they are equal, then stop
                    if (result.Best.Fitness == 0) break;

                    //Mate population
                    MatePopulation(targetSize, maxPopulationPerGeneration, elitismRate, mutationRate);

                    //and swap them
                    SwapPopulation();
                }
                Console.WriteLine("\nFINAL BEST RESULTS:");
                Console.WriteLine("{0}\t\t{1}\t\t{2}", new String(result.Best.Value), result.Best.Fitness, result.GenerationNumber);
            }

            private void GenerateInitialPopulation(int targetSize, int maxPopulationPerGeneration)
            {
                for (int i = 0; i < maxPopulationPerGeneration; i++)
                {
                    strChromosome c = new strChromosome(targetSize);
                    c.RandomValue(rand);
                    Population.Add(c);
                }
            }

            public class CompareByFitness : IComparer<strChromosome>
            {
                //Compares two string chromosome objects by their fitness values 
                public int Compare(strChromosome obj1, strChromosome obj2)
                {
                    return obj1.Fitness.CompareTo(obj2.Fitness);
                }
            }

            private void CalculateFitness(char[] target)
            {
                foreach (strChromosome c in Population)
                {
                    c.CalculateFitness(target);
                }
            }

            private strChromosome GetBest()
            {
                strChromosome best = null;
                foreach (strChromosome c in Population)
                {
                    if (best == null)
                    {
                        best = c;
                        continue;
                    }
                    if (c.Fitness < best.Fitness) best = c;
                }
                return best;
            }

            private void MatePopulation(int targetSize, int maxPopulationPerGeneration, double elitismRate, double mutationRate)
            {
                int elitSize = (int)(maxPopulationPerGeneration * elitismRate);
                Elitism(elitSize);
                for (int i = elitSize; i < maxPopulationPerGeneration; i++)
                {
                    NextGeneration.Add(new strChromosome(targetSize));
                    int i1 = (int)(rand.NextDouble() * maxPopulationPerGeneration);
                    int i2 = (int)(rand.NextDouble() * maxPopulationPerGeneration);
                    int spos = (int)(rand.NextDouble() * targetSize);
                    NextGeneration[i].Value =
                        (new String(Population[i1].Value).Substring(0, spos) +
                         new String(Population[i2].Value).Substring(spos, targetSize - spos)).ToCharArray();
                    if (rand.NextDouble() < mutationRate) Mutate(NextGeneration[i], targetSize, rand);
                }
            }

            private void Mutate(strChromosome Chromosome, int targetSize, Random rand)
            {
                int ipos = (int)(rand.NextDouble() * targetSize);
                int mutantGene = (int)(rand.NextDouble() * 126) + 32;
                Chromosome.Value[ipos] = (char)mutantGene;
            }

            private void Elitism(int elitSize)
            {
                Population.Sort(new CompareByFitness());
                for (int i = 0; i < elitSize; i++)
                { NextGeneration.Add(Population[i]); }
            }

            private void SwapPopulation()
            {
                Population.Clear();
                foreach (strChromosome c in NextGeneration)
                { Population.Add(c); }
                NextGeneration.Clear();
            }
        }

        public class GASearchResult
        {
            public strChromosome Best;
            public int GenerationNumber;
        }

        public static void GeneticAlgorithmStringSearchTest()
        {
            Console.WriteLine("Testing a genetic algorithm to search a random string for a target value.\n");

            Console.Write("Enter a target string:");
            string targetString = Convert.ToString(Console.ReadLine());

            strGA ga = new strGA();
            ga.targetWord = targetString;
            ga.maxGenerations = 180;
            ga.maxPopulationPerGeneration = 30000;
            ga.LaunchGA();

            Console.WriteLine("\nPress ENTER to terminate program");
            Console.ReadLine();
        }

        #endregion

        static void Main(string[] args)
        {
            SteepestDescentTest();
            revisedSimplexTest();
            SimulatedAnnealingTest();
            GeneticAlgorithmSearchForOptimumValueTest();
            GeneticAlgorithmStringSearchTest();
        }
    }
}
